/**************************************************************

tabparser.c (C-Munipack project)
Copyright (C) 2004 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "allparser.h"
#include "allfile.h"
#include "cmpack_common.h"
#include "cmpack_allfile.h"

#define ALLOC_BY 64

/******************    STRING BUFFER FUNCTIONS    ******************/

/* Initialize a string buffer */
static void str_init(CmpackString *str)
{
	memset(str, 0, sizeof(CmpackString));
}

/* Free allocated memory in a string buffer */
static void str_free(CmpackString *str)
{
	cmpack_free(str->buf);
	str->buf = NULL;
	str->capacity = str->count = 0;
}

/* Clear a string (do not free the memory) */
static void str_clear(CmpackString *str)
{
	str->count = 0;
}

/* Add a character to a string buffer */
static void str_add(CmpackString *str, char ch)
{
	if (str->count >= str->capacity) {
		str->capacity += ALLOC_BY;
		str->buf = (char*) cmpack_realloc(str->buf, str->capacity*sizeof(char));
	}
	str->buf[str->count++] = ch;
}

/* Convert a string buffer to null-terminated string (use cmpack_free to destroy it) */
static char *str_cstr(CmpackString *str)
{
	char *aux = (char*) cmpack_malloc(str->count+1);
	memcpy(aux, str->buf, str->count*sizeof(char));
	aux[str->count] = '\0';
	return aux;
}

/* Convert a string buffer to null-terminated string (use cmpack_free to destroy it) */
static double str_tod(CmpackString *str, double def)
{
	double	val;
	char	*endptr, *aux;

	aux = (char*) cmpack_malloc(str->count+1);
	memcpy(aux, str->buf, str->count*sizeof(char));
	aux[str->count] = '\0';
	val = strtod(aux, &endptr);
	val = (endptr!=aux ? val : def);
	cmpack_free(aux);
	return val;
}

/* Convert a string buffer to null-terminated string (use cmpack_free to destroy it) */
static long str_tol(CmpackString *str, long def)
{
	long 	val;
	char	*endptr, *aux;

	aux = (char*)cmpack_malloc(str->count+1);
	memcpy(aux, str->buf, str->count*sizeof(char));
	aux[str->count] = '\0';
	val = strtol(aux, &endptr, 10);
	val = (endptr!=aux ? val : def);
	cmpack_free(aux);
	return val;
}

/* Convert a string buffer to null-terminated string (use cmpack_free to destroy it) */
static int str_find(CmpackString *str, const char *str2)
{
	int 	val;
	char	*aux;

	aux = (char*)cmpack_malloc(str->count+1);
	memcpy(aux, str->buf, str->count*sizeof(char));
	aux[str->count] = '\0';
	val = strstr(aux, str2)!=NULL;
	cmpack_free(aux);
	return val;
}

/* Trim all trailing white characters from the string */
static void str_rtrim(CmpackString *str)
{
	size_t i;

	for (i=str->count-1; i>=0; i--) {
		if (str->buf[i]<=32)
			str->count--;
		else
			break;
	}
}

/* End of line(file) found, process parsed data for current line. */
static void process_line(CmpackParser *p)
{
	str_clear(&p->key);
	str_clear(&p->val);
	str_clear(&p->jd);
	str_clear(&p->mag);
	str_clear(&p->err);
	p->row++; 
	p->col = 1;
}	

/* Append column */
static void process_frame(CmpackParser *p, CmpackString *jdstr)
{
	if (p->file) {
		double jd = str_tod(jdstr, 0); 
		if (jd>0) {
			p->frame = all_tab_add_frame(&p->file->data, jd);
		}
	}
}

/* Append parameter */
static void process_val(CmpackParser *p, CmpackString *key, CmpackString *val)
{
	char *skey;

	if (p->file) {
		skey = str_cstr(key);
		if (skey) {
			if (strcmp(skey, "Aperture")==0)
				p->file->aperture = str_tol(val, 0);
			else if (strcmp(skey, "Filter")==0)
				p->file->filter = str_cstr(val);
			else if (strcmp(skey, "JD")==0) 
				p->file->helcor = str_find(val, "heliocentric");
			cmpack_free(skey);
		}
	}
}

/* Append star */
static void process_mag(CmpackParser *p, CmpackString *mag, CmpackString *err)
{
	int mask;
	double dmag, derr;

	if (p->frame>=0 && p->star>=0 && p->star<65536) {
		dmag = str_tod(mag, 99.999);
		derr = str_tod(err, 9.999);
		mask = ((dmag<99 && dmag>-99) ? CMPACK_MASK_MAG : 0) |
				((derr>=0 && derr<9) ? CMPACK_MASK_ERR : 0);
		all_tab_set_star(&p->file->data, p->frame, p->star, mask, dmag, derr);
	}
}

/*********************   PUBLIC FUNCTIONS   ****************************/

/* Init parser context */
void all_parser_init(CmpackParser *p, struct _CmpackAllFile *file)
{
	memset(p, 0, sizeof(CmpackParser));
	p->file = file;
	str_init(&p->key);
	str_init(&p->val);
	str_init(&p->jd);
	str_init(&p->mag);
	str_init(&p->err);
	p->row = p->col = 1;
	p->frame = -1;
}

/* Free asociated memory in the context */
void all_parser_clear(CmpackParser *p)
{
	str_free(&p->key);	
	str_free(&p->val);	
	str_free(&p->jd);	
	str_free(&p->mag);	
	str_free(&p->err);	
	p->file = NULL;
	p->frame = -1;
}
	
/* Parse buffer */
int all_parser_parse(CmpackParser *p, const char *buf, size_t buflen, int eof)
{
	size_t i;
	char ch;

	if (p==NULL || buf==NULL)
		return 0; 
	
	for (i=0; i<buflen; i++) {
		/* Take next character from the buffer */
		ch = buf[i]; 
		/* Finite automaton which processes the characters */
		switch (p->st) 
		{
		case 0:
			/* Skip first line (column names) */
			p->col = 1;	
			/* First character on the line */
			if (ch==13) {
				/* Skip empty line */
				process_line(p);
				p->st = 10;
			} else
			if (ch==10) {
				/* Skip empty line */
				if (p->lastch!=13) {
					process_line(p);
					p->st = 10;
				}
			}
		 	break;
		 	
		case 10:
			/* Start of new line */
			p->col = 1;	
			/* First character on the line */
		 	if ((ch>='A' && ch<='Z') || (ch>='a' && ch<='z') || (ch=='_')) {
				/* Start of a keyword */
				str_clear(&p->key);
				str_add(&p->key, ch);
				p->st = 12;
			} else
			if (ch==13) {
				/* Empty line ends the header */
				process_line(p);
				p->st = 40;
			} else
			if (ch==10) {
				if (p->lastch!=13) {
					process_line(p);
					p->st = 40;
				}
			} else {
				/* Write space */
				p->st = 11;
			}
		 	break;
		 	
		case 11: 
			/* Skip leading spaces */
		 	if ((ch>='A' && ch<='Z') || (ch>='a' && ch<='z') || (ch=='_')) {
				/* Start of a keyword */
				str_clear(&p->key);
				str_add(&p->key, ch);
				p->st = 12;
			} else
			if (ch==13 || ch==10) {
				/* Line with spaces only */
				process_line(p);
				p->st = 40;
			}
		 	break;

		case 12: 
		 	/* Reading keyword */
			if (ch==':') {
				/* End of keyword, old format, wait for value */
				str_rtrim(&p->key);
				p->st = 20;
		 	} else 
			if (ch==13 || ch==10) {
				/* End of line (ignore it) */
				process_line(p);
				p->st = 40;
			} else {
				/* Next character in keyword */
				str_add(&p->key, ch);
			}
			break;

		case 20:
			/* Old format */
			if (ch==13 || ch==10) {
				/* End of line */
				process_line(p);
				p->st = 40;
			} else
		 	if (ch!=' ' && ch!='\t') {
				/* Start of value (unquoted) */
				str_clear(&p->val);
				str_add(&p->val, ch); 
				p->st = 21;
			}
		 	break;
		 	
		case 21: 
			/* Reading value */
			if (ch==',') {
				/* End of value */
				str_rtrim(&p->val);
				process_val(p, &p->key, &p->val);
				p->st = 22;
			} else
			if (ch==13 || ch==10) {
				/* End of line */
				str_rtrim(&p->val);
				process_val(p, &p->key, &p->val);
				process_line(p);
				p->st = 40;
			} else {
				str_add(&p->val, ch);
			}
		 	break;

		case 22:
			/* Waiting for next keyword */
		 	if ((ch>='A' && ch<='Z') || (ch>='a' && ch<='z') || (ch=='_')) {
				/* Start of a keyword */
				str_clear(&p->key);
				str_add(&p->key, ch);
				p->st = 12;
			} else
			if (ch==13 || ch==10) {
				/* Line with spaces only */
				process_line(p);
				p->st = 40;
			}
			break;

		case 40:
			/* Waiting for table data */
			p->col = 1;	
			/* First character on the line */
		 	if (ch>='0' && ch<='9') {
				/* Start of a value */
				str_clear(&p->jd);
				str_add(&p->jd, ch);
				p->st = 42;
			} else
			if (ch==13) {
				/* Empty line ends the header */
				process_line(p);
			} else
			if (ch==10) {
				if (p->lastch!=13) {
					process_line(p);
				}
			} else {
				/* White space */
				p->st = 41;
			}
		 	break;

		case 41:
			/* Waiting for table data */
		 	if (ch>='0' && ch<='9') {
				/* Start of a keyword */
				str_clear(&p->jd);
				str_add(&p->jd, ch);
				p->st = 42;
			} else
			if (ch==13 || ch==10) {
				/* Line with spaces only */
				process_line(p);
				p->st = 40;
			}
		 	break;

		case 42:
			/* Reading value */
			if (ch==' ') {
				/* End of value */
				process_frame(p, &p->jd);
				p->star = 0;
				p->st = 43;
		 	} else 
			if (ch==13 || ch==10) {
				/* End of value */
				process_frame(p, &p->jd);
				p->star = 0;
				process_line(p);
				p->st = 40;
			} else {
				/* Next character in keyword */
				str_add(&p->jd, ch);
			}
			break;

		case 43:
			/* Waiting for next keyword */
		 	if ((ch>='0' && ch<='9') || (ch=='+') || (ch=='-')) {
				/* Start of a keyword */
				str_clear(&p->mag);
				str_add(&p->mag, ch);
				p->st = 44;
			} else
			if (ch==13 || ch==10) {
				/* Line with spaces only */
				process_line(p);
				p->st = 40;
			}
			break;

		case 44:
			/* Reading next values */
			if (ch==' ') {
				/* End of value */
				p->st = 45;
		 	} else 
			if (ch==13 || ch==10) {
				/* End of value */
				process_line(p);
				p->st = 40;
			} else {
				/* Next character in keyword */
				str_add(&p->mag, ch);
			}
			break;

		case 45:
			/* Waiting for next keyword */
		 	if ((ch>='0' && ch<='9') || (ch=='+') || (ch=='-')) {
				/* Start of a keyword */
				str_clear(&p->err);
				str_add(&p->err, ch);
				p->st = 46;
			} else
			if (ch==13 || ch==10) {
				/* Line with spaces only */
				process_line(p);
				p->st = 40;
			}
			break;

		case 46:
			/* Reading next values */
			if (ch==' ') {
				/* End of value */
				process_mag(p, &p->mag, &p->err);
				p->star++;
				p->st = 43;
		 	} else 
			if (ch==13 || ch==10) {
				/* End of value */
				process_mag(p, &p->mag, &p->err);
				process_line(p);
				p->st = 40;
			} else {
				/* Next character in keyword */
				str_add(&p->err, ch);
			}
			break;
		}
		/* Remember last character */
		p->lastch = ch;
	}

	if (eof && p->st!=0) {
		/* Last line, which doesn't end with line delimiter */
		p->st = 0;
	}
	return 0;
}
