/**************************************************************

ccdfile.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "comfun.h"
#include "console.h"
#include "trajd.h"
#include "cmpack_common.h"
#include "cmpack_ccdfile.h"
#include "ccdfile.h"
#include "konvfits.h"
#include "konvsbig.h"
#include "konvoes.h"

#define BLOCK_SIZE 4096

/************************   Table of supported formats   ***************************/

/* Registration of known image formats */
typedef struct _CmpackCcdFmt
{
	CmpackCcdFormat format;									/**< Format identifier */
	int (*test)(const char *, size_t);						/**< Test image format */
	int (*open)(tHandle*, const char*, CmpackOpenMode);		/**< Open a file */
	tHandle (*init)(void);									/**< Create a new file */
	void (*close)(tHandle);									/**< Close a file */
	char* (*get_magic)(tHandle);							/**< Get magic string */
	int (*get_size)(tHandle, int*, int *);					/**< Get image size */
	CmpackBitpix (*get_bitpix)(tHandle);					/**< Get image data type */
	int (*get_range)(tHandle, double*, double*);			/**< Get image data range */
	int (*get_image)(tHandle, CmpackBitpix, int, int, void*);		/**< Get image data */
	int (*get_datetime)(tHandle, CmpackDateTime*);			/**< Get date of observation */
	int (*get_exptime)(tHandle, double*);					/**< Get exposure duration */
	int (*get_ccdtemp)(tHandle, double*);					/**< Get CCD temperature */
	char* (*get_filter)(tHandle);							/**< Get optical filter name */
	char* (*get_object)(tHandle);							/**< Get object's designation */
	char* (*get_objra)(tHandle);							/**< Get object's right ascension */
	char* (*get_objdec)(tHandle);							/**< Get object's declination */
	char* (*get_observer)(tHandle);							/**< Get observer's name */
	char* (*get_observatory)(tHandle);						/**< Get observatory name */
	char* (*get_obslon)(tHandle);							/**< Get observer's longitude */
	char* (*get_obslat)(tHandle);							/**< Get observer's latitude */
	int (*get_frames)(tHandle, int*);						/**< Get number of combined frames */
	int (*get_str)(tHandle, const char*, char**);			/**< Get a string value */
	int (*get_int)(tHandle, const char*, int*);				/**< Get a integer value */
	int (*get_dbl)(tHandle, const char*, double*);			/**< Get a real value */
	int (*get_bool)(tHandle, const char*, int*);			/**< Get a logic value */
	int (*get_param)(tHandle, int, char**, char**, char**);	/**< Parameter enumeration */
	int (*copy)(tHandle, fitsfile*, CmpackBitpix, int,
			  int, const CmpackBorder*, CmpackConsole*);	/**< Copy content */
	int (*copy_header)(tHandle, fitsfile*);					/**< Copy header to FITS file */
	fitsfile *(*get_fits)(tHandle);							/**< Get pointer to fits file */
	int (*set_datetime)(tHandle, const CmpackDateTime*);	/**< Set date of observation */
	void (*set_exptime)(tHandle, double);					/**< Set exposure duration */
	void (*set_ccdtemp)(tHandle, double);					/**< Set CCD temperature */
	void (*set_frames)(tHandle, int);						/**< Set number of combined frames */
	void (*set_object)(tHandle, const char*);				/**< Set object's designtion */
	void (*set_str)(tHandle, const char*, const char*, const char*);	/**< Set a string value */
	void (*set_int)(tHandle, const char*, int, const char*);	/**< Set a string value */
	void (*set_dbl)(tHandle, const char*, double, int, const char*);	/**< Set a string value */
	void (*set_bool)(tHandle, const char*, int, const char*);			/**< Set a logic value */
	int (*put_history)(tHandle, const char*);				/**< Parameter enumeration */
	void (*set_origin)(tHandle);							/**< Set package name and version */
	void (*set_pcdate)(tHandle);							/**< Set current date */
	int (*prepare)(tHandle, int, int, CmpackBitpix);		/**< Set image parameters */
	int (*put_image)(tHandle, CmpackBitpix, const void*);	/**< Set image data */
	int (*clear_header)(tHandle);							/**< Clear file header */
} CmpackCcdFmt;

static const CmpackCcdFmt flist[] = {
	{ CMPACK_FORMAT_FITS, fits_test, fits_open, fits_init, fits_close, fits_getmagic, 
		fits_getsize, fits_getbitpix, fits_getrange, fits_getimage, fits_getdatetime, 
		fits_getexptime, fits_getccdtemp, fits_getfilter, fits_getobject, fits_getobjra, 
		fits_getobjdec, fits_getobserver, fits_getobservatory, fits_getobslon, fits_getobslat, 
		fits_getframes, fits_gkys, fits_gkyi, fits_gkyd, fits_gkyl, fits_gkyn, fits_copy, 
		fits_copyheader, fits_getfits, fits_setdatetime, fits_setexptime, fits_setccdtemp, 
		fits_setframes, fits_setobject, fits_pkys, fits_pkyi, fits_pkyd, fits_pkyl, 
		fits_puthistory, fits_setorigin, fits_setpcdate, fits_prepare, fits_putimage, 
		fits_clearheader },
	{ CMPACK_FORMAT_SBIG, sbig_test, sbig_open, NULL, sbig_close, sbig_getmagic, sbig_getsize, 
		sbig_getbitpix, sbig_getrange, sbig_getimage, sbig_getdatetime, sbig_getexptime, 
		sbig_getccdtemp, sbig_getfilter, NULL, NULL, NULL, sbig_getobserver, NULL, 
		NULL, NULL, NULL, sbig_gets, sbig_geti, sbig_getd, NULL, sbig_gkyn, sbig_copy, 
		sbig_copyheader },
	{ CMPACK_FORMAT_OES, oes_test, oes_open, NULL, oes_close, oes_getmagic, oes_getsize,
		oes_getbitpix, oes_getrange, oes_getimage, oes_getdatetime, oes_getexptime, 
		oes_getccdtemp, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
		oes_gets, oes_geti, oes_getd, NULL, oes_gkyn, oes_copy, oes_copyheader },
	{ 0 }
};

/************************   Private data types   ***************************/

/* Image file context */
struct _CmpackCcdFile
{
	CmpackContext gc;					/**< Global context */
	const CmpackCcdFmt *fmt;			/**< Access routines */
	tHandle handle;						/**< Format specific handle */
	char *magic, *filter;
	char *objname, *objra, *objdec;
	char *observer, *observatory, *obslon, *obslat;
};

/************************   Local data and functions   ****************************/

static void file_clean(void *ctx);

/* Create a new CCD file context */
static CmpackCcdFile *file_create(void)
{
	CmpackCcdFile *f = (CmpackCcdFile*)gc_create(sizeof(CmpackCcdFile));
	f->gc.clean = file_clean;
	return f;
}

/* Destroy a CCD file context */
static void file_close(CmpackCcdFile *f)
{
	if (f->fmt) {
		f->fmt->close(f->handle);
		f->fmt = NULL;
		f->handle = NULL;
	}
	cmpack_free(f->magic);
	f->magic = NULL;
	cmpack_free(f->filter);
	f->filter = NULL;
	cmpack_free(f->objname);
	f->objname = NULL;
	cmpack_free(f->objra);
	f->objra = NULL;
	cmpack_free(f->objdec);
	f->objdec = NULL;
	cmpack_free(f->observer);
	f->observer = NULL;
	cmpack_free(f->observatory);
	f->observatory = NULL;
	cmpack_free(f->obslon);
	f->obslon = NULL;
	cmpack_free(f->obslat);
	f->obslat = NULL;
}

/* Destroy a CCD file context */
static void file_clean(void *ctx)
{
	file_close((CmpackCcdFile*)ctx);
	gc_clean(ctx);
}
	

/************************   Public functions   **************************************/

/* Determines file format. */
int cmpack_ccd_test(const char *filename)
{
	char buf[BLOCK_SIZE+1];
	int res = 0, filesize, bytes;
	const CmpackCcdFmt *fmt;

	/* Test if source file is present and accessible */
	if (filename) {
		FILE *f = fopen(filename, "rb");
		if (f) {
			fseek(f, 0, SEEK_END);
			filesize = ftell(f);
			fseek(f, 0, SEEK_SET);
			bytes = (int)fread(buf, 1, BLOCK_SIZE, f);
			buf[bytes] = 0;
			for (fmt=flist; fmt->format!=0; fmt++) {
				if (fmt->test) {
					if (fmt->test(buf, filesize)) {
						res = 1;
						break;
					}
				}
			}
			fclose(f);
		}
	}
	return res;
}

/* Determines file format. Return 0 if successful, or error code if not. */
int cmpack_ccd_test_buffer(const char *buffer, int buflen, int filesize)
{
	int res = 0;
	char *aux;
	const CmpackCcdFmt *fmt;

	if (buffer && buflen>16) {
		/* Convert buffer to a null-terminated string */
		aux = (char*)cmpack_malloc((buflen+1)*sizeof(char));
		memcpy(aux, buffer, buflen);
		aux[buflen] = '\0';
		for (fmt=flist; fmt->test!=NULL; fmt++) {
			if (fmt->test(aux, filesize)) {
				res = 1;
				break;
			}
		}
		cmpack_free(aux);
	}
	return res;
}

/* Create a temporary file */
CmpackCcdFile *cmpack_ccd_new(void)
{
	CmpackCcdFile *fs;
	const CmpackCcdFmt *fmt;

	for (fmt=flist; fmt->test!=NULL; fmt++) {
		if (fmt->format == CMPACK_FORMAT_FITS) {
			fs = file_create();
			fs->fmt = fmt;
			fs->handle = fmt->init();
			return fs;
		}
	}
	return NULL;
}

/* Open a source file (format is specified) */
int cmpack_ccd_open(CmpackCcdFile **fc, const char *filename, CmpackOpenMode flags)
{
	int res;
	char buf[BLOCK_SIZE+1];
	CmpackCcdFile *fs;
	size_t filesize, bytes;
	FILE *f;
	const CmpackCcdFmt *fmt, *format;

	*fc = NULL;

	/* Check source file */
	if (!filename) 
		return CMPACK_ERR_INVALID_PAR;
	if (strlen(filename)==0)
		return CMPACK_ERR_INVALID_PAR;

	format = NULL;
	if ((flags & CMPACK_OPEN_CREATE)==0) {
		/* Determine file format */
		if (flags & CMPACK_OPEN_READONLY)
			f = fopen(filename, "rb");
		else
			f = fopen(filename, "rb+");
		if (!f)
			return CMPACK_ERR_CANT_OPEN_SRC;
		fseek(f, 0, SEEK_END);
		filesize = ftell(f);
		fseek(f, 0, SEEK_SET);
		bytes = (int)fread(buf, 1, BLOCK_SIZE, f);
		buf[bytes] = 0;
		for (fmt=flist; fmt->format!=0; fmt++) {
			if (fmt->test) {
				if (fmt->test(buf, filesize)) {
					format = fmt;
					break;
				}
			}
		}
		fclose(f);
	} else {
		/* New file is always FITS */
		for (fmt=flist; fmt->format!=0; fmt++) {
			if (fmt->format == CMPACK_FORMAT_FITS)
				format = fmt;
		}
	}
	if (!format)
		return CMPACK_ERR_UNKNOWN_FORMAT;

	/* Create context */
	fs = file_create();
	res = format->open(&fs->handle, filename, flags);
	if (res!=0) {
		cmpack_unref(fs);
		return res;
	}

	/* Normal return */
	fs->fmt = format;
	*fc = fs;
	return 0;
}

/* Read string parameter from file header */
int cmpack_ccd_get_str(CmpackCcdFile *fs, const char *key, char **val)
{
	if (!fs->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fs->fmt->get_str)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else
		return fs->fmt->get_str(fs, key, val);
}

/* Read double parameter from file header */
int cmpack_ccd_get_dbl(CmpackCcdFile *fs, const char *key, double *val)
{
	if (!fs->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fs->fmt->get_dbl)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else
		return fs->fmt->get_dbl(fs, key, val);
}

/* Read integer parameter from file header */
int cmpack_ccd_get_int(CmpackCcdFile *fs, const char *key, int *val)
{
	if (!fs->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fs->fmt->get_int)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else
		return fs->fmt->get_int(fs, key, val);
}

/* Read integer parameter from file header */
int cmpack_ccd_get_bool(CmpackCcdFile *fs, const char *key, int *val)
{
	if (!fs->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fs->fmt->get_bool)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else
		return fs->fmt->get_bool(fs->handle, key, val);
}

/* Get frame width */
int cmpack_ccd_get_size(CmpackCcdFile *fc, int *width, int *height)
{
	if (!fc->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fc->fmt->get_size)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else
		return fc->fmt->get_size(fc->handle, width, height);
		
}

/* Read string parameter from file header */
const char *cmpack_ccd_get_format(CmpackCcdFile *fs)
{
	if (!fs->magic && fs->fmt && fs->fmt->get_magic)
		fs->magic = fs->fmt->get_magic(fs->handle);
	return fs->magic;
}

/* Read string parameter from file header */
const char *cmpack_ccd_get_filter(CmpackCcdFile *fs)
{
	if (!fs->filter && fs->fmt && fs->fmt->get_filter)
		fs->filter = fs->fmt->get_filter(fs->handle);
	return fs->filter;
}

/* Get object's designation */
const char *cmpack_ccd_get_object(CmpackCcdFile *fs)
{
	if (!fs->objname && fs->fmt && fs->fmt->get_object) 
		fs->objname = fs->fmt->get_object(fs->handle);
	return fs->objname;
}

/* Read string parameter from file header */
const char *cmpack_ccd_get_objra(CmpackCcdFile *fs)
{
	if (!fs->objra && fs->fmt && fs->fmt->get_objra) 
		fs->objra = fs->fmt->get_objra(fs->handle);
	return fs->objra;
}

/* Get object's declination */
const char *cmpack_ccd_get_objdec(CmpackCcdFile *fs)
{
	if (!fs->objdec && fs->fmt && fs->fmt->get_objdec) 
		fs->objdec = fs->fmt->get_objdec(fs->handle);
	return fs->objdec;
}

/* Read string parameter from file header */
const char *cmpack_ccd_get_observer(CmpackCcdFile *fs)
{
	if (!fs->observer && fs->fmt && fs->fmt->get_observer)
		fs->observer = fs->fmt->get_observer(fs->handle);
	return fs->observer;
}

/* Read string parameter from file header */
const char *cmpack_ccd_get_observatory(CmpackCcdFile *fs)
{
	if (!fs->observatory && fs->fmt && fs->fmt->get_observatory)
		fs->observatory = fs->fmt->get_observatory(fs->handle);
	return fs->observatory;
}

/* Read string parameter from file header */
const char *cmpack_ccd_get_obslon(CmpackCcdFile *fs)
{
	if (!fs->obslon && fs->fmt && fs->fmt->get_obslon)
		fs->obslon = fs->fmt->get_obslon(fs->handle);
	return fs->obslon;
}

/* Read string parameter from file header */
const char *cmpack_ccd_get_obslat(CmpackCcdFile *fs)
{
	if (!fs->obslat && fs->fmt && fs->fmt->get_obslat)
		fs->obslat = fs->fmt->get_obslat(fs->handle);
	return fs->obslat;
}

/* Read string parameter from file header */
int cmpack_ccd_get_datetime(CmpackCcdFile *fs, CmpackDateTime *dt)
{
	if (!fs->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fs->fmt->get_datetime)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else 
		return fs->fmt->get_datetime(fs->handle, dt);
}

/* Read string parameter from file header */
int cmpack_ccd_get_exptime(CmpackCcdFile *fs, double *exptime)
{
	if (!fs->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fs->fmt->get_exptime)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else 
		return fs->fmt->get_exptime(fs->handle, exptime);
}

/* Read string parameter from file header */
int cmpack_ccd_get_ccdtemp(CmpackCcdFile *fs, double *ccdtemp)
{
	if (!fs->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	if (!fs->fmt->get_ccdtemp)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else
		return fs->fmt->get_ccdtemp(fs->handle, ccdtemp);
}

/* Get number of subframes averaged */
int cmpack_ccd_get_frames(CmpackCcdFile *fs, int *frames)
{
	if (!fs->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fs->fmt->get_frames)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else 
		return fs->fmt->get_frames(fs->handle, frames);
}

/* Read parameter by its index */
int cmpack_ccd_get_param(CmpackCcdFile *fs, int index, char **key, char **val, char **com)
{
	if (!fs->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fs->fmt->get_param)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else 
		return fs->fmt->get_param(fs->handle, index, key, val, com);
}

/* Convert file header to FITS file */
CmpackBitpix cmpack_ccd_get_bitpix(CmpackCcdFile *fs)
{
	if (fs->fmt && fs->fmt->get_bitpix)
		return fs->fmt->get_bitpix(fs->handle);
	else
		return CMPACK_BITPIX_UNKNOWN;
}

/* Convert file header to FITS file */
int cmpack_ccd_get_range(CmpackCcdFile *fs, double *minvalue, double *maxvalue)
{
	if (!fs->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fs->fmt->get_range)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else 
		return fs->fmt->get_range(fs->handle, minvalue, maxvalue);
}

/* Read image data in double precision */
int cmpack_ccd_get_image(CmpackCcdFile *fs, CmpackBitpix bitpix, int flipx, int flipy, void *buf)
{
	if (!fs->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fs->fmt->get_image)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else
		return fs->fmt->get_image(fs->handle, bitpix, flipx, flipy, buf);
}

/* Copy entire file */
int ccd_copy(CmpackCcdFile *source, CmpackCcdFile *target,
		CmpackBitpix bitpix, int flipx, int flipy, const CmpackBorder *border, 
		CmpackConsole *console)
{
	if (!source->fmt || !target->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!source->fmt->copy || !target->fmt->get_fits)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else {
		return source->fmt->copy(source->handle, target->fmt->get_fits(target->handle),
			bitpix, flipx, flipy, border, console);
	}
}

/* Copy entire file */
int cmpack_ccd_copy(CmpackCcdFile *source, CmpackCcdFile *target)
{
	if (!source->fmt || !target->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!source->fmt->copy || !target->fmt->get_fits)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else {
		return source->fmt->copy(source->handle, target->fmt->get_fits(target->handle),
			CMPACK_BITPIX_AUTO, 0, 0, NULL, NULL);
	}
}

/* Copy entire file */
int ccd_copy_header(CmpackCcdFile *fs, CmpackCcdFile *target)
{
	if (!fs->fmt || !target->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fs->fmt->copy_header || !target->fmt->get_fits)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else
		return fs->fmt->copy_header(fs->handle, target->fmt->get_fits(target->handle));
}

/* Cleear file header */
void ccd_clear_header(CmpackCcdFile *fs)
{
	if (fs->fmt && fs->fmt->clear_header)
		fs->fmt->clear_header(fs->handle);
}

/* Append text to the history */
int ccd_write_history(CmpackCcdFile *fs, const char *text)
{
	if (!fs->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fs->fmt->put_history)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else 
		return fs->fmt->put_history(fs->handle, text);
}

/* Set date of observation */
int ccd_set_datetime(CmpackCcdFile *fs, const CmpackDateTime *dt)
{
	if (!fs->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fs->fmt->put_history)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else 
		return fs->fmt->set_datetime(fs->handle, dt);
}

/* Set exposure duration */
void ccd_set_exptime(CmpackCcdFile *fs, double exptime)
{
	if (fs->fmt && fs->fmt->set_exptime)
		fs->fmt->set_exptime(fs->handle, exptime);
}

/* Set exposure duration */
void ccd_set_ccdtemp(CmpackCcdFile *fs, double ccdtemp)
{
	if (fs->fmt && fs->fmt->set_ccdtemp)
		fs->fmt->set_ccdtemp(fs->handle, ccdtemp);
}

/* Set number of frames combined */
void ccd_set_frames(CmpackCcdFile *fs, int frames)
{
	if (fs->fmt && fs->fmt->set_frames)
		fs->fmt->set_frames(fs->handle, frames);
}

/* Set image parameters */
int ccd_prepare(CmpackCcdFile *fs, int nx, int ny, CmpackBitpix bitpix)
{
	if (!fs->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fs->fmt->prepare)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else if (nx<=0 || ny<=0 || nx>=65536 || ny>=65536)
		return CMPACK_ERR_INVALID_SIZE;
	else if (pixformat(bitpix)==NULL)
		return CMPACK_ERR_INVALID_BITPIX;
	else
		return fs->fmt->prepare(fs->handle, nx, ny, bitpix);
}

/* Set image data */
int ccd_write_image(CmpackCcdFile *fs, CmpackBitpix bitpix, const void *buf)
{
	if (!fs->fmt)
		return CMPACK_ERR_CLOSED_FILE;
	else if (!fs->fmt->put_image)
		return CMPACK_ERR_NOT_IMPLEMENTED;
	else
		return fs->fmt->put_image(fs->handle, bitpix, buf);
}

/* Set object's designation */
void ccd_set_object(CmpackCcdFile *fs, const char *name)
{
	if (fs->fmt && fs->fmt->set_object)
		fs->fmt->set_object(fs->handle, name);
}

/* Set parameter (string value) */
void ccd_set_str(CmpackCcdFile *fs, const char *key, const char *value, const char *comment)
{
	if (fs->fmt && fs->fmt->set_str)
		fs->fmt->set_str(fs->handle, key, value, comment);
}

/* Set parameter (integer value) */
void ccd_set_int(CmpackCcdFile *fs, const char *key, int value, const char *comment)
{
	if (fs->fmt && fs->fmt->set_int)
		fs->fmt->set_int(fs->handle, key, value, comment);
}

/* Set parameter (real value) */
void ccd_set_dbl(CmpackCcdFile *fs, const char *key, double value, int prec, const char *comment)
{
	if (fs->fmt && fs->fmt->set_dbl)
		fs->fmt->set_dbl(fs->handle, key, value, prec, comment);
}

/* Set parameter (logic value) */
void ccd_set_bool(CmpackCcdFile *fs, const char *key, int value, const char *comment)
{
	if (fs->fmt && fs->fmt->set_bool)
		fs->fmt->set_bool(fs->handle, key, value, comment);
}

/* Write package name and version to file header */
void ccd_set_origin(CmpackCcdFile *fs)
{
	if (fs->fmt && fs->fmt->set_origin)
		fs->fmt->set_origin(fs->handle);
}

/* Write date of creation */
void ccd_set_pcdate(CmpackCcdFile *fs)
{
	if (fs->fmt && fs->fmt->set_pcdate)
		fs->fmt->set_pcdate(fs->handle);
}


/* Print short info */
int cmpack_ccd_print_info(CmpackCcdFile *file, CmpackConsole *con)
{
  	const char *magic, *filter, *objname, *objra, *objdec;
	const char *observer, *observatory, *obslon, *obslat;
	char msg[MAXLINE], datestr[64], timestr[64];
  	int width, height;
  	double temp, exp;
	CmpackDateTime dt;

  	/* Check parameters */
	if (!file || !con) 
		return CMPACK_ERR_INVALID_PAR;

	/* Print short info */
	magic = cmpack_ccd_get_format(file);
	if (magic) {
		sprintf(msg,"File format: %s", magic);
		printout(con, 0, msg);
	}
	if (cmpack_ccd_get_size(file, &width, &height)==0) {
  		sprintf(msg,"Image size : %d x %d pixels", width, height);
		printout(con, 0, msg);
	}
	if (cmpack_ccd_get_datetime(file, &dt)==0) {
		cmpack_datetostr(&dt.date, datestr, 64);
		cmpack_timetostr(&dt.time, timestr, 64);
  		sprintf(msg,"Date & time: %s %s UT" , datestr, timestr);
		printout(con, 0, msg);
  	}
	if (cmpack_ccd_get_exptime(file, &exp)==0) {
	  	sprintf(msg,"Exposure   : %.2f s", exp);
		printout(con, 0, msg);
  	}
	if (cmpack_ccd_get_ccdtemp(file, &temp)==0) {
	  	sprintf(msg,"Temperature: %.1f deg.C", temp);
		printout(con, 0, msg);
  	}
	filter = cmpack_ccd_get_filter(file);
	if (filter) {
      	sprintf(msg,"Filter     : %s", filter);
		printout(con, 0, msg);
  	}
	objname = cmpack_ccd_get_object(file);
	if (objname) {
      	sprintf(msg,"Object     : %s", objname);
		printout(con, 0, msg);
	}
	objra = cmpack_ccd_get_objra(file);
	if (objra) {
      	sprintf(msg,"R.A.       : %s", objra);
		printout(con, 0, msg);
	}
	objdec = cmpack_ccd_get_objdec(file);
	if (objdec) {
      	sprintf(msg,"Declination: %s", objdec);
		printout(con, 0, msg);
	}
	observer = cmpack_ccd_get_observer(file);
	if (observer) {
      	sprintf(msg,"Observer   : %s", observer);
		printout(con, 0, msg);
	}
	observatory = cmpack_ccd_get_observer(file);
	if (observatory) {
      	sprintf(msg,"Observatory: %s", observatory);
		printout(con, 0, msg);
	}
	obslon = cmpack_ccd_get_obslon(file);
	if (obslon) {
      	sprintf(msg,"Longitude  : %s", obslon);
		printout(con, 0, msg);
	}
	obslat = cmpack_ccd_get_obslat(file);
	if (obslat) {
      	sprintf(msg,"Latitude   : %s", obslat);
		printout(con, 0, msg);
	}
 	return 0;
}

/* Print full header */
int cmpack_ccd_print_head(CmpackCcdFile *file, CmpackConsole *con)
{
	int i;
	char msg[MAXLINE], *key, *val, *com;

  	/* Check parameters */
	if (!file || !con) 
		return CMPACK_ERR_INVALID_PAR;

	/* Print full info */
	sprintf(msg,"%-20s %-24s %s","---Key---","---Value---","---Comments---");
	printout(con, 0, msg);
	for (i=0; cmpack_ccd_get_param(file, i, &key, &val, &com)==0; i++) {
		sprintf(msg,"%-20.20s %-24.24s %-32.32s",
			(key ? key : ""), (val ? val : ""), (com ? com : ""));
		printout(con, 0, msg);
		cmpack_free(key);
		cmpack_free(val);
		cmpack_free(com);
	}
	return 0;
}
