/**************************************************************

phfile_file.c (C-Munipack project)
Photometry file read/write routines
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <expat.h>
#include <math.h>
#include <float.h>

#include "hash.h"
#include "console.h"
#include "comfun.h"
#include "cmpack_common.h"
#include "cmpack_phtfile.h"
#include "phthead.h"
#include "phtaper.h"
#include "phtstars.h"
#include "phtparser.h"
#include "phtloadsrt.h"
#include "phtfile.h"

/******************* Constants and definitions **********************/

#define BUFFSIZE 	8192

/********************** Local functions ****************************/

static void phfile_clean(void *ctx);

/* Create new photometry file context */
static CmpackPhtFile *phfile_create(void)
{
	CmpackPhtFile *f = (CmpackPhtFile*)gc_create(sizeof(CmpackPhtFile));
	f->gc.clean = &phfile_clean;
	pht_head_init(&f->hd);
	pht_aptab_init(&f->ap);
	pht_stars_init(&f->st);
	return f;
}

/* Frees all allocated memory in internal structures */
static void phfile_clean(void *ctx)
{
	CmpackPhtFile *f = (CmpackPhtFile*)ctx;
	pht_head_clear(&f->hd);
	pht_aptab_clear(&f->ap);
	pht_stars_clear(&f->st);
	gc_clean(f);
}

/* Writes content of the file to the stream in XML format */
static void phfile_write(CmpackPhtFile *f, FILE *to)
{
	/* Prolog */
	fprintf(to, "<?xml version=\"1.0\" encoding=\"ISO_8859-1\" ?>\n");
	fprintf(to, "<phot version=\"1\">\n");
	/* Write headers */
	if (f->hd.count>0) {
		fprintf(to, "<head>\n");
		pht_head_write(&f->hd, to);
		fprintf(to, "</head>\n");
	}
	/* Write apertures */
	if (f->ap.count>0) {
		fprintf(to,  "<apertures>\n");
		pht_aptab_write(&f->ap, to);
  		fprintf(to, "</apertures>\n");
	}
	/* Write stars */
	if (f->st.count>0) {
		fprintf(to,  "<body>\n");
		pht_stars_write(&f->st, to);
		fprintf(to,  "</body>\n");
	}
	/* Epilog */
	fprintf(to,  "</phot>\n");
}

/*****************   Public functions    *******************/

/* Creates a empty temporary file. This file is maintained only 
 * in the memory. Returns pointer to internal file structure or
 * NULL if failed. 
 */
CmpackPhtFile *cmpack_pht_init(void)
{
	return phfile_create();
}
		
/* Opens file. You can open the file for reading only or in 
 * read-write mode. Returns pointer to internal file structure 
 * handle or NULL if failed.
 */
int cmpack_pht_load(CmpackPhtFile **pfile, const char *filename, CmpackLoadMode flags)
{
	int res;
	FILE *f;
	char line[MAXLINE+1];
	CmpackPhtFile *file;

	*pfile = NULL;
	f = fopen(filename, "rb");
	if (!f) 
		return CMPACK_ERR_OPEN_ERROR;
	/* Load a first line */
	fgets(line, MAXLINE, f); 
	fseek(f, 0, SEEK_SET);
	if (strstr(line, "<?xml")) {
		/* XML photometry file */
		file = phfile_create();
		res = pht_loadxml(file, f, flags);
	} else
	if (strstr(line, "NL") && strstr(line, "NX") && strstr(line, "NY")) {
		/* Text photometry file */
		file = phfile_create();
		res = pht_loadsrt(file, f, flags);
	} else {
		/* Unknown format */
		file = NULL;
		res = CMPACK_ERR_UNKNOWN_FORMAT;
	}
	fclose(f);
	if (res!=0) {
		if (file)
			cmpack_unref(file);
	} else {
		*pfile = file;
	}
	return res;
}


/* Save data to a file */
int cmpack_pht_save(CmpackPhtFile *file, const char *filename)
{
	FILE *f;

	f = fopen(filename, "wb+");
	if (!f) 
		return CMPACK_ERR_OPEN_ERROR;

	phfile_write(file, f);
	fclose(f);
	return CMPACK_ERR_OK;
}

/* Save changes, close file and free allocated memory */
void cmpack_pht_clear(CmpackPhtFile *f)
{
	pht_head_clear(&f->hd);
	pht_aptab_clear(&f->ap);
	pht_stars_clear(&f->st);
}

/* Makes copy of the file */
void cmpack_pht_copy(CmpackPhtFile *fd, CmpackPhtFile *fs)
{
	pht_head_copy(&fd->hd, &fs->hd);
	pht_aptab_copy(&fd->ap, &fs->ap);
	pht_stars_copy(&fd->st, &fs->st);
}

/* Test if the file is a photometry file */
int cmpack_pht_test(const char *filename)
{
	int bytes, filesize;
	char buffer[2048];

	FILE *f = fopen(filename, "rb");
	if (f) {
		fseek(f, 0, SEEK_END);
		filesize = ftell(f);
		fseek(f, 0, SEEK_SET);
		bytes = (int)fread(buffer, 1, 2048, f);
		fclose(f);
		return cmpack_pht_test_buffer(buffer, bytes, filesize);
	}
	return 0;
}

/* Test if the file is a photometry file */
int cmpack_pht_test_buffer(const char *buffer, int buflen, int filesize)
{
	int res;
	char *aux;

	res = 0;
	if (buffer && buflen>16) {
		/* Convert buffer to a null-terminated string */
		aux = (char*)cmpack_malloc((buflen+1)*sizeof(char));
		memcpy(aux, buffer, buflen);
		aux[buflen] = '\0';
		if (strstr(aux, "<?xml") && strstr(aux, "<phot")) {
			/* XML photometry file */
			res = 1;
		} else
		if (strstr(aux, "NL") && strstr(aux, "NX") && strstr(aux, "NY")) {
			/* Text photometry file */
			res = 1;
		}
		cmpack_free(aux);
	}
	return res;
}

/********************   Header manipulation    ***********************/

/* Sets value and comment of specified key */
void cmpack_pht_pkys(CmpackPhtFile *f, const char *key, const char *val, const char *com)
{
	pht_head_set(&f->hd, key, val, com);
}

/* Sets value and comment of specified key */
void cmpack_pht_pkyi(CmpackPhtFile *f, const char *key, int val, const char *com)
{
	char buf[16];

	sprintf(buf, "%d", val);
	pht_head_set(&f->hd, key, buf, com);
}

/* Sets value and comment of specified key */
void cmpack_pht_pkyd(CmpackPhtFile *f, const char *key, double val, int prec, const char *com)
{
	char buf[256];

	if (prec>16) 
		prec = 16;
	if (fabs(val)<DBL_EPSILON)
		strcpy(buf, "0");
	else if (fabs(val)<1e+9)
	  sprintf(buf, "%.*f", prec, val);
	else
	  sprintf(buf, "%.*e", prec, val);
	pht_head_set(&f->hd, key, buf, com);
}

/* Gets value of specified key */
const char *cmpack_pht_gkys(CmpackPhtFile *f, const char *key)
{
	CmpackHeadItem *item = pht_head_get(&f->hd, key);
	if (item)
		return item->val;
	else
		return NULL;
}

/* Gets comment of specified key */
const char *cmpack_pht_gkyc(CmpackPhtFile *f, const char *key)
{
	CmpackHeadItem *item = pht_head_get(&f->hd, key);
	if (item)
		return item->com;
	else
		return NULL;
}

/* Gets value and comment of specified key */
int cmpack_pht_gkyi(CmpackPhtFile *f, const char *key, int defval)
{
	CmpackHeadItem *item = pht_head_get(&f->hd, key);
	if (item)
		return atoi(item->val);
	else
		return defval;
}

/* Gets value and comment of specified key */
double cmpack_pht_gkyd(CmpackPhtFile *f, const char *key, double defval)
{
	CmpackHeadItem *item = pht_head_get(&f->hd, key);
	if (item)
		return atof(item->val);
	else
		return defval;
}

/* Return number of parameters stored in the active section */
int cmpack_pht_nkey(CmpackPhtFile *f)
{
	return f->hd.count;
}

/* Get key, value and comment triplet by its index. The first record has index 0.
 * Returns nonzero on success or zero on failure.
 */
int cmpack_pht_gkyn(CmpackPhtFile *f, int index, const char **key, const char **val, const char **com)
{
	if (index>=0 && index<f->hd.count) {
		CmpackHeadItem *item = f->hd.list[index];
		if (key) *key = item->key;
		if (val) *val = item->val;
		if (com) *com = item->com;
		return 1;
	} else {
		if (key) *key = NULL;
		if (val) *val = NULL;
		if (com) *com = NULL;
		return 0;
	}
}

/* Deletes specified key */
void cmpack_pht_dkey(CmpackPhtFile *f, const char *key)
{
	pht_head_delete(&f->hd, key);
}

int cmpack_pht_get_size(CmpackPhtFile *file, int *width, int *height)
{
	if (cmpack_pht_gkys(file, "width") && cmpack_pht_gkys(file, "height")) {
		if (width) *width = cmpack_pht_gkyi(file, "width", 0);
		if (height) *height = cmpack_pht_gkyi(file, "height", 0);
		return 1;
	} else if (cmpack_pht_gkys(file, "NX") && cmpack_pht_gkys(file, "NY")) {
		if (width) *width = cmpack_pht_gkyi(file, "NX", 0);
		if (height) *height = cmpack_pht_gkyi(file, "NY", 0);
		return 1;
	} else {
		if (width) *width = 0;
		if (height) *height = 0;
		return 0;
	}
}

int cmpack_pht_get_offset(CmpackPhtFile *file, double *offsetx, double *offsety)
{
	if (cmpack_pht_gkys(file, "offsetx") && cmpack_pht_gkys(file, "offsety")) {
		if (offsetx) *offsetx = cmpack_pht_gkyi(file, "offsetx", 0);
		if (offsety) *offsety = cmpack_pht_gkyi(file, "offsety", 0);
		return 1;
	} else if (cmpack_pht_gkys(file, "OFFSETX") && cmpack_pht_gkys(file, "OFFSETY")) {
		if (offsetx) *offsetx = cmpack_pht_gkyi(file, "OFFSETX", 0);
		if (offsety) *offsety = cmpack_pht_gkyi(file, "OFFSETY", 0);
		return 1;
	} else {
		if (offsetx) *offsetx = 0;
		if (offsety) *offsety = 0;
		return 0;
	}
}

double cmpack_pht_get_jd(CmpackPhtFile *file)
{
	if (cmpack_pht_gkys(file, "jd"))
		return cmpack_pht_gkyd(file, "jd", 0);
	else if (cmpack_pht_gkys(file, "JD")==0)
		return cmpack_pht_gkyd(file, "JD", 0);
	else
		return 0;
}

double cmpack_pht_get_exptime(CmpackPhtFile *file)
{
	if (cmpack_pht_gkys(file, "exptime"))
		return cmpack_pht_gkyd(file, "exptime", 0);
	else if (cmpack_pht_gkys(file, "EXPTIME"))
		return cmpack_pht_gkyd(file, "EXPTIME", 0);
	else
		return 0;
}

const char *cmpack_pht_get_filter(CmpackPhtFile *file)
{
	if (cmpack_pht_gkys(file, "filter"))
		return cmpack_pht_gkys(file, "filter");
	else if (cmpack_pht_gkys(file, "FILTER"))
		return cmpack_pht_gkys(file, "FILTER");
	else
		return NULL;
}

/********************   Table of apertures   ***********************/

/* Returns number of apertures */
int cmpack_pht_naper(CmpackPhtFile *f)
{
	return f->ap.count;
}

/* Find aperture by identifier */
CmpackPhtAperture *cmpack_pht_find_aperture(CmpackPhtFile *f, int apid)
{
	int i;

	if (apid>=0) {
		for (i=0; i<f->ap.count; i++) {
			if (f->ap.list[i]->id == apid)
				return cmpack_addref(f->ap.list[i]);
		}
	}
	return NULL;
}

/* Return aperture by index */
CmpackPhtAperture *cmpack_pht_get_aperture(CmpackPhtFile *f, int index)
{
	if (index>=0 && index<f->ap.count)
		return cmpack_addref(f->ap.list[index]);
	else
		return NULL;
}

/* Add aperture to a file */
void cmpack_pht_add_aperture(CmpackPhtFile *f, CmpackPhtAperture *a)
{
	int i;

	for (i=0; i<f->ap.count; i++) {
		if (f->ap.list[i]->id == a->id) {
			/* Replace existing */
			cmpack_unref(f->ap.list[i]);
			f->ap.list[i] = a;
			return;
		}
	}
	/* Append new aperture */
	pht_aptab_add(&f->ap, a);
}

/********************   Table of stars   ***********************/

/* Returns number of stars */
int cmpack_pht_nstar(CmpackPhtFile *f)
{
	return f->st.count;
}

/* Return star by index */
CmpackPhtStar *cmpack_pht_get_star(CmpackPhtFile *f, int index)
{
	if (index>=0 && index<f->st.count)
		return cmpack_addref(f->st.list[index]);
	else
		return NULL;
}

/* Find star by identifier */
CmpackPhtStar *cmpack_pht_find_star(CmpackPhtFile *f, int stid)
{
	if (stid>=0) {
		int index = pht_stars_search(&f->st, stid);
		if (index>=0)
			return cmpack_addref(f->st.list[index]);
	}
	return NULL;
}

/* Find star by reference indentifier */
CmpackPhtStar *cmpack_pht_find_star_by_refid(CmpackPhtFile *f, int refid)
{
	if (refid>=0) {
		int index = pht_stars_search_by_refid(&f->st, refid);
		if (index>=0)
			return cmpack_addref(f->st.list[index]);
	}
	return NULL;
}

/* Find aperture / create new aperture */
void cmpack_pht_add_star(CmpackPhtFile *f, CmpackPhtStar *s)
{
	int i;

	for (i=0; i<f->st.count; i++) {
		if (f->st.list[i]->id == s->id) {
			/* Replace existing */
			cmpack_unref(f->st.list[i]);
			f->st.list[i] = s;
			return;
		}
	}
	/* Append new star */
	pht_stars_add(&f->st, s);
}
