/**************************************************************

phtloadstr.c (C-Munipack project)
Load SRT photometry file
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "comfun.h"
#include "cmpack_common.h"
#include "phtloadsrt.h"
#include "phtfile.h"

/*
 NL   NX   NY  LOWBAD HIGHBAD  THRESH     AP1  PH/ADU  RNOISE              JD           FILTER    EXPTIME    FWHM  OFFSETX  OFFSETY
  2  382  255     0.0 65535.0  105.92    2.00    2.30   15.00  2452909.317326            Clear      20.00    2.10    -0.00    -0.00

     0  122.034   22.033  11.1179  10.9562  10.8833  10.8484  10.8278  10.8127  10.8021  10.7941  10.7882  10.7814  99.9990  99.9990
      449.033 15.08  0.00  0.0012   0.0011   0.0011   0.0011   0.0011   0.0012   0.0013   0.0015   0.0018   0.0021   9.9990   9.9990
*/

int pht_loadsrt(CmpackPhtFile *pht, FILE *f, CmpackLoadMode flags)
{
	int		i, j, id, res, start, end, length, matched;
	char	line1[MAXLINE], line2[MAXLINE], key[MAXLINE], value[MAXLINE];
	char	*endptr, *start1, *start2;
	CmpackPhtMeasurement mag;
	
	/* Check file format */
	fgets(line1, MAXLINE, f);
	if (!strstr(line1, " NL") || !strstr(line1, "JD"))
		return CMPACK_ERR_UNKNOWN_FORMAT;

	/* Read header information */
	fgets(line2, MAXLINE, f);
	start = 0;
	length = (int)strlen(line1);
	while (start<length) {
		end = start + (int)strspn(line1+start, " \t\r\n");
		end += (int)strcspn(line1+end, " \t\r\n");
		line1[end] = line2[end] = '\0';
		key[0] = value[0] = '\0';
		sscanf(line1+start, " %[^ \t\r\n]", key);
		sscanf(line2+start, " %[^ \t\r\n]", value);
		if (strlen(key)>0)
			pht_head_set(&pht->hd, key, value, NULL);
		start = end+1;
	}
	matched = pht_head_get(&pht->hd, "OFFSETX")!=NULL;

	/* Single blank line */
	fgets(line1, 200, f);
	j = 3;

	/* Make table of apertures */
	if ((flags & CMPACK_LOAD_HEADONLY)==0) {
		for (i=0; i<12; i++) {
			pht_aptab_add(&pht->ap, pht_aper_new(i+1));
		}
	}

	/* Read magnitudes and errors of stars */
	if ((flags & CMPACK_LOAD_HEADONLY)==0) {
		while (1) {
			if (!fgets(line1, 200, f))
				break;
			if (!fgets(line2, 200, f))
				break;
			j+=2;
			id = strtol(line1, &endptr, 10);
			if (id>=0 && endptr!=line1) {
				/* Star */
				CmpackPhtStar *star = pht_star_new(id);
				if (matched) 
					star->refid = (j-5)/3;
				start1 = endptr;
				star->center[0] = strtod(start1, &endptr);
				if (start1!=endptr) {
					start1 = endptr;
					star->center[1] = strtod(start1, &endptr);
					if (start1!=endptr) {
						star->mask |= CMPACK_MASK_POS;
						start1 = endptr;
					}
				}
				start2 = line2;
				for (i=0; i<3; i++) {
					start2 += strspn(start2, " \t\r\n");
					start2 += strcspn(start2, " \t\r\n");
				}
				pht_stars_add(&pht->st, star);
				/* Read magnitudes */
				for (i=0; i<12; i++) {
					memset(&mag, 0, sizeof(CmpackPhtMeasurement));
					res = sscanf(start1, " %lf", &mag.magnitude);
					if (res==1 && mag.magnitude<99 && mag.magnitude>-99)
						mag.mask |= CMPACK_MASK_MAG;
					res = sscanf(start2, " %lf", &mag.error);
					if (res==1 && mag.error>0 && mag.error<9)
						mag.mask |= CMPACK_MASK_ERR; 
					if (mag.mask) 
						pht_star_set_data(star, i+1, &mag);
					start1 += strspn(start1, " \t\r\n");
					start1 += strcspn(start1, " \t\r\n");
					start2 += strspn(start2, " \t\r\n");
					start2 += strcspn(start2, " \t\r\n");
				}
			}
			/* Skip empty line */
			fgets(line1, 200, f);
			j++;
		}
	}
	return CMPACK_ERR_OK;
}
