/**************************************************************

tabparser.c (C-Munipack project)
Copyright (C) 2004 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cmpack_common.h"
#include "tabparser.h"
#include "tabdata.h"
#include "tabhead.h"
#include "table.h"

#define BUFFSIZE 	8192
#define ALLOC_BY	64

/***********************      Datatypes      ******************/

/* String buffer */
struct _CmpackTabString
{
	size_t count;				/**< Number of stored chars */
	size_t capacity;			/**< Number of allocated chars */
	char *buf;					/**< Memory buffer */
};
typedef struct _CmpackTabString CmpackTabString;

/* Context structure for configuration file parser */
struct _CmpackTabParser
{
	int headonly;				/**< Parse only header */
	int st;						/**< Automaton state */
	char lastch;				/**< Previous character */
	CmpackTabString key;		/**< String buffers for keyword */
	CmpackTabString val;		/**< String buffers for value */
	CmpackTable *tab;			/**< Reference to table */
	int row_index, col_index;	/**< Current position in table - row, column indices (starting by 1) */
};
typedef struct _CmpackTabParser CmpackTabParser;

/*******************   STRING BUFFER FUNCTIONS    ***************************/

/* Initialize a string buffer */
static void str_init(CmpackTabString *str)
{
	memset(str, 0, sizeof(CmpackTabString));
}

/* Free allocated memory in a string buffer */
static void str_free(CmpackTabString *str)
{
	cmpack_free(str->buf);
	str->buf = NULL;
	str->capacity = str->count = 0;
}

/* Clear a string (do not free the memory) */
static void str_clear(CmpackTabString *str)
{
	str->count = 0;
}

/* Add a character to a string buffer */
static void str_add(CmpackTabString *str, char ch)
{
	if (str->count >= str->capacity) {
		str->capacity += ALLOC_BY;
		str->buf = (char*) cmpack_realloc(str->buf, str->capacity*sizeof(char));
	}
	str->buf[str->count++] = ch;
}

/* Returns length of the string in characters (without trailing null) */
static size_t str_count(CmpackTabString *str)
{
	return str->count;
}

/* Convert a string buffer to null-terminated string (use free to destroy it) */
static const char *str_cstr(CmpackTabString *str)
{
	if (str->count >= str->capacity) {
		str->capacity += ALLOC_BY;
		str->buf = (char*) cmpack_realloc(str->buf, str->capacity*sizeof(char));
	}
	str->buf[str->count] = '\0';
	return str->buf;
}

/* Convert a string buffer to null-terminated string (use free to destroy it) */
static int str_int(CmpackTabString *str)
{
	int retval = 0;

	char *aux = (char*) cmpack_malloc(str->count+1);
	if (aux) {
		memcpy(aux, str->buf, str->count*sizeof(char));
		aux[str->count] = '\0';
		retval = atoi(aux);
		cmpack_free(aux);
	}
	return retval;
}

/* Convert a string buffer to null-terminated string (use free to destroy it) */
static double str_dbl(CmpackTabString *str)
{
	double retval = 0;

	char *aux = (char*) cmpack_malloc(str->count+1);
	if (aux) {
		memcpy(aux, str->buf, str->count*sizeof(char));
		aux[str->count] = '\0';
		retval = atof(aux);
		cmpack_free(aux);
	}
	return retval;
}

/* Trim all trailing white characters from the string */
static void str_rtrim(CmpackTabString *str)
{
	size_t i;

	for (i=str->count-1; i>=0; i--) {
		if (str->buf[i]<=32)
			str->count--;
		else
			break;
	}
}

/* End of line(file) found, process parsed data for current line. */
static void process_line(CmpackTabParser *p, int columns)
{
	const char *name0, *name1;

	if (columns) {
		/* Determine the type of the table */
		name0 = tab_cols_name(p->tab->data, 0);
		if (name0) {
			if (strcmp(name0, "JD")==0 || strcmp(name0, "HJD")==0 || 
				strcmp(name0, "GJD")==0 || strcmp(name0, "JDHEL")==0 || 
				strcmp(name0, "JDGEO")==0) {
					name1 = tab_cols_name(p->tab->data, 1);
					if (name1) {
						if (strcmp(name1, "V-C")==0 || strcmp(name1, "s1")==0) {
							p->tab->type = CMPACK_TABLE_LIGHTCURVE;
						} else
						if (strcmp(name1, "OFFSETX")==0) {
							p->tab->type = CMPACK_TABLE_TRACKLIST;
						} else
						if (strcmp(name1, "AIRMASS")==0) {
							p->tab->type = CMPACK_TABLE_AIRMASS;
						}
					}
			} else
			if (strcmp(name0, "INDEX")==0) {
				p->tab->type = CMPACK_TABLE_MUNIFIND;
			} else
			if (strcmp(name0, "APERTURE")==0) {
				p->tab->type = CMPACK_TABLE_MUNIFIND;
			}
		}
	}
	str_clear(&p->key);
	str_clear(&p->val);
}	

/* Append column */
static void process_col(CmpackTabParser *p, CmpackTabString *name)
{
	CmpackColumnType ctype;
	
	const char *sname = str_cstr(name);
	if (sname) {
		/*if (strstr(sname, "JD")) 
			ctype = CMPACK_COLUMN_JULDAT;
		else if (strcmp(sname, "HELCOR")==0) 
			ctype = CMPACK_COLUMN_HELCOR;
		else if (strcmp(sname, "AIRMASS")==0) 
			ctype = CMPACK_COLUMN_AIRMASS;
		else if (strcmp(sname, "APERTURE")==0 || strcmp(sname, "INDEX")==0) 
			ctype = CMPACK_COLUMN_ID;
		else if (strcmp(sname, "OFFSETX")==0 || strcmp(sname, "OFFSETX")==0)
			ctype = CMPACK_COLUMN_PIXEL;
		else if (strcmp(sname, "GOODPOINTS")) 
			ctype = CMPACK_COLUMN_UNSPECIFIED;
		else if (sname[0]=='V' || sname[0]=='C' || sname[0]=='K') 
			ctype = CMPACK_COLUMN_MAG;
		else if (sname[0]=='s') 
			ctype = CMPACK_COLUMN_ERR;
		else
			ctype = CMPACK_COLUMN_UNSPECIFIED;*/

		if (strstr(sname, "JD") || strcmp(sname, "HELCOR")==0 || 
			strcmp(sname, "AIRMASS")==0 || strcmp(sname, "ALTITUDE")==0 || 
			strstr(sname, "OFFSET") || sname[0]=='V' || sname[0]=='C' || 
			sname[0]=='K' || sname[0]=='s')
			ctype = CMPACK_COLUMN_DOUBLE;
		else if (strcmp(sname, "APERTURE")==0 || strcmp(sname, "INDEX")==0 ||
			strcmp(sname, "GOODPOINTS")==0)
			ctype = CMPACK_COLUMN_INTEGER;
		else 
			ctype = CMPACK_COLUMN_STRING;
		tab_data_add_col(p->tab->data, sname, ctype, 0);
	}
}

/* Append parameter */
static void process_val(CmpackTabParser *p, CmpackTabString *key, CmpackTabString *val)
{
	if (str_count(key)>0)
		tab_head_set(p->tab->head, str_cstr(key), str_cstr(val));
}

/* Append new row */
static void process_row(CmpackTabParser *p)
{
	p->row_index = tab_data_add_row(p->tab->data);
	p->col_index = 0;
}

/* Set table data */
static void process_data(CmpackTabParser *p, CmpackTabString *str)
{
	CmpackTabColumn *column = tab_data_get_col(p->tab->data, p->col_index);
	if (column) {
		switch (tab_col_type(column))
		{
		case CMPACK_COLUMN_STRING:
			tab_data_set(p->tab->data, p->col_index, p->row_index, str_cstr(str));
			break;
		case CMPACK_COLUMN_INTEGER:
			tab_data_seti(p->tab->data, p->col_index, p->row_index, str_int(str));
			break;
		case CMPACK_COLUMN_DOUBLE:
			tab_data_setd(p->tab->data, p->col_index, p->row_index, str_dbl(str));
			break;
		}
	}
	p->col_index++;
}

/********************   PUBLIC FUNCTIONS   ***************************/

/* Init parser context */
void tab_parser_init(CmpackTabParser *p, CmpackTable *tab, CmpackLoadMode flags)
{
	memset(p, 0, sizeof(CmpackTabParser));
	p->tab = tab;
	p->headonly = (flags & CMPACK_LOAD_HEADONLY)!=0;
	str_init(&p->key);
	str_init(&p->val);
}

/* Free asociated memory in the context */
void tab_parser_clear(CmpackTabParser *p)
{
	str_free(&p->key);	
	str_free(&p->val);	
	p->tab = NULL;
}
	
/* Parse buffer */
int tab_parser_parse(CmpackTabParser *p, const char *buf, size_t buflen, int eof)
{
	size_t i;
	char ch;

	if (p==NULL || buf==NULL)
		return 0; 
	
	for (i=0; i<buflen; i++) {
		ch = buf[i]; 
		/* Replace CR, LF or CR+LF by single CR character */
		if (ch==10) {
			if (p->lastch==13)
				continue;
			ch = 13;
		}
		switch (p->st) 
		{
		case 0:
			/* Start of the first line (column names) */
			if ((ch>='A' && ch<='Z') || (ch>='a' && ch<='z') || (ch=='_')) {
				/* Start of a column name */
				str_clear(&p->key);
				str_add(&p->key, ch);
				p->st = 1;
			} else if (ch==13) {
				/* First line is empty => parse error */
				return CMPACK_ERR_UNKNOWN_FORMAT;
			}
		 	break;
		 	
		case 1: 
		 	/* Reading column name */
		 	if (ch==' ') {
				/* End of colum name, wait for the next one */
				process_col(p, &p->key);
				p->st = 2;
		 	} else if (ch==13) {
				/* End of last column name */
				process_col(p, &p->key);
				process_line(p, 1);
				p->st = 10;
			} else {
				/* Next character in column name */
			 	str_add(&p->key, ch);
			}
			break;

		case 2:
			/* Waiting for the next column name */
			if ((ch>='A' && ch<='Z') || (ch>='a' && ch<='z') || (ch=='_')) {
				/* Start of a column name */
				str_clear(&p->key);
				str_add(&p->key, ch);
				p->st = 1;
			} else if (ch==13) {
				/* First line is empty => parse error */
				process_line(p, 1);
				p->st = 10;
			}
			break;
			
		case 10:
			/* Second line with parameters */
		 	if ((ch>='A' && ch<='Z') || (ch>='a' && ch<='z') || (ch=='_')) {
				/* Start of a keyword */
				str_clear(&p->key);
				str_add(&p->key, ch);
				p->st = 11;
			} else if (ch==13) {
				/* Empty second line */
				process_line(p, 2);
				p->st = 40;
			}
		 	break;
		 	
		case 11:
		 	/* Reading keyword */
			if (ch==':') {
				/* End of keyword wait for value */
				str_rtrim(&p->key);
				p->st = 20;
			} else if (ch==13) {
				/* End of line */
				process_line(p, 2);
				p->st = 40;
			} else {
				/* Next character in keyword */
				str_add(&p->key, ch);
			}
			break;

		case 20:
			/* Reading value*/
			if (ch==13) {
				/* End of line */
				process_line(p, 2);
				p->st = 40;
			} else
		 	if (ch!=' ' && ch!='\t') {
				/* Start of value (unquoted) */
				str_clear(&p->val);
				str_add(&p->val, ch); 
				p->st = 21;
			}
		 	break;
		 	
		case 21: 
			/* Reading value */
			if (ch==',') {
				/* End of value */
				str_rtrim(&p->val);
				process_val(p, &p->key, &p->val);
				p->st = 10;
			} else if (ch==13) {
				/* End of line */
				str_rtrim(&p->val);
				process_val(p, &p->key, &p->val);
				process_line(p, 2);
				p->st = 40;
			} else {
				str_add(&p->val, ch);
			}
		 	break;

		case 40:
			/* Waiting for table data */
			/* First character on the line */
		 	if ((ch>='0' && ch<='9') || (ch=='+') || (ch=='-')) {
				/* Start of a value */
				str_clear(&p->val);
				str_add(&p->val, ch);
				p->st = 42;
			} else
			if (ch==13) {
				/* Empty line ends the header */
				process_line(p, 3);
			} else
			if (ch==10) {
				if (p->lastch!=13) {
					process_line(p, 3);
				}
			} else {
				/* Write space */
				p->st = 41;
			}
		 	break;

		case 41:
			/* Waiting for table data */
		 	if ((ch>='0' && ch<='9') || (ch=='+') || (ch=='-')) {
				/* Start of a keyword */
				str_clear(&p->val);
				str_add(&p->val, ch);
				p->st = 42;
			} else
			if (ch==13 || ch==10) {
				/* Line with spaces only */
				process_line(p, 3);
				p->st = 40;
			}
		 	break;

		case 42:
			/* Reading value */
			if (ch==' ') {
				/* End of value */
				process_row(p);
				process_data(p, &p->val);
				p->st = 43;
		 	} else 
			if (ch==13 || ch==10) {
				/* End of value */
				process_row(p);
				process_data(p, &p->val);
				process_line(p, 3);
				p->st = 40;
			} else {
				/* Next character in keyword */
				str_add(&p->val, ch);
			}
			break;

		case 43:
			/* Waiting for next keyword */
		 	if ((ch>='0' && ch<='9') || (ch=='+') || (ch=='-')) {
				/* Start of a keyword */
				str_clear(&p->val);
				str_add(&p->val, ch);
				p->st = 44;
			} else
			if (ch==13 || ch==10) {
				/* Line with spaces only */
				process_line(p, 3);
				p->st = 40;
			}
			break;

		case 44:
			/* Reading next values */
			if (ch==' ') {
				/* End of value */
				process_data(p, &p->val);
				p->st = 43;
		 	} else 
			if (ch==13 || ch==10) {
				/* End of value */
				process_data(p, &p->val);
				process_line(p, 3);
				p->st = 40;
			} else {
				/* Next character in keyword */
				str_add(&p->val, ch);
			}
			break;
		}
		/* Remember last character */
		p->lastch = ch;
	}

	if (eof && p->st!=0) {
		/* Last line, which doesn't end with line delimiter */
		p->st = 0;
	}
	return 0;
}

/* Load table */
int tab_load(CmpackTable *tab, FILE *from, CmpackLoadMode flags)
{
	size_t len;
	int res = 0, done;
	char buff[BUFFSIZE];
	CmpackTabParser p;

	tab_parser_init(&p, tab, flags);
	do {
		len = fread(buff, 1, BUFFSIZE, from);
		if (ferror(from)) {
			res = CMPACK_ERR_READ_ERROR;
			break;
		}
		done = feof(from);
		res = tab_parser_parse(&p, buff, len, done);
		if (res!=0) 
			break;
	} while (!done);
	tab_parser_clear(&p);
	return res;
}

/* Detect table type */
CmpackTableType tab_format(const char *header)
{
	CmpackTabParser p;
	CmpackTableType retval = CMPACK_TABLE_UNSPECIFIED;
	CmpackTable *tab = cmpack_tab_init(CMPACK_TABLE_UNSPECIFIED);
	tab_parser_init(&p, tab, CMPACK_LOAD_HEADONLY);
	if (tab_parser_parse(&p, header, strlen(header), 1)==0) 
		retval = cmpack_tab_get_type(tab);
	tab_parser_clear(&p);
	cmpack_unref(tab);
	return retval;
}
