/**************************************************************

graph.h (C-Munipack project)
Widget which can draw a graph
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#ifndef MUNIWIN_GRAPH_H
#define MUNIWIN_GRAPH_H

#include <gtk/gtk.h>

#include "callback.h"
#include "utils.h"

//
// Base class for all graph widgets
//
class CGraph: public CCBObject
{
public:
	// Axis flags:
	enum tXYAxisFlags {
		LOG_SCALE		= (1<<0),	// Logarithmic scale
		REVERSE_SCALE	= (1<<1),	// Reverse axis
		TIME_LABELS		= (1<<2),	// Show labels in time format
		EXP_LABELS		= (1<<3),	// Show labels in exponential format
		NO_LABELS		= (1<<4),	// Hide labels
		LABELS_RIGHT	= (1<<5),	// Show labels on the right side of the graph
		LABELS_TOP		= LABELS_RIGHT,	// Show labels on the top side of the graph
		LABELS_INSIDE	= (1<<6),	// Show labels inside a panel
		NO_GRID			= (1<<7)	// Hide grid
	};

	// Column flags:
	enum tErrFlags {
		HIDDEN			= (1<<0),	// Hidden column
		NO_ERRORS		= (1<<1)	// Hide error bars
	};

	// Callback message codes
	enum tMessageId {
		CB_LEFT_CLICK,			// Left button click
		CB_LEFT_DBLCLICK,		// Left button double click
		CB_RIGHT_CLICK,			// Right button click
		CB_MOUSE_MOVE			// Mouse moved
	};

public:
	// Constructor
	CGraph();
	
	// Destructor
	virtual ~CGraph();

	// Set data 
	void SetData(class CGraphData *pData);

	// Set graph padding
	void SetGraphPadding(int left, int top, int right, int bottom);

	// After this call, the backscreen buffer will be updated before next paint.
	// And queues the redraw.
	void Invalidate(void);

	// X axis
	void SetXParams(const char *name, double Min, double Max, double Eps, 
		int flags, int Prec);
	void SetXFlags(int mask, int value);
	int GetXFlags(void) const;
	void SetXName(const char *name);
	const char *GetXName(void) const
	{ return m_X.Name; }

	// Y axis
	void SetYParams(const char *name, double Min, double Max, double Eps, 
		int flags, int Prec);
	void SetYFlags(int mask, int value);
	int GetYFlags(void) const;
	void SetYName(const char *name);
	const char *GetYName(void) const
	{ return m_X.Name; }

	// Data columns
	void ClearColumns();
	int AddColumn(int val_channel, int err_channel, int flags, tColor color = 0);
	void SetColumn(int col, int val_channel, int err_channel, int flags, tColor color = 0);
	void SetColumnFlags(int col, int mask, int value);
	void ShowColumn(int col, bool show);
	int GetColumnFlags(int col) const;
	int Columns(void) const
	{ return m_ColCount; }

	// Get handle to the drawing area
	GtkWidget *Handle(void)
	{ return m_DrawArea; }

	// Enable/disable mouse tracking
	void EnableTracking(bool enable);

	// Is mouse tracking enabled?
	bool IsTrackingEnabled(void) const
	{ return m_EnableTrack; }

	// Set/get zoom position 
	void SetXZoom(double zoom);
	void SetYZoom(double zoom);
	double GetXZoom(void) const
	{ return m_X.ZoomPos; }
	double GetYZoom(void) const
	{ return m_Y.ZoomPos; }

	// Autozoom
	void AutoZoom(bool x_axis, bool y_axis);

	// Set/get viewfield in projection units
	void SetXResolution(double res);
	void SetYResolution(double res);
	double GetXResolution(void) const;
	double GetYResolution(void) const;

	// Set/get actual position in physical units
	void SetXPosition(double pos);
	void SetYPosition(double pos);
	double GetXPosition(void) const
	{ return ProjToX(m_X.Pos); }
	double GetYPosition(void) const
	{ return ProjToY(m_Y.Pos); }

	// Mouse position in physical units
	bool GetMousePos(double *x, double *y) const;

	// Find a sample near to the specified position
	int FindByPos(double x, double y, int column) const;

	// Set focused item
	void SetFocusedItem(int index, int column);

	// Get focused item
	int GetFocusedItem(void) const
	{ return m_FocusedItem; }

	// Tags
	void SetTag(int column, int row, const char *text, tColor color);
	void DeleteTag(int column, int row);
	void DeleteAllTags(void);

protected:
	enum tFormat {
		F_DEC,					// Decimal format with fixed precision
		F_EXP,					// Exponential format
		F_TIME					// Time format
	};

	struct tAxis {
		char Name[256];			// Axis name
		bool Log;				// Log mapping
		bool Reverse;			// Reverse mapping
		double Min, Max;		// Limits in physical units
		double PMin, PMax;		// Limits in projection units
		double Eps;				// Minimum viewfield in projection units
		bool ShowLabels;		// Show labels
		bool ShowGrid;			// Show grid lines
		bool LabelsInside;		// Labels are inside a panel
		bool LabelsOpposite;	// Labels are on the top/right side
		tFormat Format;			// Format of labels
		int Prec;				// Number of decimal points
		double ZoomPos;			// Actual zoom position
		double ZMin, ZMax;		// Zoom position limits
		double PxlSize;			// Current pixel size in projection units
		double Pos;				// Left/top border of viewfield in projection units
		double Scale, Offset;
	};

	struct tColumn {
		int ValChannel;			// Data channel with values
		int	ErrChannel;			// Data channel with errors
		bool Visible;			// Show data
		bool ShowErrors;		// Show errors
		tColor Color;			// Color
	};

	struct tTag {
		int Col, Row;			// Column index, row index
		char Text[64];			// Text
		tColor Color;			// Color
	};

	struct tRect {
		int left, top, right, bottom;
	};

	CGraphData	*m_pData;				// Graph data
	double		m_ZoomBase;				// Zoom base
	tAxis		m_X, m_Y;				// Axis parameters
	int			m_ColCount;				// Number of columns
	int			m_ColCapacity;			// Number of allocated items
	tColumn		*m_Cols;				// Columns
	GtkWidget	*m_DrawArea;			// Drawing area
	GdkPixmap	*m_Pixmap;				// Backscreen buffer
	GSList		*m_Tags;				// List of tags
	int			m_PixWidth;				// Backscreen buffer width
	int			m_PixHeight;			// Backscreen buffer width
	bool		m_NeedUpdate;			// Needs update 
	tRect		m_PanelRect;			// Panel rectangle in pixels
	tRect		m_GraphRect;			// Graph rectangle in pixels
	tRect		m_GraphPadding;			// Graph padding in pixels
	tRect		m_PanelPadding;			// Panel padding in pixels
	tRect		m_XScaleRect;			// X axis scale rectangle
	tRect		m_YScaleRect;			// Y axis scale rectangle
	tRect		m_XNameRect;			// X axis name rectangle
	tRect		m_YNameRect;			// Y axis name rectangle
	int			m_GraphWidth;			// Viewfield width in pixels
	int			m_GraphHeight;			// Viewfield height in pixels
	bool		m_EnableTrack;			// Mouse tracking enabled?
	int			m_FocusedItem;			// Index of focused item
	int			m_FocusedCol;			// Column for focused item
	double		m_OffsetX, m_OffsetY;	// Reference position in pixels
	int			m_MouseX, m_MouseY;		// Mouse position in pixels

	// Mapping functions between physical units and projection units
	double XToProj(double x) const;
	double YToProj(double y) const;
	double ProjToX(double u) const;
	double ProjToY(double v) const;

	// Mapping functions between projection units and display units
	double XProjToView(double x) const;
	double YProjToView(double y) const;
	double ViewToProjX(double u) const;
	double ViewToProjY(double v) const;

	// Mapping functions between physical units and display units
	double XToView(double x) const;
	double YToView(double y) const;
	double ViewToX(double u) const;
	double ViewToY(double v) const;

	// Update pixelsize
	void UpdatePxlSizeX(void);
	void UpdatePxlSizeY(void);

	// Limit position to scrollable area
	void LimitPosX(void);
	void LimitPosY(void);

	// Autozoom
	void AutoXZoom(void);
	void AutoYZoom(void);

	// Recompute rectangles, update pixelsizes
	void UpdateRectangles(void);

	// Text extents
	int GetXScaleHeight(cairo_t *cr);
	int GetXLabelWidth(cairo_t *cr, int prec);
	int GetYScaleWidth(cairo_t *cr);
	int GetYLabelHeight(cairo_t *cr);
	int GetTextHeight(cairo_t *cr, const char *text);
	int GetTextWidth(cairo_t *cr, const char *text);
	void GetTextExtents(cairo_t *cr, const char *text, int *width, int *height);

	// Autoscaling
	void GetGridXStep(cairo_t *cr, double minval, double maxval, double *step, int *minprec);
	void GetGridYStep(cairo_t *cr, double minval, double maxval, double *step, int *minprec);

	// Print captions
	void PrintLabel(char *buf, double value, tFormat fmt, int prec);
	void PrintXLabel(char *buf, double value, int minprec);
	void PrintYLabel(char *buf, double value, int minprec);

	// Update back-screen buffer
	void Redraw(void);

	// Paiting
	void PaintXLabels(cairo_t *cr);
	void PaintXScaleLabel(cairo_t *cr, double val, int prec, int y);
    void PaintYLabels(cairo_t *cr);
	void PaintYScaleLabel(cairo_t *cr, double val, int prec, int x);
    void PaintData(cairo_t *cr);
	void PaintXGrid(cairo_t *cr);
	void PaintXGridLine(cairo_t *cr, double val);
    void PaintYGrid(cairo_t *cr);
	void PaintYGridLine(cairo_t *cr, double val);

	// Signal handlers
	static gint configure_event(GtkWidget *widget, GdkEventConfigure *event, CGraph *pMe);
	static gint expose_event(GtkWidget *widget, GdkEventExpose *event, CGraph *pMe);
	static gint leave_notify_event(GtkWidget *widget, GdkEventCrossing *event, CGraph *pMe);
	static gint button_press_event(GtkWidget *widget, GdkEventButton *event, CGraph *pMe);
	static gint motion_notify_event(GtkWidget *widget, GdkEventMotion *event, CGraph *pMe);

	void OnConfigure(GtkWidget *widget, GdkEventConfigure *event);
	void OnExpose(GtkWidget *widget, GdkEventExpose *event);
	void OnLeaveNotify(GtkWidget *widget, GdkEventCrossing *event);
	void OnButtonPress(GtkWidget *widget, GdkEventButton *event);
	void OnMotionNotify(GtkWidget *widget, GdkEventMotion *event);
	void MouseMove(GdkModifierType state, int x, int y);

	static void SetRect(tRect *rect, int left, int top, int right, int bottom);
};

#endif
