/**************************************************************

selection.cpp (C-Munipack project)
Table of selected stars
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>

#include "helper_classes.h"
#include "utils.h"

//--------------------------   HELPER FUNCTIONS   ----------------------------------

//
// Make a string
//
static char *MakeString(const char *str)
{
	if (str) {
		char *dst = (char*)g_malloc((strlen(str)+1)*sizeof(char));
		strcpy(dst, str);
		return dst;
	}
	return NULL;
}


//
// Convert a string from locale encoding to UTF-8
//
static char *DecodeString(char *dst, const char *str)
{
	char *aux = g_locale_to_utf8(str, -1, NULL, NULL, NULL);
	if (aux) {
		dst = (char*)g_realloc(dst, (strlen(aux)+1)*sizeof(char));
		strcpy(dst, aux);
		g_free(aux);
		return dst;
	} else {
		if (dst)
			g_free(dst);
		return NULL;
	}
}

// 
// Copy string
//
static char *CopyString(char *dst, const char *src)
{
	if (src) {
		dst = (char*)g_realloc(dst, (strlen(src)+1)*sizeof(char));
		strcpy(dst, src);
		return dst;
	} else {
		if (dst)
			g_free(dst);
		return NULL;
	}
}

//
// Free string
//
static void FreeString(char *str)
{
	if (str)
		g_free(str);
}

//----------------------   SELECTION OF STARS   -------------------------

//
// Default constructor
//
CSelection::CSelection():m_Count(0), m_Capacity(0), m_List(NULL)
{
}


//
// Copy constructor
//
CSelection::CSelection(const CSelection &orig)
{
	int i;

	if (orig.m_Count>0) {
		m_Count = m_Capacity = orig.m_Count;
		m_List = (tStarRec*)g_malloc(m_Capacity*sizeof(tStarRec));
		for (i=0; i<m_Count; i++) 
			m_List[i] = orig.m_List[i];
	} else {
		m_Count = m_Capacity = 0;
		m_List = NULL;
	}
}


//
// Destructor
//
CSelection::~CSelection()
{
	if (m_List)
		g_free(m_List);
}


//
// assignment operator
//
const CSelection &CSelection::operator=(const CSelection &orig)
{
	int i;

	if (orig.m_Count>0) {
		m_Count = m_Capacity = orig.m_Count;
		m_List = (tStarRec*)g_realloc(m_List, m_Capacity*sizeof(tStarRec));
		for (i=0; i<m_Count; i++)
			m_List[i] = orig.m_List[i];
	} else {
		m_Count = m_Capacity = 0;
		g_free(m_List);
		m_List = NULL;
	}
	Callback(CB_SELECT_CHANGE, -1, 0);
	return *this;
}


//
// Clear data
//
void CSelection::Clear()
{
	m_Count = 0;
	Callback(CB_SELECT_CHANGE, -1, 0);
}

//
// Select a star
//
void CSelection::Select(int id, tStarType type)
{
	int i, index = Search(id);
	if (index>=0) {
		if (m_List[index].type != type) {
			for (i=index+1; i<m_Count; i++)
				m_List[i-1] = m_List[i];
			m_Count--;
			if (type!=ST_NOT_SELECTED) {
				m_List[m_Count].id = id;
				m_List[m_Count].type = type;
				m_List[m_Count].index = 0;
				m_Count++;
			}
			Reindex();
			Callback(CB_SELECT_CHANGE, id, 0);
		}
	} else {
		if (type != ST_NOT_SELECTED) {
			if (m_Count >= m_Capacity) {
				m_Capacity += 64;
				m_List = (tStarRec*)g_realloc(m_List, m_Capacity*sizeof(tStarRec));
			}
			m_List[m_Count].id = id;
			m_List[m_Count].type = type;
			m_List[m_Count].index = 0;
			m_Count++;
			Reindex();
			Callback(CB_SELECT_CHANGE, id, 0);
		}
	}
}


//
// Unselect a star
//
void CSelection::Unselect(int id)
{
	int i, index = Search(id);
	if (index>=0) {
		for (i=index+1; i<m_Count; i++)
			m_List[i-1] = m_List[i];
		Reindex();
		Callback(CB_SELECT_CHANGE, id, 0);
	}
}


//
// Get selection
//
int CSelection::FindStar(int id) const
{
	return Search(id);
}


//
// Get star
//
bool CSelection::Get(int i, int *id, tStarType *type, int *index) const
{
	if (i>=0 && i<m_Count) {
		if (id)
			*id = m_List[i].id;
		if (type)
			*type = m_List[i].type;
		if (index)
			*index = m_List[i].index;
		return true;
	} else {
		if (id)
			*id = 0;
		if (type)
			*type = ST_NOT_SELECTED;
		if (index)
			*index = 0;
		return false;
	}
}


//
// Get star identifier
//
int CSelection::GetId(int i) const
{
	if (i>=0 && i<m_Count) 
		return m_List[i].id;
	else
		return 0;
}


//
// Get star type
//
tStarType CSelection::GetType(int i) const
{
	if (i>=0 && i<m_Count)
		return m_List[i].type;
	else
		return ST_NOT_SELECTED;
}


//
// Get star index
//
int CSelection::GetIndex(int i) const
{
	if (i>=0 && i<m_Count)
		return m_List[i].index;
	else
		return 0;
}


//
// Search record by its id
//
int CSelection::Search(int id) const
{
	int i;

	for (i=0; i<m_Count; i++) {
		if (m_List[i].id == id)
			return i;
	}
	return -1;
}


//
// Reindex stars
//
void CSelection::Reindex(void)
{
	int i, count[SELECTION_TYPES];

	memset(count, 0, SELECTION_TYPES*sizeof(int));
	for (i=0; i<m_Count; i++) 
		m_List[i].index = ++count[m_List[i].type];
}


//
// Set selection
//
void CSelection::SetStarList(tStarType type, const char *string)
{
	int i, j, id;
	const char *str;
	char *endptr;

	// Delete all stars with specified type
	for (i=0; i<m_Count; i++) {
		if (m_List[i].type == type) {
			for (j=i+1; j<m_Count; j++)
				m_List[j-1] = m_List[j];
			m_Count--;
		}
	}

	// Parse string and add all stars
	if (string) {
		str = string;
		id = strtol(str, &endptr, 10);
		while (endptr > str) {
			int index = Search(id);
			if (index>=0) {
				m_List[index].type = type;
				m_List[index].index = 0;
			} else {
				if (m_Count >= m_Capacity) {
					m_Capacity += 64;
					m_List = (tStarRec*)g_realloc(m_List, m_Capacity*sizeof(tStarRec));
				}
				m_List[m_Count].id = id;
				m_List[m_Count].type = type;
				m_List[m_Count].index = 0;
				m_Count++;
			}
			str = endptr + strcspn(endptr, "-+0123456789");
			id = strtol(str, &endptr, 10);
		}
	}
	Reindex();
	Callback(CB_SELECT_CHANGE, -1, 0);
}


//
// Set selection
//
void CSelection::SetStarList(tStarType type, const int *list, int count)
{
	int i, j;

	// Delete all stars with specified type
	for (i=0; i<m_Count; i++) {
		if (m_List[i].type == type) {
			for (j=i+1; j<m_Count; j++)
				m_List[j-1] = m_List[j];
			m_Count--;
		}
	}

	// Parse string and add all stars
	for (i=0; i<count; i++) {
		int index = Search(list[i]);
		if (index>=0) {
			m_List[index].type = type;
			m_List[index].index = 0;
		} else {
			if (m_Count >= m_Capacity) {
				m_Capacity += 64;
				m_List = (tStarRec*)g_realloc(m_List, m_Capacity*sizeof(tStarRec));
			}
			m_List[m_Count].id = list[i];
			m_List[m_Count].type = type;
			m_List[m_Count].index = 0;
			m_Count++;
		}
	}
	Reindex();
	Callback(CB_SELECT_CHANGE, -1, 0);
}

//
// Get number of stars of specified type
//
int CSelection::CountStars(tStarType type) const
{
	int i, count;

	count = 0;
	for (i=0; i<m_Count; i++) {
		if (m_List[i].type == type) 
			count++;
	}
	return count;
}


//
// Get selection
//
int CSelection::GetStarList(tStarType type, int *list, int maxlen) const
{
	int i, count;

	count = 0;
	for (i=0; i<m_Count; i++) {
		if (m_List[i].type == type) 
			list[count++] = m_List[i].id;
	}
	return count;
}


//
// Get selection
//
char *CSelection::GetStarList(tStarType type) const
{
	int i, count;
	char aux[12];

	count = 0;
	for (i=0; i<m_Count; i++) {
		if (m_List[i].type == type) 
			count++;
	}

	if (count>0) {
		char *buf = (char*)g_malloc((count*12+1)*sizeof(char));
		if (buf) {
			buf[0] = '\0';
			for (i=0; i<m_Count; i++) {
				if (m_List[i].type == type) {
					if (buf[0]!='\0')
						strcat(buf, ",");
					sprintf(aux, "%d", m_List[i].id);
					strcat(buf, aux);
				}
			}
		}
		return buf;
	}
	return NULL;
}


// Find record that matches specified type and index
// Returns the index of the record
int CSelection::FindStar(tStarType type, int index) const
{
	int i;

	for (i=0; i<m_Count; i++) {
		if (m_List[i].type==type && m_List[i].index==index) 
			return i;
	}
	return -1;
}


//
// Load selection from a file
//
void CSelection::Load(CmpackCatFile *file)
{
	CmpackCatSelection *sel;

	Clear();
	if (file) {
		sel = cmpack_cat_get_selection(file, CMPACK_SELECT_VAR);
		if (sel) {
			LoadSelection(sel, ST_VARIABLE);
			cmpack_unref(sel);
		}
		sel = cmpack_cat_get_selection(file, CMPACK_SELECT_COMP);
		if (sel) {
			LoadSelection(sel, ST_COMPARISON);
			cmpack_unref(sel);
		}
		sel = cmpack_cat_get_selection(file, CMPACK_SELECT_CHECK);
		if (sel) {
			LoadSelection(sel, ST_CHECK);
			cmpack_unref(sel);
		}
		Reindex();
	}
}


//
// Load selection (one type)
//
void CSelection::LoadSelection(CmpackCatSelection *sel, tStarType type)
{
	int count = cmpack_cat_selection_nitems(sel);
	if (count>0) {
		int *buf = (int*)g_malloc(count*sizeof(int));
		cmpack_cat_selection_get(sel, buf, count);
		SetStarList(type, buf, count);
		g_free(buf);
	}
}


//
// Save selection to a file
//
void CSelection::Save(CmpackCatFile *file) const
{
	CmpackCatSelection *sel;

	if (file) {
		cmpack_cat_clear_selection(file);
		sel = cmpack_cat_get_selection(file, CMPACK_SELECT_VAR);
		if (sel) {
			SaveSelection(sel, ST_VARIABLE);
			cmpack_unref(sel);
		}
		sel = cmpack_cat_get_selection(file, CMPACK_SELECT_COMP);
		if (sel) {
			SaveSelection(sel, ST_COMPARISON);
			cmpack_unref(sel);
		}
		sel = cmpack_cat_get_selection(file, CMPACK_SELECT_CHECK);
		if (sel) {
			SaveSelection(sel, ST_CHECK);
			cmpack_unref(sel);
		}
	}
}


//
// Save selection
//
void CSelection::SaveSelection(CmpackCatSelection *sel, tStarType type) const
{
	int count = CountStars(type);
	if (count>0) {
		int *buf = (int*)g_malloc(count*sizeof(int));
		GetStarList(type, buf, count);
		cmpack_cat_selection_set(sel, buf, count);
		g_free(buf);
	}

}

//--------------------------   CCD IMAGE INFO   ----------------------------------

//
// Default constructor
//
CImageInfo::CImageInfo():m_Width(0), m_Height(0), m_Magic(NULL), m_Filter(NULL),
	m_Object(NULL), m_Observer(NULL), m_JD(0), m_ExpTime(0), m_CCDTemp(-999.9),
	m_ObjCoordsValid(false), m_ObsCoordsValid(false), m_Frames(0)
{
}


// 
// Copy constructor
//
CImageInfo::CImageInfo(const CImageInfo &info)
{
	m_Width = info.m_Width;
	m_Height = info.m_Height;
	m_Magic = MakeString(info.m_Magic);
	m_Filter = MakeString(info.m_Filter);
	m_Object = MakeString(info.m_Object);
	m_Observer = MakeString(info.m_Observer);
	m_JD = info.m_JD;
	m_ExpTime = info.m_ExpTime;
	m_CCDTemp = info.m_CCDTemp;
	m_ObjCoordsValid = info.m_ObjCoordsValid;
	m_ObsCoordsValid = info.m_ObsCoordsValid;
	m_ObjRA = info.m_ObjRA;
	m_ObjDec = info.m_ObjDec;
	m_ObsLon = info.m_ObsLon;
	m_ObsLat = info.m_ObsLat;
	m_Frames = info.m_Frames;
}


//
// Destructor
//
CImageInfo::~CImageInfo()
{
	FreeString(m_Magic);
	FreeString(m_Filter);
	FreeString(m_Object);
	FreeString(m_Observer);
}


//
// assignment operator
//
const CImageInfo &CImageInfo::operator=(const CImageInfo &info)
{
	m_Width = info.m_Width;
	m_Height = info.m_Height;
	m_Magic = CopyString(m_Magic, info.m_Magic);
	m_Filter = CopyString(m_Filter, info.m_Filter);
	m_Object = CopyString(m_Object, info.m_Object);
	m_Observer = CopyString(m_Observer, info.m_Observer);
	m_JD = info.m_JD;
	m_ExpTime = info.m_ExpTime;
	m_CCDTemp = info.m_CCDTemp;
	m_ObjCoordsValid = info.m_ObjCoordsValid;
	m_ObsCoordsValid = info.m_ObsCoordsValid;
	m_ObjRA = info.m_ObjRA;
	m_ObjDec = info.m_ObjDec;
	m_ObsLon = info.m_ObsLon;
	m_ObsLat = info.m_ObsLat;
	m_Frames = info.m_Frames;
	return *this;
}


//
// Clear data
//
void CImageInfo::Clear()
{
	m_Width = m_Height = 0;
	FreeString(m_Magic);
	m_Magic = NULL;
	FreeString(m_Filter);
	m_Filter = NULL;
	FreeString(m_Object);
	m_Object = NULL;
	FreeString(m_Observer);
	m_Observer = NULL;
	m_JD = m_ExpTime = 0;
	m_CCDTemp = -999.9;
	m_ObjCoordsValid = false;
	m_ObsCoordsValid = false;
	m_Frames = 0;
}


//
// Get object's coordinates
//
bool CImageInfo::ObjectCoordinates(double *ra, double *dec) const
{
	if (m_ObjCoordsValid) {
		if (ra)
			*ra = m_ObjRA;
		if (dec)
			*dec = m_ObjDec;
		return true;
	}
	return false;
}

//
// Get observer's coordinates
//
bool CImageInfo::ObserverCoordinates(double *lon, double *lat) const
{
	if (m_ObsCoordsValid) {
		if (lon)
			*lon = m_ObsLon;
		if (lat)
			*lat = m_ObsLat;
		return true;
	}
	return false;
}

//
// Set image info from a file
//
void CImageInfo::Set(CmpackCcdFile *file)
{
	const char *aux;
	CmpackDateTime dt;

	Clear();

	if (cmpack_ccd_get_datetime(file, &dt)==0) 
		m_JD = cmpack_encodejd(&dt);
	cmpack_ccd_get_size(file, &m_Width, &m_Height);
	cmpack_ccd_get_exptime(file, &m_ExpTime);
	cmpack_ccd_get_ccdtemp(file, &m_CCDTemp);
	cmpack_ccd_get_frames(file, &m_Frames);
	aux = cmpack_ccd_get_format(file);
	if (aux)
		m_Magic = DecodeString(m_Magic, aux);
	aux = cmpack_ccd_get_filter(file);
	if (aux)
		m_Filter = DecodeString(m_Filter, aux);
	aux = cmpack_ccd_get_object(file);
	if (aux)
		m_Object = DecodeString(m_Object, aux);
	aux = cmpack_ccd_get_observer(file);
	if (aux)
		m_Observer = DecodeString(m_Observer, aux);
}

//--------------------------   CHART INFO   ----------------------------------

//
// Default constructor
//
CChartInfo::CChartInfo():m_Width(0), m_Height(0), m_Stars(0), m_Filter(NULL), 
	m_JD(0), m_ExpTime(0)
{
}


// 
// Copy constructor
//
CChartInfo::CChartInfo(const CChartInfo &info)
{
	m_Width = info.m_Width;
	m_Height = info.m_Height;
	m_Filter = MakeString(info.m_Filter);
	m_JD = info.m_JD;
	m_ExpTime = info.m_ExpTime;
	m_Stars = info.m_Stars;
}


//
// Destructor
//
CChartInfo::~CChartInfo()
{
	FreeString(m_Filter);
}


//
// assignment operator
//
const CChartInfo &CChartInfo::operator=(const CChartInfo &info)
{
	m_Width = info.m_Width;
	m_Height = info.m_Height;
	m_Filter = CopyString(m_Filter, info.m_Filter);
	m_JD = info.m_JD;
	m_ExpTime = info.m_ExpTime;
	m_Stars = info.m_Stars;
	return *this;
}


//
// Clear data
//
void CChartInfo::Clear()
{
	m_Width = m_Height = 0;
	FreeString(m_Filter);
	m_Filter = NULL;
	m_JD = m_ExpTime = 0;
	m_Stars = 0;
}


//
// Set image info from a file
//
void CChartInfo::Set(CmpackPhtFile *file)
{
	const char *filter;

	Clear();

	cmpack_pht_get_size(file, &m_Width, &m_Height);
	m_JD = cmpack_pht_get_jd(file);
	m_ExpTime = cmpack_pht_get_exptime(file);
	filter = cmpack_pht_get_filter(file);
	if (filter) 
		m_Filter = DecodeString(m_Filter, filter);
	m_Stars = cmpack_pht_nstar(file);
}


//
// Set image info from a file
//
void CChartInfo::Set(CmpackCatFile *file)
{
	const char *filter;

	Clear();

	cmpack_cat_get_size(file, &m_Width, &m_Height);
	m_JD = cmpack_cat_get_jd(file);
	m_ExpTime = cmpack_cat_get_exptime(file);
	filter = cmpack_cat_get_filter(file);
	if (filter) 
		m_Filter = DecodeString(m_Filter, filter);
	m_Stars = cmpack_cat_nstar(file);
}


//----------------------   OBJECT COORDINATES   -----------------------------

//
// Default constructor
//
CObjectCoords::CObjectCoords():m_Name(NULL), m_RA(NULL), m_Dec(NULL)
{
}


//
// Copy constructor
//
CObjectCoords::CObjectCoords(const CObjectCoords &orig)
{
	m_Name = g_strdup(orig.m_Name);
	m_RA = g_strdup(orig.m_RA);
	m_Dec = g_strdup(orig.m_Dec);
}


//
// Destructor
//
CObjectCoords::~CObjectCoords()
{
	g_free(m_Name);
	g_free(m_RA);
	g_free(m_Dec);
}


// assignment operator
//
const CObjectCoords &CObjectCoords::operator=(const CObjectCoords &orig)
{
	g_free(m_Name);
	g_free(m_RA);
	g_free(m_Dec);
	m_Name = g_strdup(orig.m_Name);
	m_RA = g_strdup(orig.m_RA);
	m_Dec = g_strdup(orig.m_Dec);
	return *this;
}


//
// Set object's designation
//
void CObjectCoords::SetName(const char *name)
{
	g_free(m_Name);
	m_Name = g_strdup(name);
}


//
// Set object's coordinates
//
void CObjectCoords::SetRA(const char *ra)
{
	g_free(m_RA);
	m_RA = g_strdup(ra);
}
void CObjectCoords::SetDec(const char *dec)
{
	g_free(m_Dec);
	m_Dec = g_strdup(dec);
}


//
// Set object's coordinates
//
void CObjectCoords::Load(CmpackCcdFile *file)
{
	g_free(m_Name);
	g_free(m_RA);
	g_free(m_Dec);
	m_Name = g_strdup(cmpack_ccd_get_object(file));
	m_RA = g_strdup(cmpack_ccd_get_objra(file));
	m_Dec = g_strdup(cmpack_ccd_get_objdec(file));
}


//
// Load selection from a file
//
void CObjectCoords::Load(CmpackCatFile *file)
{
	g_free(m_Name);
	g_free(m_RA);
	g_free(m_Dec);
	m_Name = g_strdup(cmpack_cat_get_object(file));
	m_RA = g_strdup(cmpack_cat_get_objra(file));
	m_Dec = g_strdup(cmpack_cat_get_objdec(file));
}


//
// Clear data
//
void CObjectCoords::Clear()
{
	g_free(m_Name);
	g_free(m_RA);
	g_free(m_Dec);
	m_Name = NULL;
	m_RA = NULL;
	m_Dec = NULL;
}

//----------------------   OBSERVER COORDINATES   -----------------------------

//
// Default constructor
//
CObserverCoords::CObserverCoords():m_Name(NULL), m_Lon(NULL), m_Lat(NULL)
{
}


//
// Copy constructor
//
CObserverCoords::CObserverCoords(const CObserverCoords &orig)
{
	m_Name = g_strdup(orig.m_Name);
	m_Lon = g_strdup(orig.m_Lon);
	m_Lat = g_strdup(orig.m_Lat);
}


//
// Destructor
//
CObserverCoords::~CObserverCoords()
{
	g_free(m_Name);
	g_free(m_Lon);
	g_free(m_Lat);
}


// assignment operator
//
const CObserverCoords &CObserverCoords::operator=(const CObserverCoords &orig)
{
	g_free(m_Name);
	g_free(m_Lon);
	g_free(m_Lat);
	m_Name = g_strdup(orig.m_Name);
	m_Lon = g_strdup(orig.m_Lon);
	m_Lat = g_strdup(orig.m_Lat);
	return *this;
}


//
// Set observer's designation
//
void CObserverCoords::SetName(const char *name)
{
	g_free(m_Name);
	m_Name = g_strdup(name);
}


//
// Set observer's coordinates
//
void CObserverCoords::SetLon(const char *ra)
{
	g_free(m_Lon);
	m_Lon = g_strdup(ra);
}
void CObserverCoords::SetLat(const char *dec)
{
	g_free(m_Lat);
	m_Lat = g_strdup(dec);
}


//
// Set observer's coordinates
//
void CObserverCoords::Load(CmpackCcdFile *file)
{
	g_free(m_Name);
	g_free(m_Lon);
	g_free(m_Lat);
	m_Name = g_strdup(cmpack_ccd_get_observatory(file));
	m_Lon = g_strdup(cmpack_ccd_get_obslon(file));
	m_Lat = g_strdup(cmpack_ccd_get_obslat(file));
}


//
// Clear data
//
void CObserverCoords::Clear()
{
	g_free(m_Name);
	g_free(m_Lon);
	g_free(m_Lat);
	m_Name = NULL;
	m_Lon = NULL;
	m_Lat = NULL;
}

//-----------------------   TABLE HEADER INFO   ----------------------------

//
// Default constructor
//
CTableInfo::CTableInfo(void):m_Filter(NULL), m_Aperture(0)
{
}


//
// Copy constructor
//
CTableInfo::CTableInfo(const CTableInfo &orig):m_Aperture(orig.m_Aperture)
{
	m_Filter = g_strdup(orig.m_Filter);
}


//
// Destructor
//
CTableInfo::~CTableInfo()
{
	g_free(m_Filter);
}


//
// Clear data
//
void CTableInfo::Clear(void)
{
	g_free(m_Filter);
	m_Filter = NULL;
	m_Aperture = 0;
}


//
// Copy data from context
//
void CTableInfo::Set(CmpackTable *tab)
{
	g_free(m_Filter);
	m_Filter = g_strdup(cmpack_tab_gkys(tab, "Filter"));
	m_Aperture = cmpack_tab_gkyi(tab, "Aperture", 0);
}


//
// Assignment operator
//
const CTableInfo &CTableInfo::operator=(const CTableInfo &orig)
{
	g_free(m_Filter);
	m_Filter = g_strdup(orig.m_Filter);
	m_Aperture = orig.m_Aperture;
	return *this;
}

//-----------------------   READ-ALL FILE INFO   ----------------------------

//
// Default constructor
//
CReadAllInfo::CReadAllInfo(void):m_Filter(NULL), m_Aperture(0), m_HelCor(0)
{
}


//
// Copy constructor
//
CReadAllInfo::CReadAllInfo(const CReadAllInfo &orig):m_Aperture(orig.m_Aperture),
	m_HelCor(orig.m_HelCor)
{
	m_Filter = g_strdup(orig.m_Filter);
}


//
// Destructor
//
CReadAllInfo::~CReadAllInfo()
{
	g_free(m_Filter);
}


//
// Clear data
//
void CReadAllInfo::Clear(void)
{
	g_free(m_Filter);
	m_Filter = NULL;
	m_Aperture = 0;
	m_HelCor = false;
}


//
// Copy data from context
//
void CReadAllInfo::Set(CmpackAllFile *file)
{
	g_free(m_Filter);
	m_Filter = g_strdup(cmpack_all_get_filter(file));
	m_Aperture = cmpack_all_get_aperture(file);
	m_HelCor = cmpack_all_get_helcor(file);
}


//
// Assignment operator
//
const CReadAllInfo &CReadAllInfo::operator=(const CReadAllInfo &orig)
{
	g_free(m_Filter);
	m_Filter = g_strdup(orig.m_Filter);
	m_Aperture = orig.m_Aperture;
	m_HelCor = orig.m_HelCor;
	return *this;
}

//------------------------   FILE INFO   --------------------------

//
// Default constructor
//
CFileInfo::CFileInfo():m_FullPath(NULL)
{
	memset(&m_MTime, 0, sizeof(GTimeVal));
}


//
// Constructor with initialization
//
CFileInfo::CFileInfo(const gchar *path):m_FullPath(NULL)
{
	char *fullpath;
	struct stat fs;

	memset(&m_MTime, 0, sizeof(GTimeVal));
	if (path) {
		if (g_path_is_absolute(path))
			fullpath = g_strdup(path);
		else
			fullpath = g_build_filename(get_temp_dir(), path, NULL);

		char *f = g_locale_from_utf8(fullpath, -1, NULL, NULL, NULL);
		if (stat(f, &fs)==0) {
			m_MTime = fs.st_mtime;
			m_FullPath = fullpath;
		} else {
			g_free(fullpath);
		}
		g_free(f);
	}
}


//
// Copy constructor
//
CFileInfo::CFileInfo(const CFileInfo &orig)
{
	m_FullPath = g_strdup(orig.m_FullPath);
	m_MTime = orig.m_MTime;
}


//
// Destructor
//
CFileInfo::~CFileInfo()
{
	g_free(m_FullPath);
}


//
// Assignment operator
//
const CFileInfo &CFileInfo::operator=(const CFileInfo &orig)
{
	g_free(m_FullPath);
	m_FullPath = g_strdup(orig.m_FullPath);
	m_MTime = orig.m_MTime;
	return *this;
}


//
// Clear data
//
void CFileInfo::Clear(void)
{
	g_free(m_FullPath);
	m_FullPath = NULL;
	memset(&m_MTime, 0, sizeof(GTimeVal));
}


//
// Load from project file
//
void CFileInfo::Load(GKeyFile *cfg, const gchar *section, const gchar *prefix)
{
	gchar key[128], *val;

	g_free(m_FullPath);
	sprintf(key, "%s.path", prefix);
	val = g_key_file_get_string(cfg, section, key, NULL);
	if (val && strlen(val)>0)
		m_FullPath = g_strdup(val);
	g_free(val);

	memset(&m_MTime, 0, sizeof(GTimeVal));
	sprintf(key, "%s.time", prefix);
	val = g_key_file_get_string(cfg, section, key, NULL);
	if (val && strlen(val)>0) {
#ifdef _MSC_VER
		if (sizeof(time_t)==sizeof(long)) 
			sscanf(val, " %ld ", (long*)&m_MTime);
		else 
			sscanf(val, " %lld ", (long long*)&m_MTime);
#else
		sscanf(val, " %ld ", (long*)&m_MTime);
#endif	
	}
	g_free(val);
}


//
// Save to project file
//
void CFileInfo::Save(GKeyFile *cfg, const gchar *section, const gchar *prefix)
{
	gchar key[128], val[128];

	sprintf(key, "%s.path", prefix);
	if (m_FullPath) 
		g_key_file_set_string(cfg, section, key, m_FullPath);
	else
		g_key_file_remove_key(cfg, section, key, NULL);

	sprintf(key, "%s.time", prefix);
	if (m_MTime!=0) {
#ifdef _MSC_VER		
		if (sizeof(time_t)==sizeof(long)) 
			sprintf(val, "%ld", (long)m_MTime);
		else
			sprintf(val, "%lld", (long long)m_MTime);
#else
		sprintf(val, "%ld", (long)m_MTime);
#endif			
		g_key_file_set_string(cfg, section, key, val);
	} else {
		g_key_file_remove_key(cfg, section, key, NULL);
	}
}


//
// Equivalence operator
//
bool operator==(const CFileInfo &a, const CFileInfo &b)
{
	if (!a.m_FullPath && !b.m_FullPath)
		return true;
	else if (!a.m_FullPath || !b.m_FullPath)
		return false;
	else if (strcmp(a.m_FullPath, b.m_FullPath)!=0)
		return false;
	else
		return a.m_MTime == b.m_MTime;
}

//
// Non-equivalence operator
//
bool operator!=(const CFileInfo &a, const CFileInfo &b)
{
	if (!a.m_FullPath && !b.m_FullPath)
		return false;
	else if (!a.m_FullPath || !b.m_FullPath)
		return true;
	else if (strcmp(a.m_FullPath, b.m_FullPath)!=0)
		return true;
	else
		return a.m_MTime == b.m_MTime;
}
