/**************************************************************

menubar.cpp (C-Munipack project)
Menu wrapper
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <glib.h>

#include "menubar.h"
#include "utils.h"

struct tMenuData
{
	CMenuBar		*pMe;
	GtkWidget		*pItem;
	int				menu_id, cmd_id;
	//guint			handler_id;
	GSList			*items;
	tMenuData		*parent;
};


// Default constructor
CMenuBar::CMenuBar()
{
	m_Data = NULL;
	m_hBar = NULL;
	m_Updating = false;
}

// Destructor
CMenuBar::~CMenuBar()
{
	g_slist_foreach(m_Data, (GFunc)FreeItem, NULL);
	g_slist_free(m_Data);
}

// Create and initialize the m_hBar
void CMenuBar::Create(const tMenu *menus, bool icons)
{
	int i;

	m_Updating = true;
	if (menus) {
		m_hBar = gtk_menu_bar_new();
		for (i=0; menus[i].text!=NULL; i++) {
			tMenuData *data = CreateMenu(&menus[i], icons);
			m_Data = g_slist_append(m_Data, data);
			gtk_menu_shell_append(GTK_MENU_SHELL(m_hBar), data->pItem);
			g_object_ref(data->pItem);
		}
	}
	gtk_widget_show(m_hBar);
	m_Updating = false;
}


tMenuData *CMenuBar::CreateMenu(const tMenu *menu, bool icons)
{
	int i;

	tMenuData *data = (tMenuData*)g_malloc0(sizeof(tMenuData));
	data->cmd_id = -1;
	data->menu_id = menu->menu_id;
	data->pMe = this;

	data->pItem = gtk_menu_item_new_with_mnemonic(menu->text);
	if ((menu->flags & MBF_RIGHT)!=0)
		gtk_menu_item_right_justify(GTK_MENU_ITEM(data->pItem));

	if (menu->items!=NULL) {
		GtkWidget *pMenu = gtk_menu_new();
		for (i=0; menu->items[i].type!=MB_END; i++) {
			tMenuData *child = CreateItem(&menu->items[i], icons);
			data->items = g_slist_append(data->items, child);
			gtk_menu_shell_append(GTK_MENU_SHELL(pMenu), child->pItem);
			gtk_widget_show(child->pItem);
			g_object_ref(child->pItem);
		}
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(data->pItem), pMenu);
		gtk_widget_show(pMenu);
	}

	gtk_widget_show(data->pItem);
	return data;
}

tMenuData *CMenuBar::CreateItem(const tMenuItem *item, bool icons)
{
	int i;

	tMenuData *data = (tMenuData*)g_malloc0(sizeof(tMenuData));
	data->cmd_id = item->cmd_id;
	data->menu_id = -1;
	data->pMe = this;

	switch (item->type)
	{
	case MB_ITEM:
		data->pItem = gtk_image_menu_item_new_with_mnemonic(item->text);
		if (icons) {
			char *icon = get_icon_file(item->icon ? item->icon : "noicon");
			gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(data->pItem), 
				gtk_image_new_from_file(icon));
			g_free(icon);
		}
		g_signal_connect(G_OBJECT(data->pItem), "activate", G_CALLBACK(item_activate), data);
		break;

	case MB_CHECKBTN:
	case MB_RADIOBTN:
		data->pItem = gtk_check_menu_item_new_with_mnemonic(item->text);
		if (item->type==MB_RADIOBTN)
			gtk_check_menu_item_set_draw_as_radio(GTK_CHECK_MENU_ITEM(data->pItem), TRUE);
		g_signal_connect(G_OBJECT(data->pItem), "toggled", G_CALLBACK(item_activate), data);
		break;

	case MB_SEPARATOR:
		data->pItem = gtk_separator_menu_item_new();
		break;

	case MB_SUBMENU:
		data->pItem = gtk_menu_item_new_with_mnemonic(item->text);
		if (item->submenu!=NULL) {
			GtkWidget *pMenu = gtk_menu_new();
			for (i=0; item->submenu[i].type!=MB_END; i++) {
				tMenuData *child = CreateItem(&item->submenu[i], icons);
				data->items = g_slist_append(data->items, child);
				gtk_menu_shell_append(GTK_MENU_SHELL(pMenu), child->pItem);
				gtk_widget_show(child->pItem);
				g_object_ref(child->pItem);
			}
			gtk_menu_item_set_submenu(GTK_MENU_ITEM(data->pItem), pMenu);
			gtk_widget_show(pMenu);
		}
		break;
		
	default:
		break;
	}

	gtk_widget_show(data->pItem);
	return data;
}

// Free memory allocated for a menu
void CMenuBar::FreeItem(tMenuData *menu, void *user_data)
{
	if (menu->items) {
		g_slist_foreach(menu->items, (GFunc)FreeItem, NULL);
		g_slist_free(menu->items);
	}
	if (menu->pItem) {
		/*if (menu->handler_id>0) {
			if (g_signal_handler_is_connected(G_OBJECT(menu->pItem), menu->handler_id))
				g_signal_handler_disconnect(G_OBJECT(menu->pItem), menu->handler_id);
		}*/
		g_object_unref(menu->pItem);
	}
	g_free(menu);
}

void CMenuBar::item_activate(GtkWidget *widget, gpointer data)
{
	((tMenuData*)data)->pMe->OnActivate(((tMenuData*)data)->cmd_id);
}

void CMenuBar::OnActivate(int cmd_id)
{
	if (!m_Updating)
		Callback(CB_ACTIVATE, cmd_id);
}

void CMenuBar::ShowMenu(int menu_id, bool show)
{
	GSList *ptr;

	m_Updating = true;
	for (ptr=m_Data; ptr!=NULL; ptr=ptr->next) {
		tMenuData *data = (tMenuData*)ptr->data;
		if (data->menu_id == menu_id) {
			if (show)
				gtk_widget_show(GTK_WIDGET(data->pItem));
			else
				gtk_widget_hide(GTK_WIDGET(data->pItem));
			break;
		}
	}
	m_Updating = false;
}

void CMenuBar::Show(int cmd_id, bool show)
{
	bool found = false;
	GSList *ptr, *ptr2;

	m_Updating = true;
	for (ptr=m_Data; ptr!=NULL && !found; ptr=ptr->next) {
		tMenuData *data = (tMenuData*)ptr->data;
		for (ptr2=data->items; ptr2!=NULL; ptr2=ptr2->next) {
			if (ShowItem((tMenuData*)ptr2->data, cmd_id, show)) {
				found = true;
				break;
			}
		}
	}
	m_Updating = false;
}

bool CMenuBar::ShowItem(tMenuData *data, int cmd_id, bool show)
{
	GSList *ptr;

	if (data->cmd_id == cmd_id) {
		if (show)
			gtk_widget_show(GTK_WIDGET(data->pItem));
		else
			gtk_widget_hide(GTK_WIDGET(data->pItem));
		return true;
	}

	for (ptr=data->items; ptr!=NULL; ptr=ptr->next) {
		if (ShowItem((tMenuData*)ptr->data, cmd_id, show))
			return true;
	}

	return false;
}

void CMenuBar::EnableMenu(int menu_id, bool enable)
{
	GSList *ptr;

	m_Updating = true;
	for (ptr=m_Data; ptr!=NULL; ptr=ptr->next) {
		tMenuData *data = (tMenuData*)ptr->data;
		if (data->menu_id == menu_id) {
			gtk_widget_set_sensitive(GTK_WIDGET(data->pItem), enable);
			break;
		}
	}
	m_Updating = false;
}

void CMenuBar::Enable(int cmd_id, bool enable)
{
	bool found = false;
	GSList *ptr, *ptr2;

	m_Updating = true;
	for (ptr=m_Data; ptr!=NULL && !found; ptr=ptr->next) {
		tMenuData *data = (tMenuData*)ptr->data;
		for (ptr2=data->items; ptr2!=NULL; ptr2=ptr2->next) {
			if (EnableItem((tMenuData*)ptr2->data, cmd_id, enable)) {
				found = true;
				break;
			}
		}
	}
	m_Updating = false;
}

bool CMenuBar::EnableItem(tMenuData *data, int cmd_id, bool enable)
{
	GSList *ptr;

	if (data->cmd_id == cmd_id) {
		gtk_widget_set_sensitive(GTK_WIDGET(data->pItem), enable);
		return true;
	}

	for (ptr=data->items; ptr!=NULL; ptr=ptr->next) {
		if (EnableItem((tMenuData*)ptr->data, cmd_id, enable))
			return true;
	}

	return false;
}

void CMenuBar::Check(int cmd_id, bool check)
{
	bool found = false;
	GSList *ptr, *ptr2;

	m_Updating = true;
	for (ptr=m_Data; ptr!=NULL && !found; ptr=ptr->next) {
		tMenuData *data = (tMenuData*)ptr->data;
		for (ptr2=data->items; ptr2!=NULL; ptr2=ptr2->next) {
			if (CheckItem((tMenuData*)ptr2->data, cmd_id, check)) {
				found = true;
				break;
			}
		}
	}
	m_Updating = false;
}

bool CMenuBar::CheckItem(tMenuData *data, int cmd_id, bool check)
{
	GSList *ptr;

	if (data->cmd_id == cmd_id) {
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(data->pItem), check);
		return true;
	}

	for (ptr=data->items; ptr!=NULL; ptr=ptr->next) {
		if (CheckItem((tMenuData*)ptr->data, cmd_id, check))
			return true;
	}

	return false;
}

bool CMenuBar::IsChecked(int cmd_id)
{
	bool checked = false, found = false;
	GSList *ptr, *ptr2;

	m_Updating = true;
	for (ptr=m_Data; ptr!=NULL && !found; ptr=ptr->next) {
		tMenuData *data = (tMenuData*)ptr->data;
		for (ptr2=data->items; ptr2!=NULL; ptr2=ptr2->next)  {
			if (IsItemChecked((tMenuData*)ptr2->data, cmd_id, &checked)) {
				found = true;
				break;
			}
		}
	}
	m_Updating = false;
	return checked;
}

bool CMenuBar::IsItemChecked(tMenuData *data, int cmd_id, bool *value)
{
	GSList *ptr;

	if (data->cmd_id == cmd_id) {
		*value = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(data->pItem))!=FALSE;
		return true;
	}

	for (ptr=data->items; ptr!=NULL; ptr=ptr->next) {
		if (CheckItem((tMenuData*)ptr->data, cmd_id, value))
			return true;
	}
	return false;
}
