/**************************************************************

newfiles_dlg.cpp (C-Munipack project)
The 'Process new files' dialog
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <cmunipack.h>

#include "configuration.h"
#include "project.h"
#include "newfiles_dlg.h"
#include "progress_dlg.h"
#include "main.h"
#include "utils.h"
#include "proc_classes.h"
#include "lightcurve_dlg.h"

struct tFileData
{
	char *path;
	char *filter, *object;
	double jd, exptime;
};

static tFileData *MakeFile(const char *fpath)
{
	char	*f, *str;
	const char *filter, *object;
	CmpackCcdFile *file;
	CmpackDateTime dt;
	tFileData *data = NULL;

	if (Global.pProject->FileInProject(fpath))
		return NULL;

	f = g_locale_from_utf8(fpath, -1, NULL, NULL, NULL);
	if (cmpack_ccd_open(&file, f, CMPACK_OPEN_HEADONLY)==0) {
		data = (tFileData*)g_malloc0(sizeof(tFileData));
		data->path = g_strdup(fpath);
		filter = cmpack_ccd_get_filter(file);
		if (filter) {
			str = g_locale_to_utf8(filter, -1, NULL, NULL, NULL);
			if (str)
				g_strstrip(str);
			if (str && strlen(str)>0)
				data->filter = str;
			else
				g_free(str);
		}
		if (cmpack_ccd_get_datetime(file, &dt)==0) 
			data->jd = cmpack_encodejd(&dt);
		cmpack_ccd_get_exptime(file, &data->exptime);
		object = cmpack_ccd_get_object(file);
		if (object) {
			str = g_locale_to_utf8(object, -1, NULL, NULL, NULL);
			if (str)
				g_strstrip(str);
			if (str && strlen(str)>0)
				data->object = str;
			else
				g_free(str);
		}
		cmpack_unref(file);
	}
	g_free(f);
	return data;
}

static void FileFree(tFileData *data)
{
	if (data) {
		g_free(data->path);
		g_free(data->filter);
		g_free(data->object);
		g_free(data);
	}
}

static gint CompareByTime(const tFileData *a, const tFileData *b)
{
	// Compare by time
	if (a->jd < b->jd)
		return -1;
	else if (a->jd > b->jd)
		return 1;
	
	// If they are equal, compare by path
	if (a->path) {
		if (!b->path)
			return -1;
		else
			return strcmp(a->path, b->path);
	} else {
		if (b->path)
			return 1;
		else
			return 0;
	}
}

//-------------------------   MAIN WINDOW   --------------------------------

CNewFilesDlg::CNewFilesDlg(GtkWindow *pParent)
{
	GtkWidget *frame, *tbox;
	GSList *group;

	// Dialog with buttons
	m_pDlg = gtk_dialog_new_with_buttons("Process new files", pParent, 
		(GtkDialogFlags)(GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR),
		GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, NULL);

	// Dialog icon
	gchar *icon = get_icon_file("newfiles");
	gtk_window_set_icon(GTK_WINDOW(m_pDlg), gdk_pixbuf_new_from_file(icon, NULL));
	g_free(icon);

	// Frame
	frame = gtk_frame_new("Process");
	gtk_container_add (GTK_CONTAINER(GTK_DIALOG(m_pDlg)->vbox), frame);
	tbox = gtk_table_new(8, 3, FALSE);
	gtk_container_add(GTK_CONTAINER(frame), tbox);
	gtk_container_set_border_width(GTK_CONTAINER(tbox), 8);
	gtk_table_set_row_spacings(GTK_TABLE(tbox), 8);
	gtk_table_set_col_spacings(GTK_TABLE(tbox), 16);
	gtk_table_set_col_spacing(GTK_TABLE(tbox), 0, 32);
	
	m_AllBtn = gtk_radio_button_new_with_label(NULL, "all unprocessed frames in the current project");
	gtk_table_attach_defaults(GTK_TABLE(tbox), m_AllBtn, 0, 3, 0, 1);

	group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(m_AllBtn));
	m_FilesBtn = gtk_radio_button_new_with_label(group, "process new files in the directory");
	gtk_table_attach_defaults(GTK_TABLE(tbox), m_FilesBtn, 0, 3, 1, 2);

	m_FolderLabel = gtk_label_new("Directory");
	gtk_misc_set_alignment(GTK_MISC(m_FolderLabel), 0.0, 0.5);
	gtk_table_attach_defaults(GTK_TABLE(tbox), m_FolderLabel, 1, 2, 2, 3);
	m_Folder = gtk_file_chooser_button_new(NULL, GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);
	gtk_table_attach_defaults(GTK_TABLE(tbox), m_Folder, 2, 3, 2, 3);

	m_SubDirs = gtk_check_button_new_with_label("Include subdirectories");
	gtk_table_attach_defaults(GTK_TABLE(tbox), m_SubDirs, 1, 3, 3, 4);

	m_ProcLabel = gtk_label_new("Process only files which meet the following conditions:");
	gtk_misc_set_alignment(GTK_MISC(m_ProcLabel), 0.0, 0.5);
	gtk_table_attach_defaults(GTK_TABLE(tbox), m_ProcLabel, 1, 3, 4, 5);

	m_NameCheck = gtk_check_button_new_with_label("File name");
	gtk_table_attach_defaults(GTK_TABLE(tbox), m_NameCheck, 1, 2, 5, 6);
	m_NameEdit = gtk_entry_new();
	gtk_table_attach_defaults(GTK_TABLE(tbox), m_NameEdit, 2, 3, 5, 6);

	m_FilterCheck = gtk_check_button_new_with_label("Filter");
	gtk_table_attach_defaults(GTK_TABLE(tbox), m_FilterCheck, 1, 2, 6, 7);
	m_FilterEdit = gtk_entry_new();
	gtk_table_attach_defaults(GTK_TABLE(tbox), m_FilterEdit, 2, 3, 6, 7);

	m_ObjectCheck = gtk_check_button_new_with_label("Object name");
	gtk_table_attach_defaults(GTK_TABLE(tbox), m_ObjectCheck, 1, 2, 7, 8);
	m_ObjectEdit = gtk_entry_new();
	gtk_table_attach_defaults(GTK_TABLE(tbox), m_ObjectEdit, 2, 3, 7, 8);

	// Default state
	int mode = CConfig::GetInt("NewFiles", "Mode", 0);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(m_AllBtn), !mode);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(m_FilesBtn), mode);

	// Go to last folder
	gchar *aux = CConfig::GetStr("NewFiles", "Folder", NULL);
	if (aux) {
		gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(m_Folder), aux);
		g_free(aux);
	}
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(m_SubDirs), 
		CConfig::GetBool("NewFiles", "SubDirs"));
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(m_NameCheck), 
		CConfig::GetBool("NewFiles", "Name"));
	aux = CConfig::GetStr("NewFiles", "NameText", "");
	if (aux) {
		gtk_entry_set_text(GTK_ENTRY(m_NameEdit), aux);
		g_free(aux);
	}
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(m_FilterCheck), 
		CConfig::GetBool("NewFiles", "Filter"));
	aux = CConfig::GetStr("NewFiles", "FilterText", "");
	if (aux) {
		gtk_entry_set_text(GTK_ENTRY(m_FilterEdit), aux);
		g_free(aux);
	}
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(m_ObjectCheck), 
		CConfig::GetBool("NewFiles", "Object"));
	aux = CConfig::GetStr("NewFiles", "ObjectText", "");
	if (aux) {
		gtk_entry_set_text(GTK_ENTRY(m_ObjectEdit), aux);
		g_free(aux);
	}

	gtk_widget_show_all(frame);

	// Signals:
	g_signal_connect(G_OBJECT(m_AllBtn), "toggled", G_CALLBACK(button_toggled), this);
	g_signal_connect(G_OBJECT(m_FilesBtn), "toggled", G_CALLBACK(button_toggled), this);
	g_signal_connect(G_OBJECT(m_NameCheck), "toggled", G_CALLBACK(button_toggled), this);
	g_signal_connect(G_OBJECT(m_FilterCheck), "toggled", G_CALLBACK(button_toggled), this);
	g_signal_connect(G_OBJECT(m_ObjectCheck), "toggled", G_CALLBACK(button_toggled), this);

	UpdateControls();
}

CNewFilesDlg::~CNewFilesDlg()
{
	gtk_widget_destroy(m_pDlg);
}

void CNewFilesDlg::button_toggled(GtkToggleButton *widget, CNewFilesDlg *pMe)
{
	pMe->OnButtonToggled(widget);
}

void CNewFilesDlg::OnButtonToggled(GtkToggleButton *widget)
{
	UpdateControls();
}

gboolean CNewFilesDlg::foreach_all_files(GtkTreeModel *model, GtkTreePath *path, 
										GtkTreeIter *iter, gpointer userdata)
{
	int state;
	GList **list = ((GList**)userdata);
	gtk_tree_model_get(model, iter, FRAME_STATE, &state, -1);
	if ((state & CFILE_CONVERSION)==0) {
		GtkTreeRowReference *rowref = gtk_tree_row_reference_new(model, path);
		*list = g_list_append(*list, rowref);
	}
	return FALSE;
}

void CNewFilesDlg::UpdateControls(void)
{
	bool ok = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(m_FilesBtn));
	gtk_widget_set_sensitive(m_FolderLabel, ok);
	gtk_widget_set_sensitive(m_SubDirs, ok);
	gtk_widget_set_sensitive(m_Folder, ok);
	gtk_widget_set_sensitive(m_ProcLabel, ok);
	gtk_widget_set_sensitive(m_FilterCheck, ok);
	gtk_widget_set_sensitive(m_FilterEdit, ok && 
		gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(m_FilterCheck)));
	gtk_widget_set_sensitive(m_NameCheck, ok);
	gtk_widget_set_sensitive(m_NameEdit, ok &&
		gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(m_NameCheck)));
	gtk_widget_set_sensitive(m_ObjectCheck, ok);
	gtk_widget_set_sensitive(m_ObjectEdit, ok &&
		gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(m_ObjectCheck)));
}

void CNewFilesDlg::Execute()
{
	int mode, res;
	char *dirpath, msg[256];
	bool recursive, name_check, object_check, filter_check;
	const char *name, *object, *filter;
	GtkFileFilter *name_filter;
	gchar **filter_list, **object_list;
	GtkWindow *pMainWnd = Global.pMainWnd->GetHandle();

	m_InFiles = m_OutFiles = 0;

	if (gtk_dialog_run(GTK_DIALOG(m_pDlg))!=GTK_RESPONSE_ACCEPT)
		return;
	gtk_widget_hide(m_pDlg);

	mode = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(m_FilesBtn));
	CConfig::SetInt("NewFiles", "Mode", (!mode ? 0 : 1));
	dirpath = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(m_Folder));
	CConfig::SetStr("NewFiles", "Folder", dirpath);
	recursive = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(m_SubDirs));
	CConfig::SetBool("NewFiles", "SubDirs", recursive);
	name_check = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(m_NameCheck));
	CConfig::SetBool("NewFiles", "Name", name_check);
	name = gtk_entry_get_text(GTK_ENTRY(m_NameEdit));
	CConfig::SetStr("NewFiles", "NameText", name);
	filter_check  = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(m_FilterCheck));
	CConfig::SetBool("NewFiles", "Filter", filter_check);
	filter = gtk_entry_get_text(GTK_ENTRY(m_FilterEdit));
	CConfig::SetStr("NewFiles", "FilterText", filter);
	object_check = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(m_ObjectCheck));
	CConfig::SetBool("NewFiles", "Object", object_check);
	object = gtk_entry_get_text(GTK_ENTRY(m_ObjectEdit));
	CConfig::SetStr("NewFiles", "ObjectText", object);

	// Make list of files
	m_FileList = NULL;
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(m_AllBtn))) {
		// Make list of unprocessed frames
		gtk_tree_model_foreach(Global.pProject->FileList(), foreach_all_files, &m_FileList);
		if (!m_FileList) {
			ShowError(pMainWnd, "There are no unprocessed files in the project.");	
			g_free(dirpath);
			return;
		}
	} else {
		if (dirpath) {
			// Make filters
			if (name_check && name)
				name_filter = MakeFileFilter(name);
			else
				name_filter = NULL;
			if (filter_check && filter)
				filter_list = g_strsplit_set(filter, " ,;", 0);
			else
				filter_list = NULL;
			if (object_check && object)
				object_list = g_strsplit_set(object, " ,;", 0);
			else
				object_list = NULL;
			// Add new files to the project
			GList *list = PrepareFolder(NULL, dirpath, recursive, name_filter,
				filter_list, object_list);
			list = g_list_sort(list, (GCompareFunc)CompareByTime);
			for (GList *ptr=list; ptr!=NULL; ptr=ptr->next) {
				int frame = Global.pProject->AddFile(((tFileData*)ptr->data)->path);
				if (frame>0) {
					GtkTreePath *path = Global.pProject->GetFilePath(frame);
					if (path) {
						m_FileList = g_list_append(m_FileList, 
							gtk_tree_row_reference_new(Global.pProject->FileList(), path));
						gtk_tree_path_free(path);
					}
				}
				FileFree((tFileData*)ptr->data);
			}
			if (name_filter)
				g_object_unref(name_filter);
			g_strfreev(filter_list);
			g_strfreev(object_list);
			g_list_free(list);
		}
		if (!m_FileList) {
			ShowError(pMainWnd, "There are no unprocessed files that fit the specified conditions.");
			g_free(dirpath);
			return;
		}
	}
	g_free(dirpath);

	if (m_FileList) {
		CProgressDlg pDlg(Global.pMainWnd->GetHandle(), "Converting files...");
		pDlg.SetMinMax(0, g_list_length(m_FileList));
		res = pDlg.Execute(ExecuteProc, this);
		if (res!=0) {
			char *msg = cmpack_formaterror(res);
			ShowError(pMainWnd, msg);
			cmpack_free(msg);
		} else
		if (m_OutFiles==0) {
			ShowError(pMainWnd, "No file was successfully processed.");
		} else {
			const CSelection *sel = Global.pProject->Selection();
			if (Global.pProject->GetReference(NULL, NULL) &&
				sel->CountStars(ST_VARIABLE)>0 && sel->CountStars(ST_COMPARISON)>0) {
					// Show light curve 
					CLightCurveDlg pDlg(GTK_WINDOW(Global.pMainWnd->GetHandle()));
					pDlg.Execute();
			} else {
				// Show information about processed files
				if (m_OutFiles!=m_InFiles) {
					sprintf(msg, "%d file(s) was successfully processed, %d file(s) failed.", 
						m_OutFiles, m_InFiles-m_OutFiles);
					ShowWarning(pMainWnd, msg);
				} else {
					sprintf(msg, "All %d file(s) was successfully processed.", m_OutFiles);
					ShowInformation(pMainWnd, msg);
				}
			}
		}
		g_list_foreach(m_FileList, (GFunc)gtk_tree_row_reference_free, NULL);
		g_list_free(m_FileList);
		m_FileList = NULL;
	}
	Global.pProject->Save();
	Global.pProject->UpdateAllIcons();
}

GList *CNewFilesDlg::PrepareFolder(GList *list, const char *dirpath, bool recursive,
		GtkFileFilter *filenames, char **filters, char **objects)
{
	char *filepath;
	const gchar *filename;
	GDir *dir;

	dir = g_dir_open(dirpath, 0, NULL);
	if (dir) {
		filename = g_dir_read_name(dir);
		while (filename) {
			filepath = g_build_filename(dirpath, filename, NULL);
			if (g_file_test(filepath, G_FILE_TEST_IS_DIR)) {
				if (recursive)
					list = PrepareFolder(list, filepath, true, filenames, filters, objects);
			} else 
			if (g_file_test(filepath, G_FILE_TEST_IS_REGULAR)) {
				if (CheckFileName(filepath, filenames)) {
					tFileData *data = MakeFile(filepath);
					if (data) {
						if (CheckFileData(data, filters, objects))
							list = g_list_prepend(list, data);
					}
				}
			}
			g_free(filepath);
			filename = g_dir_read_name(dir);
		}
		g_dir_close(dir);
	}
	return list;
}

bool CNewFilesDlg::CheckFileName(const char *fpath, GtkFileFilter *filenames)
{
	GtkFileFilterInfo info;

	// Check file name
	if (filenames) {
		info.contains = GTK_FILE_FILTER_DISPLAY_NAME;
		info.display_name = g_path_get_basename(fpath);
		if (!gtk_file_filter_filter(filenames, &info))
			return false;
	}

	// Check if the file is already in the project
	if (Global.pProject->FileInProject(fpath))
		return false;

	// File is OK
	return true;
}

bool CNewFilesDlg::CheckFileData(const tFileData *data, char **filters, char **objects)
{
	bool ok = true;

	if (filters) 
		ok = TestString(filters, data->filter);
	if (ok && objects) 
		ok = TestString(objects, data->object);
	return ok;
}

GtkFileFilter *CNewFilesDlg::MakeFileFilter(const char *text)
{
	GtkFileFilter *filter = gtk_file_filter_new();
	char **list = g_strsplit_set(text, " ,;", 0);
	for (int i=0; list[i]!=NULL; i++) 
		gtk_file_filter_add_pattern(filter, list[i]);
	g_strfreev(list);
	return (GtkFileFilter*)g_object_ref(filter);
}

bool CNewFilesDlg::TestString(char **list, const char *str)
{
	if (list && str) {
		for (char **ptr=list; *ptr!=NULL; ptr++) {
			if (strcmp(*ptr, str)==0)
				return true;
		}
	}
	return false;
}

int CNewFilesDlg::ExecuteProc(CProgressDlg *sender, void *userdata)
{
	return ((CNewFilesDlg*)userdata)->ProcessFiles(sender);
}

int CNewFilesDlg::ProcessFiles(CProgressDlg *sender)
{
	bool do_konv, do_tcor, do_bias, do_dark, do_flat, do_phot, do_match;
	int res = 0, state, ref_id, file_res;
	char *fpath, *fbase, *tpath;
	const gchar *ref_file;
	CProject::tStatus status;
	GtkTreePath *path;
	CConvertProc konv;
	CTimeCorrProc tcor;
	CBiasCorrProc bias;
	CDarkCorrProc dark;
	CFlatCorrProc flat;
	CPhotometryProc phot;
	CMatchingProc match;

	m_InFiles = m_OutFiles = 0;

	// Initialization
	Global.pProject->GetStatus(&status);
	do_konv = status.converted>0;
	do_tcor = Global.pProject->GetTimeCorrection()!=0;
	do_bias = Global.pProject->GetOrigBiasFile()!=NULL;
	do_dark = Global.pProject->GetOrigDarkFile()!=NULL;
	do_flat = Global.pProject->GetTempFlatFile()!=NULL;
	do_phot = status.photometred>0;
	do_match = Global.pProject->GetReference(&ref_id, &ref_file);
	if (do_konv) 
		res = konv.Init(sender);
	if (res==0 && do_tcor)
		res = tcor.Init(sender, Global.pProject->GetTimeCorrection());
	if (res==0 && do_bias) 
		res = bias.Init(sender, Global.pProject->GetOrigBiasFile());
	if (res==0 && do_dark) 
		res = dark.Init(sender, Global.pProject->GetOrigDarkFile());
	if (res==0 && do_flat) 
		res = flat.Init(sender, Global.pProject->GetOrigFlatFile());
	if (res==0 && do_phot) 
		res = phot.Init(sender);
	if (res==0 && do_match)
		res = match.Init(sender, ref_id, ref_file);

	// Process files
	if (res==0) {
		for (GList *node=m_FileList; node!=NULL; node = node->next) {
			path = gtk_tree_row_reference_get_path((GtkTreeRowReference*)node->data);
			state = Global.pProject->GetState(path);
			if ((state & CFILE_CONVERSION)==0) {
				fpath = Global.pProject->GetSourceFile(path);
				fbase = g_filename_display_basename(fpath);
				sender->SetFileName(fbase);
				g_free(fbase);
				g_free(fpath);
			} else {
				tpath = Global.pProject->GetImageFileName(path);
				sender->SetFileName(tpath);
				g_free(tpath);
			}
			sender->SetProgress(m_InFiles++);
			file_res = 0;
			if (file_res==0 && do_konv && (state & CFILE_CONVERSION)==0)
				file_res = konv.Execute(path);
			if (file_res==0 && do_tcor && (state & CFILE_TIMECORR | CFILE_PHOTOMETRY)==0)
				file_res = tcor.Execute(path);
			if (file_res==0 && do_bias && (state & (CFILE_BIASCORR | CFILE_DARKCORR | CFILE_FLATCORR | CFILE_PHOTOMETRY))==0) 
				file_res = bias.Execute(path);
			if (file_res==0 && do_dark && (state & (CFILE_DARKCORR | CFILE_FLATCORR | CFILE_PHOTOMETRY))==0) 
				file_res = dark.Execute(path);
			if (file_res==0 && do_flat && (state & (CFILE_FLATCORR | CFILE_PHOTOMETRY))==0) 
				file_res = flat.Execute(path);
			if (file_res==0 && do_phot && (state & CFILE_PHOTOMETRY)==0)
				file_res = phot.Execute(path);
			if (file_res==0 && do_match && (state & CFILE_MATCHING)==0)
				file_res = match.Execute(path);
			if (file_res==0)
				m_OutFiles++;
			gtk_tree_path_free(path);
			if (sender->Cancelled()) 
				break;
		}
	}

	// Finished
	return res;
}
