/**************************************************************

selection.cpp (C-Munipack project)
Table of selected stars
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <glib/gstdio.h>

#include "proc_classes.h"
#include "configuration.h"
#include "progress_dlg.h"
#include "utils.h"
#include "main.h"

//--------------------------   CONVERSION CONTEXT   --------------------------

//
// Constructor
//
CConvertProc::CConvertProc(void):m_Progress(NULL)
{
	m_Konv = cmpack_konv_init();
}


//
// Destructor
//
CConvertProc::~CConvertProc()
{
	cmpack_unref(m_Konv);
}


//
// Initialization
//
int CConvertProc::Init(CProgressDlg *pProgress)
{
	m_Progress = pProgress;

	// Check directory
	force_directory(get_temp_dir());

	// Data format
	CmpackBitpix bitpix = (CmpackBitpix)CConfig::GetInt("Conversion", "Format", CMPACK_BITPIX_AUTO);
	cmpack_konv_set_bitpix(m_Konv, bitpix);

	// Frame border
	CmpackBorder border;
	cmpack_border_set(&border, CConfig::GetInt("CCD", "BorderLeft", 0),
		CConfig::GetInt("CCD", "BorderTop", 0),
		CConfig::GetInt("CCD", "BorderRight", 0),
		CConfig::GetInt("CCD", "BorderBottom", 0));
	cmpack_konv_set_border(m_Konv, &border);

	// Output handling
	if (pProgress) {
		CmpackConsole *con = cmpack_con_init_cb(CProgressDlg::OutputProc, pProgress);
		if (CConfig::GetBool("Environment", "Debug"))
			cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
		cmpack_konv_set_console(m_Konv, con);
		cmpack_unref(con);
	}

	return 0;
}


//
// Convert file
//
int CConvertProc::Execute(GtkTreePath *pPath)
{
	int res, refid, fileid, state, frames;
	char *f, temp[128], *filter, *msg, *fpath, *tpath;
	double jd, exptime, ccdtemp;
	const char *aux;
	GtkTreeModel *pList = Global.pProject->FileList();
	GtkTreeIter iter;
	CmpackCcdFile *infile, *outfile;
	CmpackDateTime dt;

	gtk_tree_model_get_iter(pList, &iter, pPath);

	// Get frame data
	gtk_tree_model_get(pList, &iter, FRAME_ID, &fileid, FRAME_ORIGFILE, &fpath, -1);
	state = 0;

	// Print source file name
	m_Progress->Print(fpath);
	
	// Clear internal frame data
	gtk_list_store_set(GTK_LIST_STORE(pList), &iter, FRAME_STATE, 0, 
		FRAME_JULDAT, 0.0, FRAME_TEMPFILE, NULL, FRAME_PHOTFILE, NULL, 
		FRAME_REPORT, NULL, FRAME_BIASFILE, NULL, FRAME_DARKFILE, NULL, 
		FRAME_FLATFILE, NULL, FRAME_FILTER, NULL, FRAME_TIMECORR, 0.0, 
		FRAME_EXPTIME, 0.0, FRAME_SUBFRAMES, 0, FRAME_CCDTEMP, -999.9, 
		FRAME_STARS, 0, FRAME_MSTARS, 0, -1);

	// Clear reference frame data (when converting reference frame)
	if (Global.pProject->GetReference(&refid, NULL)) {
		if (refid == fileid)
			Global.pProject->ClearReference();
	}

	// Open source file
	f = g_locale_from_utf8(fpath, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&infile, f, CMPACK_OPEN_READONLY);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg, CFILE_CONVERSION | CFILE_PHOTOMETRY | CFILE_MATCHING, 0);
		cmpack_free(msg);
		g_free(fpath);
		return res;
	}
	g_free(fpath);

	// Open target file
	sprintf(temp, "tmp%05d.fts", fileid);
	tpath = g_build_filename(get_temp_dir(), temp, NULL);
	f = g_locale_from_utf8(tpath, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&outfile, f, CMPACK_OPEN_CREATE);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg, CFILE_CONVERSION | CFILE_PHOTOMETRY | CFILE_MATCHING, 0);
		cmpack_free(msg);
		g_free(tpath);
		cmpack_unref(infile);
		return res;
	}
	g_free(tpath);

	// Conversion
	res = cmpack_konv(m_Konv, infile, outfile);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg, CFILE_CONVERSION | CFILE_PHOTOMETRY | CFILE_MATCHING, 0);
		cmpack_free(msg);
		cmpack_unref(infile);
		cmpack_unref(outfile);
		return res;
	}
	cmpack_unref(infile);

	// OK
	jd = exptime = 0.0;
	ccdtemp = -999.9;
	filter = NULL;
	frames = 1;
	Global.pProject->SetResult(&iter, "Conversion OK", CFILE_CONVERSION | CFILE_PHOTOMETRY | CFILE_MATCHING, CFILE_CONVERSION);
	if (cmpack_ccd_get_datetime(outfile, &dt)==0)
		jd = cmpack_encodejd(&dt);
	cmpack_ccd_get_exptime(outfile, &exptime);
	cmpack_ccd_get_ccdtemp(outfile, &ccdtemp);
	cmpack_ccd_get_frames(outfile, &frames);
	aux = cmpack_ccd_get_filter(outfile);
	if (aux)
		filter = g_locale_to_utf8(aux, -1, NULL, NULL, NULL);
	if (jd>0 && exptime>0)
		jd += 0.5*exptime/86400.0;
	gtk_list_store_set(GTK_LIST_STORE(pList), &iter, 
		FRAME_TEMPFILE, temp, FRAME_JULDAT, jd, FRAME_FILTER, filter, 
		FRAME_EXPTIME, exptime, FRAME_CCDTEMP, ccdtemp, FRAME_SUBFRAMES, frames, -1);
	g_free(filter);

	// Print short info
	CmpackConsole *con = cmpack_con_init_cb(CProgressDlg::OutputProc, m_Progress);
	if (CConfig::GetBool("Environment", "Debug"))
		cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
	cmpack_ccd_print_info(outfile, con);
	cmpack_unref(con);

	cmpack_unref(outfile);
	return 0;
}

//-----------------------   BIAS CORRECTION CONTEXT   ----------------------

//
// Constructor
//
CBiasCorrProc::CBiasCorrProc():m_Progress(NULL)
{
	m_Bias = cmpack_bias_init();
}


//
// Destructor
//
CBiasCorrProc::~CBiasCorrProc()
{
	cmpack_unref(m_Bias);
}


//
// Initialization
//
int CBiasCorrProc::Init(CProgressDlg *pProgress, const gchar *fbias)
{
	int res = 0;
	char *temp, msg[512];
	const char *tpath;
	CmpackCcdFile *bias;

	m_Progress = pProgress;

	if (!fbias || strlen(fbias)==0 || !g_file_test(fbias, G_FILE_TEST_IS_REGULAR))
		return CMPACK_ERR_CANT_OPEN_BIAS;

	// Check directory
	force_directory(get_temp_dir());

	sprintf(msg, "Bias frame: %s", fbias);
	m_Progress->Print(msg);

	// Bias frame
	if (!Global.pProject->CheckBiasFile(fbias)) {
		temp = g_build_filename(get_temp_dir(), "bias.fts", NULL);
		res = ConvertBiasFrame(pProgress, fbias, temp);
		g_free(temp);
		if (res!=0)
			return res;
	}
	tpath = Global.pProject->GetTempBiasFile();
	if (tpath) {
		gchar *f = g_locale_from_utf8(tpath, -1, NULL, NULL, NULL);
		res = cmpack_ccd_open(&bias, f, CMPACK_OPEN_READONLY);
		if (res==0) {
			res = cmpack_bias_rbias(m_Bias, bias);
			cmpack_unref(bias);
		}
		g_free(f);
		if (res!=0)
			return res;
	}
	
	// Frame border
	CmpackBorder border;
	cmpack_border_set(&border, CConfig::GetInt("CCD", "BorderLeft", 0),
		CConfig::GetInt("CCD", "BorderTop", 0),
		CConfig::GetInt("CCD", "BorderRight", 0),
		CConfig::GetInt("CCD", "BorderBottom", 0));
	cmpack_bias_set_border(m_Bias, &border);

	// Output handling
	if (pProgress) {
		CmpackConsole *con = cmpack_con_init_cb(CProgressDlg::OutputProc, pProgress);
		if (CConfig::GetBool("Environment", "Debug"))
			cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
		cmpack_bias_set_console(m_Bias, con);
		cmpack_unref(con);
	}

	return res;
}


//
// Convert bias frame
//
int CBiasCorrProc::ConvertBiasFrame(CProgressDlg *pProgress, const gchar *orig_file, const gchar *temp_file)
{
	int res;
	char *msg;
	gchar *f;
	CmpackKonv *konv = cmpack_konv_init();
	CmpackCcdFile *infile, *outfile;

	// Delete temporary file
	g_unlink(temp_file);

	// Frame border
	CmpackBorder border;
	cmpack_border_set(&border, CConfig::GetInt("CCD", "BorderLeft", 0),
		CConfig::GetInt("CCD", "BorderTop", 0),
		CConfig::GetInt("CCD", "BorderRight", 0),
		CConfig::GetInt("CCD", "BorderBottom", 0));
	cmpack_konv_set_border(konv, &border);

	// Output handling
	if (pProgress) {
		CmpackConsole *con = cmpack_con_init_cb(CProgressDlg::OutputProc, pProgress);
		if (CConfig::GetBool("Environment", "Debug"))
			cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
		cmpack_konv_set_console(konv, con);
		cmpack_unref(con);
	}

	// Data format
	CmpackBitpix bitpix = (CmpackBitpix)CConfig::GetInt("Conversion", "Format", CMPACK_BITPIX_AUTO);
	cmpack_konv_set_bitpix(konv, bitpix);

	// Open source file
	f = g_locale_from_utf8(orig_file, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&infile, f, CMPACK_OPEN_READONLY);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		cmpack_unref(konv);
		return res;
	}

	// Open target file
	f = g_locale_from_utf8(temp_file, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&outfile, f, CMPACK_OPEN_CREATE);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		cmpack_unref(infile);
		cmpack_unref(konv);
		return res;
	}

	// Conversion
	res = cmpack_konv(konv, infile, outfile);
	cmpack_unref(infile);
	cmpack_unref(outfile);
	cmpack_unref(konv);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		return res;
	}

	// OK
	Global.pProject->SetTempBiasFile(temp_file);
	return 0;
}


//
// Bias correction
//
int CBiasCorrProc::Execute(GtkTreePath *pPath)
{
	int res, state;
	char *msg, *tpath, *f;
	GtkTreeModel *pList = Global.pProject->FileList();
	GtkTreeIter iter;
	CmpackCcdFile *infile;

	gtk_tree_model_get_iter(pList, &iter, pPath);

	// Check frame status
	state = Global.pProject->GetState(&iter);
	if ((state & CFILE_CONVERSION)==0) {
		msg = "Image hasn't been converted.";
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg);
		return -1;
	}
	if ((state & CFILE_BIASCORR)!=0) {
		msg = "Image has been bias-frame corrected already.";
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		return -1;
	}
	if ((state & CFILE_DARKCORR)!=0) {
		msg = "You cannot do bias-frame correction after dark-frame correction.";
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		return -1;
	}
	if ((state & CFILE_FLATCORR)!=0) {
		msg = "You cannot do bias-frame correction after flat-frame correction.";
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		return -1;
	}
	if ((state & CFILE_PHOTOMETRY)!=0) {
		msg = "You cannot do time correction after photometry.";
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		return -1;
	}

	// Open source file 
	tpath = Global.pProject->GetImageFile(&iter);
	f = g_locale_from_utf8(tpath, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&infile, f, CMPACK_OPEN_READWRITE);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg);
		cmpack_free(msg);
		g_free(tpath);
		return res;
	}
	g_free(tpath);

	// Bias correction
	res = cmpack_bias(m_Bias, infile);
	if (res!=0) {
		// Failed
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg);
		cmpack_free(msg);
		cmpack_unref(infile);
		return res;
	}
	cmpack_unref(infile);

	// Success
	Global.pProject->SetResult(&iter, "Bias-frame correction OK", CFILE_BIASCORR, CFILE_BIASCORR);
	return 0;
}

//-----------------------   TIME CORRECTION CONTEXT   ----------------------

//
// Constructor
//
CTimeCorrProc::CTimeCorrProc():m_Progress(NULL)
{
	m_Time = cmpack_tcorr_init();
	m_Seconds = 0;
}


//
// Destructor
//
CTimeCorrProc::~CTimeCorrProc()
{
	cmpack_unref(m_Time);
}


//
// Initialization
//
int CTimeCorrProc::Init(CProgressDlg *pProgress, double seconds)
{
	char msg[256];

	m_Progress = pProgress;

	// Set parameters
	m_Seconds = seconds;
	cmpack_tcorr_set_corr(m_Time, m_Seconds);

	// Check directory
	force_directory(get_temp_dir());

	// Print correction
	sprintf(msg, "Time correction: %.3f seconds", m_Seconds);
	m_Progress->Print(msg);

	// Output handling
	if (pProgress) {
		CmpackConsole *con = cmpack_con_init_cb(CProgressDlg::OutputProc, pProgress);
		if (CConfig::GetBool("Environment", "Debug"))
			cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
		cmpack_tcorr_set_console(m_Time, con);
		cmpack_unref(con);
	}
	return 0;
}


//
// Time correction
//
int CTimeCorrProc::Execute(GtkTreePath *pPath)
{
	double jd, tc, exptime;
	int res, state;
	char *msg, *tpath, *f, buf[256];
	GtkTreeModel *pList = Global.pProject->FileList();
	GtkTreeIter iter;
	CmpackCcdFile *infile;
	CmpackDateTime dt;

	gtk_tree_model_get_iter(pList, &iter, pPath);

	// Check frame status
	state = Global.pProject->GetState(&iter);
	if ((state & CFILE_CONVERSION)==0) {
		msg = "Image hasn't been converted.";
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg);
		return -1;
	}
	if ((state & CFILE_PHOTOMETRY)!=0) {
		msg = "You cannot do time correction after photometry.";
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		return -1;
	}

	// Open source file 
	tpath = Global.pProject->GetImageFile(&iter);
	f = g_locale_from_utf8(tpath, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&infile, f, CMPACK_OPEN_READWRITE);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg);
		cmpack_free(msg);
		g_free(tpath);
		return res;
	}
	g_free(tpath);

	// Time correction
	res = cmpack_tcorr(m_Time, infile);
	if (res!=0) {
		// Failed
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg);
		cmpack_free(msg);
		cmpack_unref(infile);
		return res;
	}

	// Update frame parameters
	jd = exptime = 0;
	cmpack_ccd_get_datetime(infile, &dt);
	jd = cmpack_encodejd(&dt);
	cmpack_ccd_get_exptime(infile, &exptime);
	if (jd>0 && exptime>0)
		jd += 0.5*exptime/86400.0;
	gtk_tree_model_get(pList, &iter, FRAME_TIMECORR, &tc, -1);
	tc += m_Seconds;
	gtk_list_store_set(GTK_LIST_STORE(pList), &iter, FRAME_JULDAT, jd, 
		FRAME_TIMECORR, tc, -1);
	cmpack_unref(infile);
	sprintf(buf, "New JD: %.7f", jd);
	m_Progress->Print(buf);

	// Success
	Global.pProject->SetResult(&iter, "Time correction OK", CFILE_TIMECORR, CFILE_TIMECORR);
	return 0;
}

//-----------------------   DARK CORRECTION CONTEXT   ----------------------

//
// Constructor
//
CDarkCorrProc::CDarkCorrProc():m_Progress(NULL)
{
	m_Dark = cmpack_dark_init();
}


//
// Destructor
//
CDarkCorrProc::~CDarkCorrProc()
{
	cmpack_unref(m_Dark);
}


//
// Initialization
//
int CDarkCorrProc::Init(CProgressDlg *pProgress, const gchar *fdark)
{
	int res = 0;
	const char *tpath;
	char *temp, msg[512];
	CmpackCcdFile *dark;

	m_Progress = pProgress;

	if (!fdark || strlen(fdark)==0 || !g_file_test(fdark, G_FILE_TEST_IS_REGULAR))
		return CMPACK_ERR_CANT_OPEN_DARK;

	// Check directory
	force_directory(get_temp_dir());

	sprintf(msg, "Dark frame: %s", fdark);
	m_Progress->Print(msg);

	// Dark frame
	if (!Global.pProject->CheckDarkFile(fdark)) {
		temp = g_build_filename(get_temp_dir(), "dark.fts", NULL);
		res = ConvertDarkFrame(pProgress, fdark, temp);
		g_free(temp);
		if (res!=0)
			return res;
	}
	tpath = Global.pProject->GetTempDarkFile();
	if (tpath) {
		gchar *f = g_locale_from_utf8(tpath, -1, NULL, NULL, NULL);
		res = cmpack_ccd_open(&dark, f, CMPACK_OPEN_READONLY);
		if (res==0) {
			res = cmpack_dark_rdark(m_Dark, dark);
			cmpack_unref(dark);
		}
		g_free(f);
		if (res!=0)
			return res;
	}
	
	// Frame border
	CmpackBorder border;
	cmpack_border_set(&border, CConfig::GetInt("CCD", "BorderLeft", 0),
		CConfig::GetInt("CCD", "BorderTop", 0),
		CConfig::GetInt("CCD", "BorderRight", 0),
		CConfig::GetInt("CCD", "BorderBottom", 0));
	cmpack_dark_set_border(m_Dark, &border);

	// Enable scaling
	bool scaling = CConfig::GetBool("Calibration", "Advanced");
	cmpack_dark_set_scaling(m_Dark, scaling);

	// Output handling
	if (pProgress) {
		CmpackConsole *con = cmpack_con_init_cb(CProgressDlg::OutputProc, pProgress);
		if (CConfig::GetBool("Environment", "Debug"))
			cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
		cmpack_dark_set_console(m_Dark, con);
		cmpack_unref(con);
	}

	return res;
}


//
// Convert dark frame
//
int CDarkCorrProc::ConvertDarkFrame(CProgressDlg *pProgress, const gchar *orig_file, const gchar *temp_file)
{
	int res;
	char *msg;
	gchar *f;
	CmpackKonv *konv = cmpack_konv_init();
	CmpackCcdFile *infile, *outfile;

	// Delete temporary file
	g_unlink(temp_file);

	// Frame border
	CmpackBorder border;
	cmpack_border_set(&border, CConfig::GetInt("CCD", "BorderLeft", 0),
		CConfig::GetInt("CCD", "BorderTop", 0),
		CConfig::GetInt("CCD", "BorderRight", 0),
		CConfig::GetInt("CCD", "BorderBottom", 0));
	cmpack_konv_set_border(konv, &border);

	// Output handling
	if (pProgress) {
		CmpackConsole *con = cmpack_con_init_cb(CProgressDlg::OutputProc, pProgress);
		if (CConfig::GetBool("Environment", "Debug"))
			cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
		cmpack_konv_set_console(konv, con);
		cmpack_unref(con);
	}

	// Data format
	CmpackBitpix bitpix = (CmpackBitpix)CConfig::GetInt("Conversion", "Format", CMPACK_BITPIX_AUTO);
	cmpack_konv_set_bitpix(konv, bitpix);

	// Open source file
	f = g_locale_from_utf8(orig_file, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&infile, f, CMPACK_OPEN_READONLY);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		cmpack_unref(konv);
		return res;
	}

	// Open target file
	f = g_locale_from_utf8(temp_file, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&outfile, f, CMPACK_OPEN_CREATE);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		cmpack_unref(infile);
		cmpack_unref(konv);
		return res;
	}

	// Conversion
	res = cmpack_konv(konv, infile, outfile);
	cmpack_unref(infile);
	cmpack_unref(outfile);
	cmpack_unref(konv);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		return res;
	}

	Global.pProject->SetTempDarkFile(temp_file);
	return 0;
}


//
// Dark correction
//
int CDarkCorrProc::Execute(GtkTreePath *pPath)
{
	int res, state;
	char *msg, *tpath, *f;
	GtkTreeModel *pList = Global.pProject->FileList();
	GtkTreeIter iter;
	CmpackCcdFile *infile;

	gtk_tree_model_get_iter(pList, &iter, pPath);

	// Check frame status
	state = Global.pProject->GetState(&iter);
	if ((state & CFILE_CONVERSION)==0) {
		msg = "Image hasn't been converted.";
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		return -1;
	}
	if ((state & CFILE_DARKCORR)!=0) {
		msg = "Image has been dark-frame corrected already.";
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		return -1;
	}
	if ((state & CFILE_FLATCORR)!=0) {
		msg = "You cannot do dark-frame correction after flat-frame correction.";
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		return -1;
	}
	if ((state & CFILE_PHOTOMETRY)!=0) {
		msg = "You cannot do time correction after photometry.";
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		return -1;
	}

	// Open source file 
	tpath = Global.pProject->GetImageFile(&iter);
	f = g_locale_from_utf8(tpath, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&infile, f, CMPACK_OPEN_READWRITE);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		cmpack_free(msg);
		g_free(tpath);
		return res;
	}
	g_free(tpath);

	// Dark correction
	res = cmpack_dark(m_Dark, infile);
	if (res!=0) {
		// Failed
		msg = cmpack_formaterror(res);
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		cmpack_free(msg);
		cmpack_unref(infile);
		return res;
	}
	cmpack_unref(infile);

	// Success
	Global.pProject->SetResult(&iter, "Dark-frame correction OK", CFILE_DARKCORR, CFILE_DARKCORR);
	return 0;
}

//-----------------------   FLAT CORRECTION CONTEXT   ----------------------

//
// Constructor
//
CFlatCorrProc::CFlatCorrProc():m_Progress(NULL)
{
	m_Flat = cmpack_flat_init();
}


//
// Destructor
//
CFlatCorrProc::~CFlatCorrProc()
{
	cmpack_unref(m_Flat);
}


//
// Initialization
//
int CFlatCorrProc::Init(CProgressDlg *pProgress, const gchar *fflat)
{
	int res = 0;
	const char *tpath;
	char *temp, msg[512];
	CmpackCcdFile *flat;

	m_Progress = pProgress;

	if (!fflat || strlen(fflat)==0 || !g_file_test(fflat, G_FILE_TEST_IS_REGULAR))
		return CMPACK_ERR_CANT_OPEN_FLAT;

	// Check directory
	force_directory(get_temp_dir());

	sprintf(msg, "Flat frame: %s", fflat);
	m_Progress->Print(msg);

	// Flat frame
	if (!Global.pProject->CheckFlatFile(fflat)) {
		temp = g_build_filename(get_temp_dir(), "flat.fts", NULL);
		res = ConvertFlatFrame(pProgress, fflat, temp);
		g_free(temp);
		if (res!=0)
			return res;
	}
	tpath = Global.pProject->GetTempFlatFile();
	if (tpath) {
		gchar *f = g_locale_from_utf8(tpath, -1, NULL, NULL, NULL);
		res = cmpack_ccd_open(&flat, f, CMPACK_OPEN_READONLY);
		if (res==0) {
			res = cmpack_flat_rflat(m_Flat, flat);
			cmpack_unref(flat);
		}
		g_free(f);
		if (res!=0)
			return res;
	}

	// Frame border
	CmpackBorder border;
	cmpack_border_set(&border, CConfig::GetInt("CCD", "BorderLeft", 0),
		CConfig::GetInt("CCD", "BorderTop", 0),
		CConfig::GetInt("CCD", "BorderRight", 0),
		CConfig::GetInt("CCD", "BorderBottom", 0));
	cmpack_flat_set_border(m_Flat, &border);

	// Output handling
	if (pProgress) {
		CmpackConsole *con = cmpack_con_init_cb(CProgressDlg::OutputProc, pProgress);
		if (CConfig::GetBool("Environment", "Debug"))
			cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
		cmpack_flat_set_console(m_Flat, con);
		cmpack_unref(con);
	}

	return res;
}

//
// Convert dark frame
//
int CFlatCorrProc::ConvertFlatFrame(CProgressDlg *pProgress, const gchar *orig_file, const gchar *temp_file)
{
	int res;
	gchar *f;
	char *msg;
	CmpackKonv *konv = cmpack_konv_init();
	CmpackCcdFile *infile, *outfile;

	// Delete temporary file
	g_unlink(temp_file);

	// Frame border
	CmpackBorder border;
	cmpack_border_set(&border, CConfig::GetInt("CCD", "BorderLeft", 0),
		CConfig::GetInt("CCD", "BorderTop", 0),
		CConfig::GetInt("CCD", "BorderRight", 0),
		CConfig::GetInt("CCD", "BorderBottom", 0));
	cmpack_konv_set_border(konv, &border);

	// Output handling
	if (pProgress) {
		CmpackConsole *con = cmpack_con_init_cb(CProgressDlg::OutputProc, pProgress);
		if (CConfig::GetBool("Environment", "Debug"))
			cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
		cmpack_konv_set_console(konv, con);
		cmpack_unref(con);
	}

	// Data format
	CmpackBitpix bitpix = (CmpackBitpix)CConfig::GetInt("Conversion", "Format", CMPACK_BITPIX_AUTO);
	cmpack_konv_set_bitpix(konv, bitpix);
	// Open source file
	f = g_locale_from_utf8(orig_file, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&infile, f, CMPACK_OPEN_READONLY);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		cmpack_unref(konv);
		return res;
	}

	// Open target file
	f = g_locale_from_utf8(temp_file, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&outfile, f, CMPACK_OPEN_CREATE);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		cmpack_unref(infile);
		cmpack_unref(konv);
		return res;
	}

	// Conversion
	res = cmpack_konv(konv, infile, outfile);
	cmpack_unref(infile);
	cmpack_unref(outfile);
	cmpack_unref(konv);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		return res;
	}

	Global.pProject->SetTempFlatFile(temp_file);
	return 0;
}


//
// Flat correction
//
int CFlatCorrProc::Execute(GtkTreePath *pPath)
{
	int res, state;
	char *msg, *tpath, *f;
	GtkTreeModel *pList = Global.pProject->FileList();
	GtkTreeIter iter;
	CmpackCcdFile *infile;

	gtk_tree_model_get_iter(pList, &iter, pPath);

	// Check frame status
	state = Global.pProject->GetState(&iter);
	if ((state & CFILE_CONVERSION)==0) {
		msg = "Image hasn't been converted.";
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		return -1;
	}
	if ((state & CFILE_FLATCORR)!=0) {
		msg = "Image has been flat-frame corrected already.";
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		return -1;
	}
	if ((state & CFILE_PHOTOMETRY)!=0) {
		msg = "You cannot do time correction after photometry.";
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		return -1;
	}

	// Open source file 
	tpath = Global.pProject->GetImageFile(&iter);
	f = g_locale_from_utf8(tpath, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&infile, f, CMPACK_OPEN_READWRITE);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		cmpack_free(msg);
		g_free(tpath);
		return res;
	}
	g_free(tpath);

	// Dark correction
	res = cmpack_flat(m_Flat, infile);
	if (res!=0) {
		// Failed
		msg = cmpack_formaterror(res);
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		cmpack_free(msg);
		cmpack_unref(infile);
		return res;
	}
	cmpack_unref(infile);

	// Success
	Global.pProject->SetResult(&iter, "Flat-frame correction OK", CFILE_FLATCORR, CFILE_FLATCORR);
	return 0;
}

//-----------------------   PHOTOMETRY CONTEXT   ----------------------

//
// Constructor
//
CPhotometryProc::CPhotometryProc():m_Progress(NULL)
{
	m_Phot = cmpack_phot_init();
}


//
// Destructor
//
CPhotometryProc::~CPhotometryProc()
{
	cmpack_unref(m_Phot);
}


//
// Initialization
//
int CPhotometryProc::Init(CProgressDlg *pProgress)
{
	int i;
	double aper[MAX_APERTURES];

	m_Progress = pProgress;

	// Check directory
	force_directory(get_temp_dir());

	// Photometry parameters
	cmpack_phot_set_rnoise(m_Phot, CConfig::GetDbl("Photometry", "ReadNoise", 15.0));
	cmpack_phot_set_adcgain(m_Phot, CConfig::GetDbl("Photometry", "ADCGain", 2.3));
	cmpack_phot_set_minval(m_Phot, CConfig::GetDbl("CCD", "MinValue", 0.0));
	cmpack_phot_set_maxval(m_Phot, CConfig::GetDbl("CCD", "MaxValue", 65535.0));
	cmpack_phot_set_fwhm(m_Phot, CConfig::GetDbl("Photometry", "Fwhm", 3.0));
	cmpack_phot_set_thresh(m_Phot, CConfig::GetDbl("Photometry", "Threshold", 4.0));
	cmpack_phot_set_minrnd(m_Phot, CConfig::GetDbl("Photometry", "MinRoundness", -1.0));
	cmpack_phot_set_maxrnd(m_Phot, CConfig::GetDbl("Photometry", "MaxRoundness", 1.0));
	cmpack_phot_set_minshrp(m_Phot, CConfig::GetDbl("Photometry", "MinSharpness", 0.2));
	cmpack_phot_set_maxshrp(m_Phot, CConfig::GetDbl("Photometry", "MaxSharpness", 1.0));
	cmpack_phot_set_skyin(m_Phot, CConfig::GetDbl("Photometry", "SkyInRadius", 20.0));
	cmpack_phot_set_skyout(m_Phot, CConfig::GetDbl("Photometry", "SkyOutRadius", 30.0));

	// Frame Border
	CmpackBorder border;
	cmpack_border_set(&border, CConfig::GetInt("CCD", "BorderLeft", 0),
		CConfig::GetInt("CCD", "BorderTop", 0),
		CConfig::GetInt("CCD", "BorderRight", 0),
		CConfig::GetInt("CCD", "BorderBottom", 0));
	cmpack_phot_set_border(m_Phot, &border);

	// Apertures
	for (i=0; i<MAX_APERTURES; i++) 
		aper[i] = GetApertureRadius(i+1);
	cmpack_phot_set_aper(m_Phot, aper, MAX_APERTURES);

	// Output handling
	if (pProgress) {
		CmpackConsole *con = cmpack_con_init_cb(CProgressDlg::OutputProc, pProgress);
		if (CConfig::GetBool("Environment", "Debug"))
			cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
		cmpack_phot_set_console(m_Phot, con);
		cmpack_unref(con);
	}

	return 0;
}


//
// Photometry
//
int CPhotometryProc::Execute(GtkTreePath *pPath)
{
	int res, state, fileid, refid, nstars;
	char *msg, *tpath, *f, pht[128], txt[128], *ppath;
	GtkTreeModel *pList = Global.pProject->FileList();
	GtkTreeIter iter;
	CmpackCcdFile *infile;
	CmpackPhtFile *outfile;

	gtk_tree_model_get_iter(pList, &iter, pPath);

	// Check frame status
	state = Global.pProject->GetState(&iter);
	if ((state & CFILE_CONVERSION)==0) {
		msg = "Image hasn't been converted.";
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		return -1;
	}

	// Clear reference frame data (when processing reference frame)
	fileid = Global.pProject->GetFrameID(&iter);
	if (Global.pProject->GetReference(&refid, NULL)) {
		if (refid == fileid)
			Global.pProject->ClearReference();
	}

	// Open source file 
	tpath = Global.pProject->GetImageFile(&iter);
	f = g_locale_from_utf8(tpath, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&infile, f, CMPACK_OPEN_READONLY);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg);
		cmpack_free(msg);
		g_free(tpath);
		return res;
	}

	// Delete target file
	sprintf(pht, "tmp%05d.pht", fileid);
	ppath = g_build_filename(get_temp_dir(), pht, NULL);
	f = g_locale_from_utf8(ppath, -1, NULL, NULL, NULL);
	remove(f);
	g_free(f);

	// Make photometry
	outfile = cmpack_pht_init();
	res = cmpack_phot(m_Phot, infile, outfile, &nstars);
	if (res==0) {
		// Save the result to a file
		f = g_locale_from_utf8(ppath, -1, NULL, NULL, NULL);
		res = cmpack_pht_save(outfile, f);
		g_free(f);
	}
	if (res==0) {
		// Success
		Global.pProject->SetResult(&iter, "Photometry OK", CFILE_PHOTOMETRY | CFILE_MATCHING, CFILE_PHOTOMETRY);
		gtk_list_store_set(GTK_LIST_STORE(pList), &iter, 
			FRAME_PHOTFILE, pht, FRAME_STARS, nstars, FRAME_MSTARS, 0, -1);
		sprintf(txt, "%d stars found.", nstars);
		m_Progress->Print(txt);
	} else {
		// Failed
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg, CFILE_PHOTOMETRY | CFILE_MATCHING, 0);
		gtk_list_store_set(GTK_LIST_STORE(pList), &iter,
			FRAME_PHOTFILE, NULL, FRAME_STARS, 0, FRAME_MSTARS, 0, -1);
		cmpack_free(msg);
	}

	g_free(tpath);
	g_free(ppath);
	cmpack_unref(infile);
	cmpack_unref(outfile);
	return res;
}


//-----------------------   MATCHING CONTEXT   ----------------------

//
// Constructor
//
CMatchingProc::CMatchingProc():m_Progress(NULL)
{
	m_Match = cmpack_match_init();
}


//
// Destructor
//
CMatchingProc::~CMatchingProc()
{
	cmpack_unref(m_Match);
}


//
// Initialization
//
int CMatchingProc::Init(CProgressDlg *pProgress, int frame_id, const char *cat_file)
{
	int res = 0;
	char *tmp_file = NULL, *f, *tpath, *fpath, *msg, txt[512];
	CmpackPhtFile *pht;
	CmpackCatFile *cat;
	CmpackCcdFile *ccd;

	m_Progress = pProgress;

	m_Selection.Clear();
	m_Object.Clear();
	m_Location.Clear();
	force_directory(get_temp_dir());

	if (frame_id>0) {
		sprintf(txt, "Reference frame: #%d", frame_id);
		m_Progress->Print(txt);
		// Read reference frame
		tpath = Global.pProject->GetPhotFile(frame_id);
		if (tpath) {
			f = g_locale_from_utf8(tpath, -1, NULL, NULL, NULL);
			res = cmpack_pht_load(&pht, f, CMPACK_LOAD_DEFAULT);
			g_free(f);
			if (res==0) {
				res = cmpack_match_readref_pht(m_Match, pht);
				cmpack_unref(pht);
			}
			g_free(tpath);
		} else {
			res = CMPACK_ERR_REF_NOT_FOUND;
		}
		// Object's and observer's coordinates
		fpath = Global.pProject->GetImageFile(frame_id);
		if (fpath) {
			f = g_locale_from_utf8(fpath, -1, NULL, NULL, NULL);
			res = cmpack_ccd_open(&ccd, f, CMPACK_OPEN_HEADONLY);
			g_free(f);
			if (res==0) {
				m_Object.Load(ccd);
				m_Location.Load(ccd);
				cmpack_unref(ccd);
			}
			g_free(fpath);
		}
	} else if (cat_file) {
		sprintf(txt, "Catalogue file: %s", cat_file);
		m_Progress->Print(txt);
		if (g_file_test(cat_file, G_FILE_TEST_IS_REGULAR)) {
			// Make copy of catalogue file
			tmp_file = g_build_filename(get_temp_dir(), "reference.xml", NULL);
			if (!Global.pProject->CheckCatFile(cat_file)) 
				ConvertCatFile(cat_file, tmp_file);
			// Read catalogue file
			f = g_locale_from_utf8(tmp_file, -1, NULL, NULL, NULL);
			res = cmpack_cat_load(&cat, f, CMPACK_LOAD_DEFAULT);
			g_free(f);
			if (res==0) {
				res = cmpack_match_readref_cat(m_Match, cat);
				m_Selection.Load(cat);
				m_Object.Load(cat);
				cmpack_unref(cat);
			}
			if (res==0)
				Global.pProject->SetTempCatFile(tmp_file);
			g_free(tmp_file);
		} else {
			res = CMPACK_ERR_REF_NOT_FOUND;
		}
	}
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		return res;
	}

	// Set configuration
	cmpack_match_set_maxstars(m_Match, CConfig::GetInt("Matching", "RStars", 10));
	cmpack_match_set_vertices(m_Match, CConfig::GetInt("Matching", "IStars", 5));
	cmpack_match_set_threshold(m_Match, CConfig::GetDbl("Matching", "Clip", 2.5));
	cmpack_match_set_method(m_Match, CConfig::GetInt("Matching", "SpFields", 0));
	cmpack_match_set_maxoffset(m_Match, CConfig::GetDbl("Matching", "MaxOffset", 2));

	// Output handling
	if (pProgress) {
		CmpackConsole *con = cmpack_con_init_cb(CProgressDlg::OutputProc, pProgress);
		if (CConfig::GetBool("Environment", "Debug"))
			cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
		cmpack_match_set_console(m_Match, con);
		cmpack_unref(con);
	}

	return res;
}

void CMatchingProc::ConvertCatFile(const gchar *orig_file, const gchar *temp_file)
{
	bool ok = false;
	char *fts_path, *src_path, *dst_path;
	gsize length;
	gchar *contents;

	// Delete existing files
	g_unlink(temp_file);
	fts_path = SetFileExtension(temp_file, "fts");
	g_unlink(fts_path);
	g_free(fts_path);

	// Copy catalogue photometry file
	if (g_file_test(orig_file, G_FILE_TEST_IS_REGULAR)) {
		if (g_file_get_contents(orig_file, &contents, &length, NULL)) {
			g_file_set_contents(temp_file, contents, length, NULL);
			g_free(contents);
			ok = true;
		}
	}

	// Copy catalogue image file
	if (ok) {
		src_path = SetFileExtension(orig_file, "fts");
		dst_path = SetFileExtension(temp_file, "fts");
		if (g_file_test(src_path, G_FILE_TEST_IS_REGULAR)) {
			if (g_file_get_contents(src_path, &contents, &length, NULL)) {
				g_file_set_contents(dst_path, contents, length, NULL);
				g_free(contents);
			}
		}
		g_free(src_path);
		g_free(dst_path);
	}
}

int CMatchingProc::Execute(GtkTreePath *pPath)
{
	int res, nstars, mstars, state;
	char *f, *msg, *fpath, txt[128];
	GtkTreeModel *pList = Global.pProject->FileList();
	GtkTreeIter iter;
	CmpackPhtFile *pht;

	mstars = nstars = 0;
	gtk_tree_model_get_iter(pList, &iter, pPath);

	// Check frame status
	state = Global.pProject->GetState(&iter);
	if ((state & CFILE_CONVERSION)==0) {
		msg = "Image hasn't been converted.";
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		return -1;
	}
	if ((state & CFILE_PHOTOMETRY)==0) {
		msg = "Missing photometry file.";
		Global.pProject->SetError(&iter, msg);
		m_Progress->Print(msg);
		return -1;
	}

	// Make photometry file
	fpath = Global.pProject->GetPhotFile(pPath);
	f = g_locale_from_utf8(fpath, -1, NULL, NULL, NULL);
	res = cmpack_pht_load(&pht, f, CMPACK_LOAD_DEFAULT);
	g_free(f);
	if (res==0)
		res = cmpack_match(m_Match, pht, &mstars);
	if (res==0) {
		// Write the result
		f = g_locale_from_utf8(fpath, -1, NULL, NULL, NULL);
		res = cmpack_pht_save(pht, f);
		g_free(f);
	}
	if (res==0) {
		// Success
		Global.pProject->SetResult(&iter, "Matching OK", CFILE_MATCHING, CFILE_MATCHING);
		gtk_tree_model_get(pList, &iter, FRAME_STARS, &nstars, -1);
		gtk_list_store_set(GTK_LIST_STORE(pList), &iter, FRAME_MSTARS, mstars, -1);
		sprintf(txt, "%d stars matched (%.0f %%).", mstars, 100.0*((double)mstars/nstars));
		m_Progress->Print(txt);
	} else {
		// Failed
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg, CFILE_MATCHING, 0);
		cmpack_free(msg);
		gtk_list_store_set(GTK_LIST_STORE(pList), &iter, FRAME_MSTARS, 0, -1);
	}
	cmpack_unref(pht);
	g_free(fpath);
	return res;
}

//-----------------------   MASTER BIAS CONTEXT   ----------------------

//
// Constructor
//
CMasterBiasProc::CMasterBiasProc():m_Progress(NULL)
{
	m_MBias = cmpack_mbias_init();
}


//
// Destructor
//
CMasterBiasProc::~CMasterBiasProc()
{
	cmpack_unref(m_MBias);
}


//
// Initialization
//
int CMasterBiasProc::Open(CProgressDlg *pProgress, const gchar *outfile)
{
	int res;
	char txt[512], *msg;
	CmpackCcdFile *file;

	m_Progress = pProgress;

	if (!outfile || strlen(outfile)==0)
		return CMPACK_ERR_CANT_OPEN_OUT;

	sprintf(txt, "Output file: %s", outfile);
	m_Progress->Print(txt);

	// Image data type
	CmpackBitpix bitpix = (CmpackBitpix)CConfig::GetInt("MasterBias", "Format", CMPACK_BITPIX_AUTO);
	cmpack_mbias_set_bitpix(m_MBias, bitpix);

	// Frame border
	CmpackBorder border;
	cmpack_border_set(&border, CConfig::GetInt("CCD", "BorderLeft", 0),
		CConfig::GetInt("CCD", "BorderTop", 0),
		CConfig::GetInt("CCD", "BorderRight", 0),
		CConfig::GetInt("CCD", "BorderBottom", 0));
	cmpack_mbias_set_border(m_MBias, &border);

	// Output file
	gchar *f = g_locale_from_utf8(outfile, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&file, f, CMPACK_OPEN_CREATE);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		return res;
	}
	res = cmpack_mbias_open(m_MBias, file);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		cmpack_unref(file);
		file = NULL;
		return res;
	}
	cmpack_unref(file);

	// Output handling
	if (pProgress) {
		CmpackConsole *con = cmpack_con_init_cb(CProgressDlg::OutputProc, pProgress);
		if (CConfig::GetBool("Environment", "Debug"))
			cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
		cmpack_mbias_set_console(m_MBias, con);
		cmpack_unref(con);
	}

	return 0;
}


//
// Flat correction
//
int CMasterBiasProc::Add(GtkTreePath *pPath)
{
	int res, state;
	char *msg, *tpath, *f;
	GtkTreeModel *pList = Global.pProject->FileList();
	GtkTreeIter iter;
	CmpackCcdFile *infile;

	gtk_tree_model_get_iter(pList, &iter, pPath);

	// Check frame status
	state = Global.pProject->GetState(&iter);
	if ((state & CFILE_CONVERSION)==0) 
		return -1;

	// Open source file 
	tpath = Global.pProject->GetImageFile(&iter);
	f = g_locale_from_utf8(tpath, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&infile, f, CMPACK_OPEN_READONLY);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg);
		cmpack_free(msg);
		g_free(tpath);
		return res;
	}
	g_free(tpath);

	// Read input file
	res = cmpack_mbias_read(m_MBias, infile);
	if (res!=0) {
		// Failed
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg);
		cmpack_free(msg);
		cmpack_unref(infile);
		return res;
	}
	cmpack_unref(infile);

	// Success
	return 0;
}

//
// Finalization
//
int CMasterBiasProc::Close(void)
{
	int res = cmpack_mbias_close(m_MBias);
	if (res!=0) {
		char *msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
	}
	return res;
}

//-----------------------   MASTER DARK CONTEXT   ----------------------

//
// Constructor
//
CMasterDarkProc::CMasterDarkProc():m_Progress(NULL)
{
	m_MDark = cmpack_mdark_init();
}


//
// Destructor
//
CMasterDarkProc::~CMasterDarkProc()
{
	cmpack_unref(m_MDark);
}


//
// Initialization
//
int CMasterDarkProc::Open(CProgressDlg *pProgress, const gchar *outfile)
{
	int res;
	char txt[512], *msg;
	CmpackCcdFile *file;

	m_Progress = pProgress;

	if (!outfile || strlen(outfile)==0)
		return CMPACK_ERR_CANT_OPEN_OUT;

	sprintf(txt, "Output file: %s", outfile);
	m_Progress->Print(txt);

	// Scalable darks
	bool scalable = CConfig::GetBool("MasterDark", "Scalable");
	cmpack_mdark_set_scalable(m_MDark, scalable);

	// Image data type
	CmpackBitpix bitpix = (CmpackBitpix)CConfig::GetInt("MasterDark", "Format", CMPACK_BITPIX_AUTO);
	cmpack_mdark_set_bitpix(m_MDark, bitpix);

	// Frame border
	CmpackBorder border;
	cmpack_border_set(&border, CConfig::GetInt("CCD", "BorderLeft", 0),
		CConfig::GetInt("CCD", "BorderTop", 0),
		CConfig::GetInt("CCD", "BorderRight", 0),
		CConfig::GetInt("CCD", "BorderBottom", 0));
	cmpack_mdark_set_border(m_MDark, &border);

	// Output file
	gchar *f = g_locale_from_utf8(outfile, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&file, f, CMPACK_OPEN_CREATE);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		return res;
	}
	res = cmpack_mdark_open(m_MDark, file);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		cmpack_unref(file);
		file = NULL;
		return res;
	}
	cmpack_unref(file);

	// Output handling
	if (pProgress) {
		CmpackConsole *con = cmpack_con_init_cb(CProgressDlg::OutputProc, pProgress);
		if (CConfig::GetBool("Environment", "Debug"))
			cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
		cmpack_mdark_set_console(m_MDark, con);
		cmpack_unref(con);
	}

	return res;
}


//
// Flat correction
//
int CMasterDarkProc::Add(GtkTreePath *pPath)
{
	int res, state;
	char *msg, *tpath, *f;
	GtkTreeModel *pList = Global.pProject->FileList();
	GtkTreeIter iter;
	CmpackCcdFile *infile;

	gtk_tree_model_get_iter(pList, &iter, pPath);

	// Check frame status
	state = Global.pProject->GetState(&iter);
	if ((state & CFILE_CONVERSION)==0) 
		return -1;

	// Open source file 
	tpath = Global.pProject->GetImageFile(&iter);
	f = g_locale_from_utf8(tpath, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&infile, f, CMPACK_OPEN_READONLY);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg);
		cmpack_free(msg);
		g_free(tpath);
		return res;
	}
	g_free(tpath);

	// Read input file
	res = cmpack_mdark_read(m_MDark, infile);
	if (res!=0) {
		// Failed
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg);
		cmpack_free(msg);
		cmpack_unref(infile);
		return res;
	}
	cmpack_unref(infile);

	// Success
	return 0;
}

//
// Finalization
//
int CMasterDarkProc::Close(void)
{
	int res = cmpack_mdark_close(m_MDark);
	if (res!=0) {
		char *msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
	}
	return res;
}

//-----------------------   MASTER FLAT CONTEXT   ----------------------

//
// Constructor
//
CMasterFlatProc::CMasterFlatProc():m_Progress(NULL)
{
	m_MFlat = cmpack_mflat_init();
}


//
// Destructor
//
CMasterFlatProc::~CMasterFlatProc()
{
	cmpack_unref(m_MFlat);
}


//
// Initialization
//
int CMasterFlatProc::Open(CProgressDlg *pProgress, const gchar *outfile)
{
	int res;
	char txt[512], *msg;
	CmpackCcdFile *file;

	m_Progress = pProgress;

	if (!outfile || strlen(outfile)==0)
		return CMPACK_ERR_CANT_OPEN_OUT;

	sprintf(txt, "Output file: %s", outfile);
	m_Progress->Print(txt);

	// Mean output level
	double level = CConfig::GetDbl("MasterFlat", "Level", 10000.0);
	cmpack_mflat_set_level(m_MFlat, level);

	// Image data type
	CmpackBitpix bitpix = (CmpackBitpix)CConfig::GetInt("MasterFlat", "Format", CMPACK_BITPIX_AUTO);
	cmpack_mflat_set_bitpix(m_MFlat, bitpix);

	// Frame border
	CmpackBorder border;
	cmpack_border_set(&border, CConfig::GetInt("CCD", "BorderLeft", 0),
		CConfig::GetInt("CCD", "BorderTop", 0),
		CConfig::GetInt("CCD", "BorderRight", 0),
		CConfig::GetInt("CCD", "BorderBottom", 0));
	cmpack_mflat_set_border(m_MFlat, &border);

	// Output file
	gchar *f = g_locale_from_utf8(outfile, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&file, f, CMPACK_OPEN_CREATE);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		return res;
	}
	res = cmpack_mflat_open(m_MFlat, file);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		cmpack_unref(file);
		file = NULL;
		return res;
	}
	cmpack_unref(file);

	// Output handling
	if (pProgress) {
		CmpackConsole *con = cmpack_con_init_cb(CProgressDlg::OutputProc, pProgress);
		if (CConfig::GetBool("Environment", "Debug"))
			cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
		cmpack_mflat_set_console(m_MFlat, con);
		cmpack_unref(con);
	}

	return res;
}


//
// Flat correction
//
int CMasterFlatProc::Add(GtkTreePath *pPath)
{
	int res, state;
	char *msg, *tpath, *f;
	GtkTreeModel *pList = Global.pProject->FileList();
	GtkTreeIter iter;
	CmpackCcdFile *infile;

	gtk_tree_model_get_iter(pList, &iter, pPath);

	// Check frame status
	state = Global.pProject->GetState(&iter);
	if ((state & CFILE_CONVERSION)==0) 
		return -1;

	// Open source file 
	tpath = Global.pProject->GetImageFile(&iter);
	f = g_locale_from_utf8(tpath, -1, NULL, NULL, NULL);
	res = cmpack_ccd_open(&infile, f, CMPACK_OPEN_READONLY);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg);
		cmpack_free(msg);
		g_free(tpath);
		return res;
	}
	g_free(tpath);

	// Read input file
	res = cmpack_mflat_read(m_MFlat, infile);
	if (res!=0) {
		// Failed
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg);
		cmpack_free(msg);
		cmpack_unref(infile);
		return res;
	}
	cmpack_unref(infile);

	// Success
	return 0;
}

//
// Finalization
//
int CMasterFlatProc::Close(void)
{
	int res = cmpack_mflat_close(m_MFlat);
	if (res!=0) {
		char *msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
	}
	return res;
}

//-----------------------   MAKE TABLE OF APERTURES   ----------------------

//
// Constructor
//
CListProc::CListProc():m_Progress(NULL)
{
	m_Table = NULL;
	m_List = cmpack_list_init();
	m_HelCor = cmpack_hcorr_init();
	m_AirMass = cmpack_amass_init();
}


//
// Destructor
//
CListProc::~CListProc()
{
	cmpack_unref(m_Table);
	cmpack_unref(m_List);
	cmpack_unref(m_HelCor);
	cmpack_unref(m_AirMass);
}


//
// Initialization
//
int CListProc::Open(CProgressDlg *pProgress, CmpackTable *pTable, CmpackListFormat format)
{
	int		count, stars[MAX_SELECTION], res = 0;
	double	ra, dec, lon, lat;
	char	*msg;

	m_Progress = pProgress;
	m_Table = (CmpackTable*)cmpack_addref(pTable);

	// Set aperture
	if (format!=CMPACK_LIST_STDDEV) {
		int aperture = Global.pProject->Aperture();
		cmpack_list_set_aperture(m_List, aperture);
	}

	// Set selected stars
	if (format!=CMPACK_LIST_TRACKLST) {
		const CSelection *sel = Global.pProject->Selection();
		count = sel->GetStarList(ST_VARIABLE, stars, MAX_SELECTION);
		cmpack_list_set_var(m_List, stars, count);
		count = sel->GetStarList(ST_COMPARISON, stars, MAX_SELECTION);
		cmpack_list_set_comp(m_List, stars, count);
		count = sel->GetStarList(ST_CHECK, stars, MAX_SELECTION);
		cmpack_list_set_check(m_List, stars, count);
	}

	// Output handling
	if (pProgress) {
		CmpackConsole *con = cmpack_con_init_cb(CProgressDlg::OutputProc, pProgress);
		if (CConfig::GetBool("Environment", "Debug"))
			cmpack_con_set_level(con, CMPACK_LEVEL_DEBUG);
		cmpack_list_set_console(m_List, con);
		cmpack_hcorr_set_console(m_HelCor, con);
		cmpack_amass_set_console(m_AirMass, con);
		cmpack_unref(con);
	}

	// Heliocentric correction
	m_DoHelCor = m_AddHelCor = m_AddAirMass = m_AddAltitude = false;
	if (format==CMPACK_LIST_DIFFMAG || format==CMPACK_LIST_INSTMAG) {
		m_DoHelCor = CConfig::GetInt("LightCurve", "JDFormat")==1;
		m_AddHelCor = (CConfig::GetInt("LightCurve", "OptColumns") & 0x01)!=0;
		m_AddAirMass = (CConfig::GetInt("LightCurve", "OptColumns") & 0x02)!=0;
		m_AddAltitude = (CConfig::GetInt("LightCurve", "OptColumns") & 0x04)!=0;
	}
	if (m_DoHelCor || m_AddHelCor || m_AddAirMass || m_AddAltitude) {
		CObjectCoords obj = *Global.pProject->ObjectCoords();
		if (cmpack_strtora(obj.RA(), &ra)!=0)
			return CMPACK_ERR_INVALID_RA;
		cmpack_hcorr_set_ra(m_HelCor, ra);
		cmpack_amass_set_ra(m_AirMass, ra);
		if (cmpack_strtodec(obj.Dec(), &dec)!=0)
			return CMPACK_ERR_INVALID_DEC;
		cmpack_hcorr_set_dec(m_HelCor, dec);
		cmpack_amass_set_dec(m_AirMass, dec);
	}
	if (m_AddAirMass || m_AddAltitude) {
		CObserverCoords obs;
		if (Global.pProject->UseDefaultLocation())
			obs = Global.pProject->DefaultLocation();
		else
			obs = *Global.pProject->UserLocation();
		if (cmpack_strtolon(obs.Lon(), &lon)!=0)
			return CMPACK_ERR_INVALID_LON;
		cmpack_amass_set_lon(m_AirMass, lon);
		if (cmpack_strtolat(obs.Lat(), &lat)!=0)
			return CMPACK_ERR_INVALID_LAT;
		cmpack_amass_set_lat(m_AirMass, lat);
	}
	/* Check coordinates */
	if (m_DoHelCor || m_AddHelCor) 
		res = cmpack_hcorr_jd(m_HelCor, 2450000.0, NULL);
	if (res==0 && (m_AddAirMass || m_AddAltitude)) 
		res = cmpack_amass_jd(m_AirMass, 2450000.0, NULL, NULL);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
		return res;
	}

	/* Start listing */
	res = cmpack_list_open(m_List, pTable, format, 0);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
	}
	return res;
}


//
// Flat correction
//
int CListProc::Add(GtkTreePath *pPath)
{
	int res, frameid, state;
	char *msg, *tpath, *f;
	GtkTreeModel *pList = Global.pProject->FileList();
	GtkTreeIter iter;
	CmpackPhtFile *infile;

	gtk_tree_model_get_iter(pList, &iter, pPath);

	// Check frame status
	state = Global.pProject->GetState(&iter);
	if ((state & CFILE_CONVERSION)==0) 
		return -1;

	// Open source file 
	frameid = Global.pProject->GetFrameID(&iter);
	tpath = Global.pProject->GetPhotFile(&iter);
	f = g_locale_from_utf8(tpath, -1, NULL, NULL, NULL);
	res = cmpack_pht_load(&infile, f, CMPACK_LOAD_DEFAULT);
	g_free(f);
	if (res!=0) {
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg);
		cmpack_free(msg);
		g_free(tpath);
		return res;
	}
	g_free(tpath);

	// Read input file
	res = cmpack_list_read(m_List, infile, frameid);
	if (res!=0) {
		// Failed
		msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		Global.pProject->SetError(&iter, msg);
		cmpack_free(msg);
		cmpack_unref(infile);
		return res;
	}
	cmpack_unref(infile);

	// Success
	return 0;
}

//
// Finalization
//
int CListProc::Close(void)
{
	int res;

	// Finish writing the file 
	res = cmpack_list_close(m_List);
	// Airmass coefficient and altitude
	if (res==0 && (m_AddAirMass || m_AddAltitude)) {
		res = cmpack_amass(m_AirMass, m_Table, 0, (CmpackAMassFlags)(
			(m_AddAirMass ? 0 : CMPACK_AMASS_NOAIRMASS) |
			(m_AddAltitude ? CMPACK_AMASS_ALTITUDE : 0)));
	}
	// Heliocentric correction
	if (res==0 && (m_DoHelCor || m_AddHelCor)) {
		res = cmpack_hcorr(m_HelCor, m_Table, 0, (CmpackHCorrFlags)(
			(m_DoHelCor ? 0 : CMPACK_HCORR_NOMODIFY) |
			(m_AddHelCor ? CMPACK_HCORR_ADDVALUE : 0)));
	}
	if (res!=0) {
		char *msg = cmpack_formaterror(res);
		m_Progress->Print(msg);
		cmpack_free(msg);
	}
	cmpack_unref(m_Table);
	m_Table = NULL;
	return res;
}

//-----------------------   AIR-MASS COEFFICIENT TABLE   ----------------------

//
// Constructor
//
CAirMassProc::CAirMassProc()
{
	m_Table = NULL;
	m_AirMass = cmpack_amass_init();
}


//
// Destructor
//
CAirMassProc::~CAirMassProc()
{
	cmpack_unref(m_Table);
	cmpack_unref(m_AirMass);
}


//
// Initialization
//
int CAirMassProc::Open(CmpackTable *pTable)
{
	int		res;
	double	ra, dec, lon, lat;
	CObjectCoords obj;
	CObserverCoords obs;

	m_Table = (CmpackTable*)cmpack_addref(pTable);

	// Object coordinates
	obj = *Global.pProject->ObjectCoords();
	if (cmpack_strtora(obj.RA(), &ra)!=0)
		return CMPACK_ERR_INVALID_RA;
	cmpack_amass_set_ra(m_AirMass, ra);
	if (cmpack_strtodec(obj.Dec(), &dec)!=0)
		return CMPACK_ERR_INVALID_DEC;
	cmpack_amass_set_dec(m_AirMass, dec);

	// Observer coordinates
	if (Global.pProject->UseDefaultLocation())
		obs = Global.pProject->DefaultLocation();
	else
		obs = *Global.pProject->UserLocation();
	if (cmpack_strtolon(obs.Lon(), &lon)!=0)
		return CMPACK_ERR_INVALID_LON;
	cmpack_amass_set_lon(m_AirMass, lon);
	if (cmpack_strtolat(obs.Lat(), &lat)!=0)
		return CMPACK_ERR_INVALID_LAT;
	cmpack_amass_set_lat(m_AirMass, lat);
	
	//Check coordinates 
	res = cmpack_amass_jd(m_AirMass, 2450000.0, NULL, NULL);
	if (res!=0) 
		return res;

	// Initialize table
	cmpack_tab_clear(m_Table);
	cmpack_tab_set_type(m_Table, CMPACK_TABLE_AIRMASS);
	cmpack_tab_add_column(m_Table, "JD", CMPACK_COLUMN_DOUBLE, 
		cmpack_get_param_long(CMPACK_PARAM_JD_PRECISION));
	cmpack_tab_add_column(m_Table, "AIRMASS", CMPACK_COLUMN_DOUBLE, 
		cmpack_get_param_long(CMPACK_PARAM_AMASS_PRECISION));
	cmpack_tab_add_column(m_Table, "ALTITUDE", CMPACK_COLUMN_DOUBLE, 
		cmpack_get_param_long(CMPACK_PARAM_ALT_PRECISION));
	return 0;
}


//
// Flat correction
//
int CAirMassProc::Add(GtkTreePath *pPath)
{
	int row, state;
	double jd, x, alt;
	GtkTreeModel *pList = Global.pProject->FileList();
	GtkTreeIter iter;

	gtk_tree_model_get_iter(pList, &iter, pPath);
	gtk_tree_model_get(pList, &iter, FRAME_STATE, &state, FRAME_JULDAT, &jd, -1);

	// Check frame status
	if ((state & CFILE_CONVERSION)==0) 
		return -1;

	row = cmpack_tab_add_row(m_Table);
	cmpack_tab_ptdd(m_Table, row, 0, jd);
	if (cmpack_amass_jd(m_AirMass, jd, &alt, &x)==0) {
		cmpack_tab_ptdd(m_Table, row, 1, x);
		cmpack_tab_ptdd(m_Table, row, 2, alt);
	}

	// Success
	return 0;
}

//
// Finalization
//
void CAirMassProc::Close(void)
{
	cmpack_unref(m_Table);
	m_Table = NULL;
}
