/**************************************************************

utils.cpp (C-Munipack project)
Helper functions
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <cmunipack.h>

#include "utils.h"
#include "configuration.h"

#define BLOCKSIZE 2048

/* File access codes */
#ifndef S_IRUSR
#define S_IRUSR _S_IREAD
#endif
#ifndef S_IWUSR
#define S_IWUSR _S_IWRITE
#endif
#ifndef S_IXUSR
#define S_IXUSR _S_IEXEC
#endif

#ifndef S_IRGRP
#define S_IRGRP _S_IREAD
#endif
#ifndef S_IXGRP
#define S_IXGRP _S_IEXEC
#endif
#ifndef S_IROTH
#define S_IROTH _S_IREAD
#endif
#ifndef S_IXOTH
#define S_IXOTH _S_IEXEC
#endif

// Default apertures
static const double DefApertureRadius[MAX_APERTURES] = 
{ 
	2.0, 2.7273, 3.8182, 5.2727, 7.0909, 9.2727, 11.8182, 14.7273, 18.0, 21.6364, 25.6364, 30.0 
};

const char *get_user_dir(void)
{
	static char *user_dir = NULL;
	if (!user_dir) {
		const char *dir = g_get_user_config_dir();
		user_dir = g_build_filename(dir, "c-munipack-1.2", NULL);
	}
	return user_dir;
}

const char *get_temp_dir(void)
{
	static char *temp_dir = NULL;
	if (!temp_dir) {
		temp_dir = g_build_filename(get_user_dir(), "temp", NULL);
	}
	return temp_dir;
}

const char *get_share_dir(void)
{
#ifdef _WIN32
	static char *icons_dir = NULL;
	if (!icons_dir)
		icons_dir = g_build_filename("..", "share", NULL);
	return icons_dir;
#else
	return DATA_DIR;
#endif
}

static const char *get_help_file(void)
{
	static char *help_file = NULL;
	if (!help_file)
#ifdef _WIN32
		help_file = g_build_filename("..", "doc", "htmlhelp", "cmpack12.chm", NULL);
#else
		help_file = g_build_filename(DATA_DIR, "doc", "html", "index.html", NULL);
#endif
	return help_file;
}

char *get_icon_file(const gchar *icon)
{
	char fname[512];
	g_snprintf(fname, 512, "%s.png", icon);
	return g_build_filename(get_share_dir(), "icons", fname, NULL);
}

// Make directory tree
bool force_directory(const gchar *path)
{
	if (!g_file_test(path, G_FILE_TEST_EXISTS))
		return g_mkdir_with_parents(path, S_IRUSR | S_IXUSR | S_IWUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)==0;
	else
		return true;
}

// Open help
void ShowHelp(GtkWindow *pParent)
{
	const gchar *cmd[3];
	GError *err = NULL;

	if (!g_file_test(get_help_file(), G_FILE_TEST_IS_REGULAR)) {
		ShowError(pParent, "The help file(s) are not installed.");
		return;
	}

#ifdef _WIN32
	cmd[0] = "hh.exe";
#else
	cmd[0] = "xdg-open";
#endif
	cmd[1] = get_help_file();
	cmd[2] = NULL;
	if (!gdk_spawn_on_screen(gtk_window_get_screen(pParent), NULL, 
		(gchar**)cmd, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, &err)) {
			ShowError(pParent, err->message);
		g_error_free(err);
	}
}

// Show message box with error message
void ShowError(GtkWindow *pParent, const char *message)
{
	GtkWidget *pDialog = gtk_message_dialog_new(pParent, GTK_DIALOG_DESTROY_WITH_PARENT, 
		GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, message);
	gtk_window_set_title(GTK_WINDOW(pDialog), "Error");
	gtk_dialog_run(GTK_DIALOG(pDialog));
	gtk_widget_destroy(pDialog);
}

// Show message box with warning message
void ShowWarning(GtkWindow *pParent, const char *message)
{
	GtkWidget *pDialog = gtk_message_dialog_new(pParent, GTK_DIALOG_DESTROY_WITH_PARENT, 
		GTK_MESSAGE_WARNING, GTK_BUTTONS_OK, message);
	gtk_window_set_title(GTK_WINDOW(pDialog), "Warning");
	gtk_dialog_run(GTK_DIALOG(pDialog));
	gtk_widget_destroy(pDialog);
}

// Show message box with information
void ShowInformation(GtkWindow *pParent, const char *message)
{
	GtkWidget *pDialog = gtk_message_dialog_new(pParent, GTK_DIALOG_DESTROY_WITH_PARENT, 
		GTK_MESSAGE_INFO, GTK_BUTTONS_OK, message);
	gtk_window_set_title(GTK_WINDOW(pDialog), "Information");
	gtk_dialog_run(GTK_DIALOG(pDialog));
	gtk_widget_destroy(pDialog);
}

// Show configuration message box
bool ShowConfirmation(GtkWindow *pParent, const char *message)
{
	int res;
	GtkWidget *pDialog = gtk_message_dialog_new(pParent, GTK_DIALOG_DESTROY_WITH_PARENT, 
		GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO, message);
	gtk_window_set_title(GTK_WINDOW(pDialog), "Confirmation");
	res = gtk_dialog_run(GTK_DIALOG(pDialog));
	gtk_widget_destroy(pDialog);
	return (res==GTK_RESPONSE_YES);
}

// Set/change file extension
char *SetFileExtension(const char *path, const char *ext)
{
	char *dir, *basename, *newname, *ptr, *res;

	res = NULL;
	if (path && ext) {
		basename = g_path_get_basename(path);
		if (basename) {
			// Trim existing extension
			ptr = strrchr(basename, '.');
			if (ptr)
				*ptr = '\0';
			newname = (char*)g_malloc((strlen(basename)+strlen(ext)+2)*sizeof(char));
			strcpy(newname, basename);
			if (strlen(ext)>0) {
				strcat(newname, ".");
				strcat(newname, ext);
			} 
			dir = g_path_get_dirname(path);
			res = g_build_filename(dir, newname, NULL);
			g_free(dir);
			g_free(newname);
			g_free(basename);
		}
	}
	return res;
}

// Set/change file extension
bool CheckFileExtension(const char *path, const char *ext)
{
	bool res;
	char *basename, *ptr;

	res = false;
	if (path && ext) {
		basename = g_path_get_basename(path);
		if (basename) {
			// Trim existing extension
			ptr = strrchr(basename, '.');
			if (ptr) 
				res = (strcmp(ptr+1, ext)==0);
			else
				res = (strlen(ext)==0);
			g_free(basename);
		}
	}
	return res;
}

GtkWidget *get_dialog_widget_by_response(GtkWidget *pDlg, gint response_id)
{
	GtkWidget *retval = NULL;

	GList *children = gtk_container_get_children(GTK_CONTAINER((GTK_DIALOG(pDlg)->action_area)));
	for (GList *l=children; l!=NULL; l=l->next) {
		if (gtk_dialog_get_response_for_widget(GTK_DIALOG(pDlg), GTK_WIDGET(l->data)) == response_id) {
			retval = GTK_WIDGET(l->data);
			break;
		}
	}
	g_list_free(children);
	return retval;
}

// Round number to the integer with limit checking
int RoundToInt(double x)
{
	if (x>INT_MAX)
		return INT_MAX;
	else if (x<INT_MIN)
		return INT_MIN;
	else if (x<0)
		return (int)(x-0.5);
	else
		return (int)(x+0.5);
}

// Autodetect file type
tFileType FileType(const char *fpath)
{
	char buffer[BLOCKSIZE], *f1;
	int bytes, filesize;

	tFileType type = FILE_TYPE_UNKNOWN;
	f1 = g_locale_from_utf8(fpath, -1, NULL, NULL, NULL);
	if (f1) {
		FILE *f = fopen(f1, "rb");
		if (f) {
			fseek(f, 0, SEEK_END);
			filesize = ftell(f);
			fseek(f, 0, SEEK_SET);
			bytes = fread(buffer, 1, BLOCKSIZE, f);
			if (CheckFileExtension(fpath, "srt") || 
				CheckFileExtension(fpath, "pht") || 
				CheckFileExtension(fpath, "mat") ||
				CheckFileExtension(fpath, "xml")) {
					// XML based files
					if (cmpack_pht_test_buffer(buffer, bytes, filesize))
						type = FILE_TYPE_PHOT;
					else if (cmpack_cat_test_buffer(buffer, bytes, filesize))
						type = FILE_TYPE_CAT;
			} else 
			if (CheckFileExtension(fpath, "dat") ||
				CheckFileExtension(fpath, "txt")) {
					// Text based files
					if (cmpack_tab_test_buffer(buffer, bytes, filesize))
						type = FILE_TYPE_TABLE;
					else if (cmpack_all_test_buffer(buffer, bytes, filesize))
						type = FILE_TYPE_READALL;
			} else {
				// Binary files
				if (cmpack_ccd_test_buffer(buffer, bytes, filesize))
					type = FILE_TYPE_IMAGE;
			}
			fclose(f);
		}
		g_free(f1);
	}
	return type;
}

// Get aperture radius
double GetApertureRadius(int aperture)
{
	char key[64];

	if (aperture>0 && aperture<=MAX_APERTURES) {
		sprintf(key, "Aperture%d", aperture);
		return CConfig::GetDbl("Photometry", key, DefApertureRadius[aperture-1]);
	}
	return 0;
}

double Min(double a, double b)
{
	return (a<b ? a : b);
}

double Max(double a, double b)
{
	return (a>b ? a : b);
}

int Min(int a, int b)
{
	return (a<b ? a : b);
}

int Max(int a, int b)
{
	return (a>b ? a : b);
}

double LimitValue(double value, double min, double max)
{
	if (value > max)
		return max;
	else if (value < min)
		return min;
	else
		return value;
}

int LimitValue(int value, int min, int max)
{
	if (value > max)
		return max;
	else if (value < min)
		return min;
	else
		return value;
}

// Compare strings, they can be NULL
int StrCmp0(const char *str1, const char *str2)
{
	if (str1==NULL && str2==NULL)
		return 0;
	else if (str1==NULL)
		return -1;
	else if (str2==NULL)
		return 1;
	else
		return strcmp(str1, str2);
}

// Make a file name from object's name
char *MakeFileName(const char *objname, const char *ext)
{
	int len;
	char ch, *buf, *dst;
	const char *ptr;

	if (objname) {
		len = strlen(objname)+1;
		if (ext) 
			len += strlen(ext)+1;
		buf = (char*)g_malloc(len);
		for (ptr=objname, dst=buf; *ptr!='\0'; ptr++) {
			ch = *ptr;
			if (ch>=' ' && ch!='/' && ch!='\\' && ch!='?' && ch!='%' && ch!='*' &&
				ch!=':' && ch!='|' && ch!='"' && ch!='<' && ch!='>' && ch!='>')
				*dst++ = ch;
		}
		if (ext) {
			*dst++ = '.';
			strcpy(dst, ext);
		} else {
			*dst = '\0';
		}
		return buf;
	}
	return NULL;
}
